#! /usr/bin/perl -w
################################################################################
#
# Usage: node.query.ciscoios.pl
#
# This script is used to query switch properties and generic event and metric
# information.
#
#   Copyright (C) 2005-2016 by Adaptive Computing Enterprises, Inc. All Rights Reserved.
#
################################################################################
use strict;
use Expect;
use FindBin;
use lib "/make_install/lib/perl5", "/make_install/etc";             # Installed
use lib "${FindBin::Bin}/../../lib/perl5", "${FindBin::Bin}/../../etc";   # Configured
use Getopt::Long;
use Moab::Tools;    # Required before including config to set $homeDir
BEGIN { require "config.vlan.pl"; }
our ($logLevel, $router, $password, $loginPrompt, $timeout, $capacity);

if ($logLevel >= 1)
{
    logPrint("$0 ", join(' ', @ARGV), "\n");
}

if ($logLevel <= 1)
{
    # Turn off expect output from spawned commands
    $Expect::Log_Stdout = 0;
}

# Parse Command Line Arguments
my ($help);
my $usage = "Usage: $0\n";
GetOptions('help|?' => \$help) or logDie($usage);
die $usage if $help;

my $status = $loginPrompt;

# Connect to the switch router
my $exp = Expect->spawn("telnet $router")
  or logDie("Failed connecting to switch router: $!\n");

# Log into the router
$exp->expect(
    $timeout,
    [
        qr/Password: /i => sub {
            my $self = shift;
            $self->send("$password\n");
          }
    ]
  )
  or logDie("Unable to log into switch router\n");

# Check environmental information
$exp->expect(
    $timeout,
    [
        qr/$loginPrompt>/ => sub {
            my $self = shift;
            $self->send("show env all\n");
          }
    ]
  )
  or logDie("Unable to query environmental information\n");

$exp->notransfer(1);
$exp->expect($timeout, '-re', qr/$loginPrompt/)
  or logDie("Unable to capture output\n");
$exp->notransfer(0);

my ($fan, $temp, $power, $tmetric);
foreach my $line (split /\n/, $exp->exp_before())
{
    if ($line =~ /FAN is (\S+)/)
    {
         $fan = $1;
    }
    elsif ($line =~ /TEMPERATURE is (\S+)/)
    {
         $temp = $1;
    }
    elsif ($line =~ /POWER is (\S+)/)
    {
         $power = $1;
    }
    elsif ($line =~ /Temperature Value: (\d+) Degree Celsius/)
    {
         $tmetric = $1;
    }
}

# Handle fan event
if (defined $fan && $fan ne "OK")
{
  $status .= " GEVENT[FANFAIL]='fan alert on switch ($fan)'";
}

# Handle temperature event
if (defined $temp && $temp ne "OK")
{
  $status .= " GEVENT[TEMPFAIL]='temperature alert on switch ($temp)'";
}

# Handle power event
if (defined $power && $power ne "OK")
{
  $status .= " GEVENT[POWERFAIL]='power alert on switch ($power)'";
}

# Provide temperature metric
if (defined $tmetric)
{
  $status .= " GMETRIC[TEMP]=$tmetric";
}

# Check utilization
$exp->expect(
    $timeout,
    [
        qr/$loginPrompt>/ => sub {
            my $self = shift;
            $self->send("show controllers utilization | include Bandwidth\n");
          }
    ]
  )
  or logDie("Unable to query controllers utilization\n");

$exp->notransfer(1);
$exp->expect($timeout, '-re', qr/$loginPrompt/)
  or logDie("Unable to capture output\n");
$exp->notransfer(0);

my ($receive, $transmit);
foreach my $line (split /\n/, $exp->exp_before())
{
    if ($line =~ /Switch Receive Bandwidth Percentage Utilization\s+: (\d+)/)
    {
         $receive = $1;
    }
    elsif ($line =~ /Switch Transmit Bandwidth Percentage Utilization\s+: (\d+)/)
    {
         $transmit = $1;
    }
}

# Post configured resources
if (defined $capacity)
{
  $status .= " CRES=bandwidth:$capacity";
}

# Post configured and available resources
if (defined $receive && defined $transmit)
{
  my $bandwidth = $capacity - $capacity * ($receive + $transmit) / 200;
  $status .= " ARES=bandwidth:$bandwidth";
}

# Close the connection
$exp->send("exit\n");
$exp->soft_close();

# Print the switch status to STDOUT and exit
print $status, "\n";
exit 0;
1;

__END__

=head1 NAME

=head1 SYNOPSIS

=head1 DESCRIPTION

=head2 EXPORT

=head2 EXPORT_OK

=head1 SEE ALSO

Additional information can be found at http://www.adaptivecomputing.com/

=head1 AUTHOR

Adaptive Computing, Inc., E<lt>http://www.adaptivecomputing.com/E<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2005-2016 by Adaptive Computing Enterprises, Inc. All Rights Reserved.

=cut
