#! /usr/bin/perl -w
################################################################################
#
# Usage: cisco.create.ciscoios.pl <vlan id> <node list>
#
# FORMAT:  <node list> -> <node id>[,<node id>]...
#
# NOTE: Currently, router is set in the config.vlan.pl file
# and it is ignored from the nodes file
#
#   Copyright (C) 2005-2016 by Adaptive Computing Enterprises, Inc. All Rights Reserved.
#
################################################################################
use strict;
use Expect;
use FindBin;
use lib "/make_install/lib/perl5", "/make_install/etc";             # Installed
use lib "${FindBin::Bin}/../../lib/perl5", "${FindBin::Bin}/../../etc";   # Configured
use Getopt::Long;
use Moab::Tools;    # Required before including config to set $homeDir
BEGIN { require "config.vlan.pl"; }
our ($logLevel, $nodesFile, $router, $globalVlanId, $password, $secret,
    $loginPrompt, $timeout);

if ($logLevel >= 1)
{
    logPrint("$0 ", join(' ', @ARGV), "\n");
}

if ($logLevel <= 1)
{
    # Turn off expect output from spawned commands
    $Expect::Log_Stdout = 0;
}

# Parse Command Line Arguments
my ($help);
my $usage = "Usage: $0 <vlan id> <node list>\n";
GetOptions('help|?' => \$help) or die $usage;
die $usage if $help;
die $usage unless @ARGV == 2;
my ($vlanId, $nodeList) = @ARGV;

# Refuse to create the global vlan
if (defined $globalVlanId && $vlanId eq $globalVlanId)
{
    logDie("Creating the Global VLAN ($globalVlanId) is not allowed\n");
}

# Connect to the switch router
my $exp = Expect->spawn("telnet $router")
  or logDie("Failed connecting to switch router: $!\n");

# Log into the router
$exp->expect(
    $timeout,
    [
        qr/Password: /i => sub {
            my $self = shift;
            $self->send("$password\n");
          }
    ]
  )
  or logDie("Unable to log into switch router\n");

# Make sure vlan does not already exist
$exp->expect(
    $timeout,
    [
        qr/$loginPrompt>/ => sub {
            my $self = shift;
            $self->send("show vlan id $vlanId\n      "); # Spaces for --More--
          }
    ]
  )
  or logDie("Unable to check for existence of vlan $vlanId\n");
$exp->expect(
    $timeout,
    [
       qr/not found/i => sub {
            my $self = shift;
          }
    ]
  )
  or logDie("Vlan $vlanId already exists!\n");

# Enter privileged mode
$exp->expect(
    $timeout,
    [
        qr/$loginPrompt>/i => sub {
            my $self = shift;
            $self->send("enable\n");
            exp_continue;
          }
    ],
    [
        qr/Password: /i => sub {
            my $self = shift;
            $self->send("$secret\n");
          }
    ]
  )
  or logDie("Unable to enter privileged mode\n");

# Enter global configuration mode
$exp->expect(
    $timeout,
    [
        qr/$loginPrompt#/ => sub {
            my $self = shift;
            $self->send("config\n\n");
          }
    ]
  )
  or logDie("Unable to enter global configuration mode\n");

# Create the new vlan
$exp->expect(
    $timeout,
    [
        qr/$loginPrompt\(config\)#/ => sub {
            my $self = shift;
            $self->send("vlan $vlanId\n");
            exp_continue;
          }
    ],
    [
        qr/$loginPrompt\(config-vlan\)#/i => sub {
            my $self = shift;
            $self->send("exit\n");
          }
    ]
  )
  or logDie("Unable to create vlan $vlanId\n");

# Parse the nodes file and read in port information
my %nodeInfo = ();
if (open NODESFILE, "<$nodesFile")
{
    foreach my $line (<NODESFILE>)
    {
        $line =~ s/^\s+//;    # Remove leading whitespace
        $line =~ s/\#.*//;    # Strip out comments
        next unless $line;    # Ignore blank (or pure comment) lines
        my @words = split /\s+/, $line;
        my $nodeId = shift @words;
        foreach my $word (@words)
        {
            my ($key, $value) = split /=/, $word;
            $nodeInfo{$nodeId}->{$key} = $value;
        }
    }
    close NODESFILE;
}
else
{
    logPrint("Unable to open nodes file ($nodesFile) for reading: $!\n");
}

# Add ports to the vlan
foreach my $nodeId (split /,/, $nodeList)
{
    if (exists $nodeInfo{$nodeId} && defined $nodeInfo{$nodeId}->{PORT})
    {
        my $port        = $nodeInfo{$nodeId}->{PORT};
        my $interfaceId = "gigabitEthernet";
        if ($port =~ /^Gi/)
        {
            $port =~ s/Gi//;    # Strip off the Interface prefix
            $interfaceId = "gigabitEthernet";
        }
        $exp->expect(
            $timeout,
            [
                qr/$loginPrompt\(config\)#/ => sub {
                    my $self = shift;
                    $self->send("interface $interfaceId $port\n");
                  }
            ]
          )
          or logDie("Unable to enter interface mode ($interfaceId $port)\n");
        $exp->expect(
            $timeout,
            [
                qr/$loginPrompt\(config-if\)#/i => sub {
                    my $self = shift;
                    $self->send("switchport access vlan $vlanId\n");
                  }
            ]
          )
          or logDie("Unable to switch port $port to vlan $vlanId\n");
        $exp->expect(
            $timeout,
            [
                qr/$loginPrompt\(config-if\)#/i => sub {
                    my $self = shift;
                    $self->send("exit\n");
                  }
            ]
          )
          or logDie("Unable to add port $port to vlan\n");
    }
    else
    {
        logDie("No port information could be found for node $nodeId\n");
    }
}

# Close the connection
$exp->expect(
    $timeout,
    [
        qr/$loginPrompt\(config\)#/ => sub {
            my $self = shift;
            $self->send("exit\n");
            exp_continue;
          }
    ],
    [
        qr/$loginPrompt#/i => sub {
            my $self = shift;
            $self->send("exit\n");
          }
    ]
  )
  or logDie("Unable to close the connection\n");
$exp->soft_close();

exit 0;
1;

__END__

=head1 NAME

=head1 SYNOPSIS

=head1 DESCRIPTION

=head2 EXPORT

=head2 EXPORT_OK

=head1 SEE ALSO

Additional information can be found at http://www.adaptivecomputing.com/

=head1 AUTHOR

Adaptive Computing, Inc., E<lt>http://www.adaptivecomputing.com/E<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2005-2016 by Adaptive Computing Enterprises, Inc. All Rights Reserved.

=cut
