#! /usr/bin/perl -w
################################################################################
#
# Usage: vlan.delete.ciscoios.pl <vlan id>
#
# NOTE: Currently, router is set in the config.vlan.pl file
# and it is ignored from the nodes file
#
#   Copyright (C) 2005-2016 by Adaptive Computing Enterprises, Inc. All Rights Reserved.
#
################################################################################
use strict;
use Expect;
use FindBin;
use lib "/make_install/lib/perl5", "/make_install/etc";             # Installed
use lib "${FindBin::Bin}/../../lib/perl5", "${FindBin::Bin}/../../etc";   # Configured
use Getopt::Long;
use Moab::Tools;    # Required before including config to set $homeDir
BEGIN { require "config.vlan.pl"; }
our ($logLevel, $nodesFile, $router, $globalVlanId, $password, $secret,
    $loginPrompt, $timeout);

if ($logLevel >= 1)
{
    logPrint("$0 ", join(' ', @ARGV), "\n");
}

if ($logLevel <= 1)
{
    # Turn off expect output from spawned commands
    $Expect::Log_Stdout = 0;
}

# Parse Command Line Arguments
my ($help);
my $usage = "Usage: $0 <vlan id>\n";
GetOptions('help|?' => \$help) or die $usage;
die $usage if $help;
die $usage unless @ARGV == 1;
my ($vlanId) = @ARGV;

# Refuse to delete the global vlan
if (defined $globalVlanId && $vlanId eq $globalVlanId)
{
    logDie("Deleting the Global VLAN ($globalVlanId) is not allowed\n");
}

# Connect to the switch router
my $exp = Expect->spawn("telnet $router")
  or logDie("Failed connecting to switch router: $!\n");

# Log into the router
$exp->expect(
    $timeout,
    [
        qr/Password: /i => sub {
            my $self = shift;
            $self->send("$password\n");
          }
    ]
  )
  or logDie("Unable to log into switch router\n");

# Enter privileged mode
$exp->expect(
    $timeout,
    [
        qr/$loginPrompt>/i => sub {
            my $self = shift;
            $self->send("enable\n");
            exp_continue;
          }
    ],
    [
        qr/Password: /i => sub {
            my $self = shift;
            $self->send("$secret\n");
          }
    ]
  )
  or logDie("Unable to enter privileged mode\n");

# Add ports back to the global vlan if globalVlanId is defined
my @ports = ();
if ($globalVlanId ne "")
{
    # Determine all ports currently in vlan
    $exp->expect(
        $timeout,
        [
            qr/$loginPrompt#/ => sub {
                my $self = shift;
                $self->send("show vlan brief\n      ");    # Spaces for --More--
              }
        ]
      )
      or logWarn("Unable to query vlan\n");

    $exp->notransfer(1);
    $exp->expect($timeout, '-re', qr/$loginPrompt/)
      or logWarn("Unable to capture output\n");
    $exp->notransfer(0);

    foreach my $line (split /\n/, $exp->exp_before())
    {
        # FORMAT: VLAN Name                             Status    Ports
        my ($id, $name, $status, @portsWithCommas) = split /\s+/, $line;
        if (defined $id && $id eq $vlanId)
        {
            map {s/,//} (@ports = @portsWithCommas);
            last;
        }
    }
    logWarn("No ports defined for vlan $vlanId\n") unless @ports;
}

# Enter global configuration mode
$exp->expect(
    $timeout,
    [
        qr/$loginPrompt#/ => sub {
            my $self = shift;
            $self->send("config\n\n");
          }
    ]
  )
  or logDie("Unable to enter global configuration mode\n");

# Add ports back to the global vlan
foreach my $port (@ports)
{
    my $interfaceId = "gigabitEthernet";
    if ($port =~ /^Gi/)
    {
        $port =~ s/Gi//;    # Strip off the Interface prefix
        $interfaceId = "gigabitEthernet";
    }
    $exp->expect(
        $timeout,
        [
            qr/$loginPrompt\(config\)#/ => sub {
                my $self = shift;
                $self->send("interface $interfaceId $port\n");
              }
        ]
      )
      or logDie("Unable to enter interface mode ($interfaceId $port)\n");
    $exp->expect(
        $timeout,
        [
            qr/$loginPrompt\(config-if\)#/i => sub {
                my $self = shift;
                $self->send("switchport access vlan $globalVlanId\n");
              }
        ]
      )
      or logDie("Unable to switch port $port to global vlan $globalVlanId\n");
    $exp->expect(
        $timeout,
        [
            qr/$loginPrompt\(config-if\)#/i => sub {
                my $self = shift;
                $self->send("exit\n");
              }
        ]
      )
      or logDie("Unable to add port $port back to global vlan\n");
}

# Delete the old vlan
$exp->expect(
    $timeout,
    [
        qr/$loginPrompt\(config\)#/ => sub {
            my $self = shift;
            $self->send("no vlan $vlanId\n");
          }
    ]
  )
  or logDie("Unable to delete vlan $vlanId\n");

# Close the connection
$exp->expect(
    $timeout,
    [
        qr/$loginPrompt\(config\)#/ => sub {
            my $self = shift;
            $self->send("exit\n");
            exp_continue;
          }
    ],
    [
        qr/$loginPrompt#/i => sub {
            my $self = shift;
            $self->send("exit\n");
          }
    ]
  )
  or logDie("Unable to close the connection\n");
$exp->soft_close();

exit 0;
1;

__END__

=head1 NAME

=head1 SYNOPSIS

=head1 DESCRIPTION

=head2 EXPORT

=head2 EXPORT_OK

=head1 SEE ALSO

Additional information can be found at http://www.adaptivecomputing.com/

=head1 AUTHOR

Adaptive Computing, Inc., E<lt>http://www.adaptivecomputing.com/E<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2005-2016 by Adaptive Computing Enterprises, Inc. All Rights Reserved.

=cut
