#! /usr/bin/perl -w
################################################################################
#
# Usage: vlan.query.ciscoios.pl <vlan id>
#
# NOTE: Currently, router is set in the config.vlan.pl file
# and it is ignored from the nodes file
#
#   Copyright (C) 2005-2016 by Adaptive Computing Enterprises, Inc. All Rights Reserved.
#
################################################################################
use strict;
use Expect;
use FindBin;
use lib "/make_install/lib/perl5", "/make_install/etc";             # Installed
use lib "${FindBin::Bin}/../../lib/perl5", "${FindBin::Bin}/../../etc";   # Configured
use Getopt::Long;
use Moab::Tools;    # Required before including config to set $homeDir
BEGIN { require "config.vlan.pl"; }
our (
    $logLevel, $nodesFile,   $router, $globalVlanId,
    $password, $loginPrompt, $timeout
);

if ($logLevel >= 1)
{
    logPrint("$0 ", join(' ', @ARGV), "\n");
}

if ($logLevel <= 1)
{
    # Turn off expect output from spawned commands
    $Expect::Log_Stdout = 0;
}

# Parse Command Line Arguments
my ($help);
my $usage = "Usage: $0 <vlan id>\n";
GetOptions('help|?' => \$help) or die $usage;
die $usage if $help;
die $usage unless @ARGV == 1;
my ($vlanId) = @ARGV;

# Refuse to query the global vlan
if (defined $globalVlanId && $vlanId eq $globalVlanId)
{
    logDie("Querying the Global VLAN ($globalVlanId) is not allowed\n");
}

# Connect to the switch router
my $exp = Expect->spawn("telnet $router")
  or logDie("Failed connecting to switch router: $!\n");

# Log into the router
$exp->expect(
    $timeout,
    [
        qr/Password: /i => sub {
            my $self = shift;
            $self->send("$password\n");
          }
    ]
  )
  or logDie("Unable to log into switch router\n");

# Issue query
$exp->expect(
    $timeout,
    [
        qr/$loginPrompt>/ => sub {
            my $self = shift;
            $self->send("show vlan brief\n      ");    # Spaces for --More--
          }
    ]
  )
  or logDie("Unable to query vlan\n");

$exp->notransfer(1);
$exp->expect($timeout, '-re', qr/$loginPrompt/)
  or logDie("Unable to capture output\n");
$exp->notransfer(0);

my @ports = ();
foreach my $line (split /\n/, $exp->exp_before())
{
    # FORMAT: VLAN Name                             Status    Ports
    my ($id, $name, $status, @portsWithCommas) = split /\s+/, $line;
    if (defined $id && $id eq $vlanId)
    {
        map {s/,//} (@ports = @portsWithCommas);
        last;
    }
}
logDie("No ports defined for vlan $vlanId\n") unless @ports;

# Parse the nodes file and read in port information
my %nodeInfo = ();
if (open NODESFILE, "<$nodesFile")
{
    foreach my $line (<NODESFILE>)
    {
        $line =~ s/^\s+//;    # Remove leading whitespace
        $line =~ s/\#.*//;    # Strip out comments
        next unless $line;    # Ignore blank (or pure comment) lines
        my @words = split /\s+/, $line;
        my $nodeId = shift @words;
        foreach my $word (@words)
        {
            my ($key, $value) = split /=/, $word;
            $nodeInfo{$nodeId}->{$key} = $value;
        }
    }
    close NODESFILE;
}
else
{
    logPrint("Unable to open nodes file ($nodesFile) for reading: $!\n");
}

# Map ports to nodes
my @nodes = ();
foreach my $port (@ports)
{
    my $located = 0;
    foreach my $nodeId (keys %nodeInfo)
    {
        if ($nodeInfo{$nodeId}->{PORT} eq $port)
        {
            $located = 1;
            push @nodes, $nodeId;
        }
    }
    logDie("Unable to locate node id for port $port\n") unless $located;
}

# Close the connection
$exp->send("exit\n");
$exp->soft_close();

# Print the node list to STDOUT and exit
print join(',', @nodes), "\n";
exit 0;
1;

__END__

=head1 NAME

=head1 SYNOPSIS

=head1 DESCRIPTION

=head2 EXPORT

=head2 EXPORT_OK

=head1 SEE ALSO

Additional information can be found at http://www.adaptivecomputing.com/

=head1 AUTHOR

Adaptive Computing, Inc., E<lt>http://www.adaptivecomputing.com/E<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2005-2016 by Adaptive Computing Enterprises, Inc. All Rights Reserved.

=cut
