#! /usr/bin/perl -w
################################################################################
# TYPE A CONDOR SUBMISSION
#
# Usage: msub.wrapper.condor.pl <msub args>
#
# It is recommended to symlink or alias this file as msub_condor or something
# similar.
#
# This script does a type A Condor submission.  You submit a single file,
# just like using msub, including msub arguments.  This will be put into a 
# format that job.submit.condor.pl will understand to create an associated
# Condor submit file.
#
# This script will set the first line as "---ARGUMENTS <args>" and will
# also set filetype=<type>, where <type> is either the file extension or the 
# entire file name (excluding path), depending on configuration.
#
# Arguments:
# To put in arguments that will be placed in the Condor submit file, put 
# quotes around the file name and place any arguments inside the quotes, with
# quotes in the arguments exactly how you want them in the Condor submit file,
# but you will have to escape quotes so that the shell doesn't interpret them.
#     Example:
#     If you want the arguments 1 two "III", you would do it like this:
#         msub.wrapper.condor.pl "myprogram.py 1 two \"III\""
#
#   Copyright (C) 2010-2016 by Adaptive Computing Enterprises, Inc. All Rights Reserved.
#
################################################################################
use strict;
use FindBin;
use lib "/usr/local/lib/perl5",      "/opt/moab/etc";        # Installed
use lib "${FindBin::Bin}/../lib/perl5", "${FindBin::Bin}/../etc";   # Configured
use Getopt::Long;
use Moab::Tools;
use Sys::Hostname;
use Cwd;
BEGIN { require "config.condor.pl"; }
our ($logLevel, $obscureFilename, $writableDir, $useJobDirectory);
logPrint("$0 ", join(' ', @ARGV), "\n") if $logLevel;

# Parse Command Line Arguments
my $help;
my $usage = "Usage: $0 {msub args}\n";

#Configurable vars
#Configurable, must be globally writable
# Must end with the '/'
my $tmpDir = $writableDir;

my $args      = "";
my $flagUsed  = 0;
my $exec      = "";
my $execAdded = 0;
my $envUsed   = 0;
my %envs;

foreach my $argnum (0 .. $#ARGV)
{
    my $curArg = $ARGV[$argnum];
    $execAdded = 0;

    if (($flagUsed == 0) && (substr($curArg, 0, 1) ne "-"))
    {
        #Found the executable?
        my @words = split(/\s+/, $curArg);
        if ((-e $words[0]) && (-s $words[0]) && (-x $words[0]))
        {
            #It's a non-zero executable file -> assume it's the executable
            if ($exec eq "")
            {
                $exec      = $curArg;
                $execAdded = 1;
            }
        }
    }

    if ($execAdded == 0)
    {
        $args .= " $curArg";
    }

    #Set if the next arg is part of current flag or not
    #Skip flags that don't require an arg
    if (   (substr($curArg, 0, 1) eq "-")
        && (substr($curArg, 1, 1) ne "-")
        && (substr($curArg, 1, 1) ne "E")
        && (substr($curArg, 1, 1) ne "h")
        && (substr($curArg, 1, 1) ne "I")
        && (substr($curArg, 1, 1) ne "K")
        && (substr($curArg, 1, 1) ne "r")
        && (substr($curArg, 1, 1) ne "V")
        && (substr($curArg, 1, 1) ne "z"))
    {
        $flagUsed = 1;
    }
    else
    {
        $flagUsed = 0;
    }

    if ($envUsed == 1)
    {
        #User is requesting an env variable
        $envs{$curArg} = $ENV{$curArg};
    }

    if (   (substr($curArg, 0, 1) eq "-")
        && (substr($curArg, 1, 1) eq "v"))
    {
        #Need to set env vars again for the msub command
        $envUsed = 1;
    }
    else
    {
        $envUsed = 0;
    }
}

my $vars       = "";
my $actualExec = "";
my $newExec    = "";
#Use type to just pass the filetype through
my $type = "";
my $justFileName = "";

#Create temp file, pass in args
if ($exec ne "")
{
    #Find type of file
    my @eArgs      = split(/\s+/, $exec);
    my $actualExec = $eArgs[0];
    my $pos        = rindex($actualExec, ".");

    if (($obscureFilename) && ((!defined $useJobDirectory) || ($useJobDirectory == 0)))
    {
        if ($pos != -1)
        {
            $type = substr($actualExec, rindex($actualExec, "."));
            $vars .= "filetype=$type";
            $justFileName = substr($actualExec, rindex($actualExec, "/") + 1);
        }
    }
    else
    {
        #Just use the original filename (exclude path)
        if (rindex($actualExec, "/") != -1)
        {
            $type = substr($actualExec, rindex($actualExec, "/") + 1);
        }
        else
        {
            $type = $actualExec;
        }

        $justFileName = substr($actualExec, rindex($actualExec, "/") + 1);
        $vars .= "filetype=$type";
    }

    $exec =~ s/$actualExec\w*//;    #Get just the args

    #Create a new submit file.
    # This file is the same as the original, but includes a #ARGUMENTS line
    #  to hold the arguments.  These will be parsed out at submission
    #  and put into the Condor submission file.
    $newExec = $tmpDir . "tmp$justFileName";

    open(REALEXEC, ">$newExec") or die "Unable to open $newExec for writing: $!\n";
    print REALEXEC "---ARGUMENTS $exec\n";

    #Copy original file
    open(ORIGEXEC, "<$actualExec");
    my @origLines = <ORIGEXEC>;
    foreach my $line (@origLines)
    {
        #chomp($line);
        print REALEXEC $line;
    }

    close(ORIGEXEC);
    close(REALEXEC);
}
else
{
    logPrint("Could not find the executable!\n");
}

#Must put a space on the end
my $subEnvs = "";
foreach my $key (keys %envs)
{
    $subEnvs .= "$key=$envs{$key} ";
}

my $hostname = hostname();
my $cwd = getcwd();

my $setVars = "-W x='VAR=submitmachine=$hostname' ";
$setVars .= "-W x='VAR=cwd=$cwd' ";

#Submit to Moab
my $cmd = "$subEnvs" . "msub $args -W x='VAR=$vars' $setVars $newExec";

#print "\n$cmd\n\n";
logPrint("Invoking command: $cmd\n") if $logLevel >= 2;
my $output = `$cmd 2>&1` || `sh -c "$cmd 2>&1"`;
my $rc = $?;
if ($rc)
{
    unlink(($newExec));
    logDie("Job submission failed ($cmd) [rc=$rc]: $output\n");
}
elsif ($logLevel >= 2 && $output)
{
    logPrint($output);
}

print $output;

unlink(($newExec));
exit 0;
1;
__END__

=head1 NAME

=head1 SYNOPSIS

=head1 DESCRIPTION

=head2 EXPORT

=head2 EXPORT_OK

=head1 SEE ALSO

Additional information can be found at http://www.adaptivecomputing.com/

=head1 AUTHOR

Adaptive Computing, Inc., E<lt>http://www.adaptivecomputing.com/E<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2010-2016 by Adaptive Computing Enterprises, Inc. All Rights Reserved.

=cut
