#! /usr/bin/perl -w
################################################################################
# TYPE B CONDOR SUBMISSION
#
# Usage: msub.wrapper.condorB.pl <condorSubmitFile>
#
# NOTE: Requires the "shar" utility
# NOTE: Must be called from the same directory as the submit file
# NOTE: Because this creates a shar of the entire directory from the directory
#         below, the user must have write priveleges for the current directory
#         and '..'
#
# It is recommended to symlink or alias this file as condor_msub or something
# similar.
#
# This script creates a self extracting file of the current directory and scp's
# it to the configured destination Moab machine (if the current machine is the
# destination machine, then it is merely copied to the destination directory
# ($writableDir).
#
# The purpose of creating a self extracting file of the directory is that
# Condor jobs can sometimes be defined in relation to the current directory
# structure, so the only way to ensure that the job will run is to transport
# the entire directory tree.  This is made into a self extracting file 
# instead of just a tarball because Moab adds #PBS directives into the 
# submitted file, which throws of untarring, even if these directives are
# removed.  With a self exctracting file, we can remove the directives and
# still access the original content.
#
# Variables set:
#     filetype -> the name of the created shar file
#     submitfile -> the name of the original Condor submit file
#     chdir -> the name of the current directory - helps when extracting
#     cwd -> the full path of the original directory (moving output)
#     submitmachine -> the hostname of the submit machine
#
#   Copyright (C) 2010-2016 by Adaptive Computing Enterprises, Inc. All Rights Reserved.
#
################################################################################
use strict;
use FindBin;
use lib "/usr/local/lib/perl5",      "/opt/moab/etc";        # Installed
use lib "${FindBin::Bin}/../lib/perl5", "${FindBin::Bin}/../etc";   # Configured
use Getopt::Long;
use Moab::Tools;
use Cwd;
use File::Copy;
use Sys::Hostname;
use IO::Socket;
BEGIN { require "config.condor.pl"; }
our ($logLevel, $obscureFilename, $writableDir, $useJobDirectory, $dstMoab, $copyCmd);
logPrint("$0 ", join(' ', @ARGV), "\n") if $logLevel;

# Parse Command Line Arguments
my $help;
my $usage = "Usage: $0 {msub args}\n";

#Configurable vars
#Configurable, must be globally writable
# Must end with the '/'
my $tmpDir = $writableDir;
my $currentDir = getcwd();
my $currentDirName = substr($currentDir,rindex($currentDir,"/") + 1);
$currentDir .= "/";
my $submitFile = "";

if ($ARGV[0])
{
    $submitFile = $ARGV[0];
    if (index($submitFile, "/") >= 0)
    {
        #We are not in the same directory as the submit file
        print "Must submit from the same directory as the Condor submit file\n";
        logDie ("Must submit from the same directory as the Condor submit file\n");
    }
}
else
{
    logDie ("No submit file was specified");
}

#Set up the tar file stuff
my $newExec = &genSharFileName();
chdir("..");

#Can't do the 2>&1 stuff here because it messes up the shar output
# Instead, test if the reslting file exists or not for verification
my $cmd = "shar -zQq $currentDirName > $writableDir$newExec";
my $output = `$cmd` || `sh -c "$cmd"`;

if (! (-e "$writableDir$newExec"))
{
    print "ERROR:  could not create share of directory - $output\n";
    logDie("ERROR:  could not create share of directory - $output\n");
}

$cmd = "mv $writableDir$newExec $currentDirName";
$output = `$cmd 2>&1` || `sh -c "$cmd 2>&1"`;
chdir($currentDirName);

#
#  SCP or copy the file
#

my $hostname = hostname();
my $ip = inet_ntoa((gethostbyname($hostname))[4]);

if (($dstMoab eq $hostname) || ($dstMoab eq $ip))
{
    #Same machine, just copy the file
    copy($newExec, $writableDir);    
}
else
{
    $cmd = "$copyCmd $newExec $dstMoab:$writableDir";
    $output = `$cmd 2>&1` || `sh -c "$cmd 2>&1"`;
    my $scprc = $?;
    if ($scprc)
    {
        logDie("Could not scp in job submission - $output\n");
    }
}


#VARIABLES:
#  submitfile=$submitFile -> Name of the original submission file
#  filetype=$newExec -> Name of the created shar file
#  chdir=$currentDirName -> Where to chdir to (since this directory will be tarred
#  cwd=getcwd -> Current directory for returning the resulitng tar file
#  submitmachine=$hostname -> Machine to return tar file to

#Submit to Moab
my $vars = "-W x='VAR=filetype=$newExec' ";
$vars .= "-W x='VAR=chdir=$currentDirName' ";
$vars .= "-W x='VAR=submitfile=$submitFile' ";
$vars .= "-W x='VAR=cwd=" . getcwd() . "' ";
$vars .= "-W x='VAR=submitmachine=$hostname' ";

#Don't actually submit the file so that it isn't altered, run, etc.
$cmd = "echo sleep 5 | msub $vars $submitFile";

logPrint("Invoking command: $cmd\n") if $logLevel >= 2;
$output = `$cmd 2>&1` || `sh -c "$cmd 2>&1"`;
my $rc = $?;
if ($rc)
{
    unlink(($newExec));
    logDie("Job submission failed ($cmd) [rc=$rc]: $output\n");
}
elsif ($logLevel >= 2 && $output)
{
    logPrint($output);
}

print $output;

unlink(($newExec));
exit 0;


###################################################################
###################################################################

sub genSharFileName
{
    my $length = 8;
    my @chars = ('a'..'z','A'..'Z','0'..'9');
    my $randomString;

    foreach (1..$length)
    {
        $randomString.=$chars[rand @chars];
    }

    # .sh because it will be a shell archive
    return $randomString . ".sh";
}
1;
__END__

=head1 NAME

=head1 SYNOPSIS

=head1 DESCRIPTION

=head2 EXPORT

=head2 EXPORT_OK

=head1 SEE ALSO

Additional information can be found at http://www.adaptivecomputing.com/

=head1 AUTHOR

Adaptive Computing, Inc., E<lt>http://www.adaptivecomputing.com/E<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2010-2016 by Adaptive Computing Enterprises, Inc. All Rights Reserved.

=cut
