#!/usr/bin/env python

import argparse
from argparse import RawTextHelpFormatter
import os.path
from datetime import datetime, timedelta
from psutil import virtual_memory
import cPickle as pickle
import re
import portalocker
import time
import random
import socket
import fcntl
import struct
import subprocess
import logging
import math
import traceback
import glob
import fileinput
import shutil
from lib.mongo import MongoConnection, get_job_from_mongo, add_mongo_parser_arguments

if os.geteuid() != 0:
    exit("You need to have root privileges to run this script.")

dir_path = os.path.dirname(os.path.realpath(__file__))
logger = logging.getLogger('elastic-trigger')
logger.setLevel(logging.DEBUG)
multi_mom_config_dir = '/var/spool/elastic-torque/'

#Changes to the elastic_node_prefix will affect the clean.sh script
elastic_node_prefix = "elastic_node"
elastic_ids_file = "%s/elastic_node_ids" % dir_path
log_file = "%s/elastic-trigger.log" % dir_path
fh = logging.FileHandler(log_file)

#Debug level
fh.setLevel(logging.DEBUG)

'''
Python logger format options:

asctime %(asctime)s
created %(created)f
filename    %(filename)s
funcName    %(funcName)s
levelname   %(levelname)s
levelno %(levelno)s
lineno  %(lineno)d
module  %(module)s
msecs   %(msecs)d
message %(message)s
name    %(name)s
pathname    %(pathname)s
process %(process)d
processName %(processName)s
relativeCreated %(relativeCreated)d
thread  %(thread)d
threadName  %(threadName)s
'''
formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
fh.setFormatter(formatter)
logger.addHandler(fh)


def main(**args):
    elastic_node_ids = get_elastic_node_ids()
    hostname = socket.gethostname()
    ip_address = socket.gethostbyname(hostname)
    args['hostname'] = hostname
    args['ip_address'] = ip_address
    if args['action'] == "add":
        if len(args['requirements']) > 0:
            for req in args['requirements']:
                node_count = req['node_count']
                proc_count = req['proc_count']
                if node_count > 1:
                    logger.debug("Creating %d nodes with %d procs each because taskCount=%d, tasksPerNode=%d, dedicatedProcessors=%d." % (node_count, proc_count, req['task_count'], req['tasks_per_node'], req['dedicated_procs'] ))
                ensure_memory_available(node_count)
                node_names = []
                for x in range(0, node_count):
                    node_name = add_elastic_node(elastic_node_ids)
                    node_names.append(node_name)
                add_etc_hosts_lines(node_names, hostname, ip_address)
                if node_names > 0:
                    request_id = node_names[0]
                    req['torque_bin'] = args['torque_bin']
                    req['torque_sbin'] = args['torque_sbin']
                    req['hostname'] = hostname
                    req['ip_address'] = ip_address
                    req['ttl'] = args['ttl']
                    req['exclude_ttl'] = args['exclude_ttl']
                    for node in node_names:
                        addNode(node, proc_count, req, elastic_node_ids, request_id)
                        persist_ids(elastic_node_ids)
        else:
            node_count = int(math.ceil(float(args['proc_count']) / float(args['procs_per_node'])))
            proc_count = int(math.ceil(float(args['proc_count']) / float(node_count)))
            if node_count > 1:
                logger.debug("Creating %d nodes with %d procs each because procs_per_node parameter is %d." % (node_count, proc_count, args['procs_per_node']))
            ensure_memory_available(node_count)
            node_names = []
            for x in range(0, node_count):
                node_name = add_elastic_node(elastic_node_ids)
                node_names.append(node_name)
            add_etc_hosts_lines(node_names, hostname, ip_address)
            if node_names > 0:
                request_id = node_names[0]
                for node in node_names:
                    addNode(node, proc_count, args, elastic_node_ids, request_id)
                persist_ids(elastic_node_ids)
    elif args['action'] == "remove":
        nodeid = remove_elastic_node(args['node_name'], elastic_node_ids)
        manager_port=40000 + int(nodeid)
        node_name = "%s%s" % (elastic_node_prefix, nodeid)
        out, err = call("%s/momctl -s -p %d" % (args['torque_sbin'], manager_port))
        if err and 'cannot connect' not in err:
            out, err = call("lsof -P +c 8 -i :%d" % manager_port)
            if out:
                try:
                    pid = int(re.match('.*?pbs_mom (\d+).*', out.replace('\n', '')).group(1))
                    message = "Force kill pbs_mom on manager port %d: kill -9 %d" % (manager_port, pid)
                    print message
                    logger.error(message)
                    call("kill -9 %d" % pid)
                except:
                    pass
        call("%s/qmgr -c \"delete node %s\"" % (args['torque_bin'], node_name))
        remove_etc_hosts_line("\n%s %s %s" % (ip_address, hostname, node_name))
        remove_mom_dir(node_name)
        persist_ids(elastic_node_ids)

def addNode(node, proc_count, args, elastic_node_ids, request_id):
    print "Add %s" % node
    nodeid = get_node_id_from_name(node)
    service_port=30000 + int(nodeid)
    manager_port=40000 + int(nodeid)
    command =  "%s/qmgr -c \"create node %s "  % (args['torque_bin'], node)
    if 'exclude_ttl' in args and not args['exclude_ttl']:
        command += "ttl=%s,requestid=%s," % (args['ttl'], request_id)
    if 'features' in args and args['features'] and len(args['features']) > 0:
        command += "properties=%s," % ":".join(args['features'])
    command += "np=%d,mom_service_port=%s,mom_manager_port=%s\"" % (proc_count, service_port, manager_port)
    out, err = call(command)
    if 'name already exists' in err:
        remove_etc_hosts_line("\n%s %s %s" % (args['ip_address'], args['hostname'], node))
        ensure_memory_available(1)
        node_name = add_elastic_node(elastic_node_ids)
        add_etc_hosts_lines([node_name], args['hostname'], args['ip_address'])
        addNode(node_name, proc_count, args, elastic_node_ids, request_id)
        return
    mom_dir = make_mom_dir(node, args)
    out, err = call("%s/pbs_mom -d %s -m -M %d -R %d -A %s" % (args['torque_sbin'], mom_dir, service_port, manager_port, node))
    if ('another mom running' in err) or ('already in use' in err):
        call("%s/qmgr -c \"delete node %s\"" % (args['torque_bin'], node))
        remove_mom_dir(node)
        remove_etc_hosts_line("\n%s %s %s" % (args['ip_address'], args['hostname'], node))
        ensure_memory_available(1)
        node_name = add_elastic_node(elastic_node_ids)
        add_etc_hosts_lines([node_name], args['hostname'], args['ip_address'])
        addNode(node_name, proc_count, args, elastic_node_ids, request_id)
        return

def remove_mom_dir(node_name):
    config_dir = os.path.join(multi_mom_config_dir, node_name)
    if os.path.exists(config_dir):
        shutil.rmtree(config_dir)

def make_mom_dir(node_name, args):
    config_dir = os.path.join(multi_mom_config_dir, node_name)
    spool_dir = os.path.join(config_dir, 'spool')
    aux_dir = os.path.join(config_dir, 'aux')
    mom_logs_dir = os.path.join(config_dir, 'mom_logs')
    mom_priv_dir = os.path.join(config_dir, 'mom_priv')
    jobs_dir = os.path.join(mom_priv_dir, 'jobs')
    undelivered_dir = os.path.join(config_dir, 'undelivered')
    checkpoint_dir = os.path.join(config_dir, 'checkpoint')
    config_lines = []
    config_lines.append('$pbsserver %s' % args['hostname'])
    config_lines.append('$exec_with_exec true')
    if 'greses' in args and args['greses'] and len(args['greses'].items()) > 0:
        for key, value in args['greses'].items():
            config_lines.append("%s %s" % (key, value))
    if 'arch' in args and args['arch']:
        config_lines.append("arch %s" % args['arch'])
    if 'os' in args and args['os']:
        config_lines.append("opsys %s" % args['os'])
    if not os.path.exists(spool_dir):
        os.makedirs(spool_dir)
    if not os.path.exists(aux_dir):
        os.makedirs(aux_dir)
    if not os.path.exists(mom_logs_dir):
        os.makedirs(mom_logs_dir)
    if not os.path.exists(mom_priv_dir):
        os.makedirs(mom_priv_dir)
    if not os.path.exists(jobs_dir):
        os.makedirs(jobs_dir)
    if not os.path.exists(undelivered_dir):
        os.makedirs(undelivered_dir)
    if not os.path.exists(checkpoint_dir):
        os.makedirs(checkpoint_dir)
    os.chmod(spool_dir, 07777)
    os.chmod(undelivered_dir, 07777)
    os.chmod(checkpoint_dir, 07777)
    with open(os.path.join(mom_priv_dir, "config"), 'w') as f:
        f.writelines('\n'.join(config_lines)+'\n')
    with open(os.path.join(config_dir, "pbs_environment"), 'w') as f:
        f.write("PATH=/bin:/usr/bin\nLANG=C")
    return config_dir

def call(cmd):
    logger.debug("Running command %s" % cmd)
    p = subprocess.Popen(cmd, stdin=subprocess.PIPE,
                         stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    out, err = p.communicate()
    if out:
        message = "cmd='%s';stdout='%s'" % (cmd,out)
        logger.debug(message)
    if err:
        message = "cmd='%s';stderr='%s'" % (cmd,err)
        print message
        logger.error(message)
    return (out, err)

def add_etc_hosts_lines(node_names, hostname, ip_address):
    lines = ""
    for name in node_names:
        lines += "\n%s %s %s" % (ip_address, hostname, name)
    with open('/etc/hosts', 'a') as f:
        f.writelines(lines)

def remove_etc_hosts_line(line):
    f = open('/etc/hosts','r')
    filedata = f.read()
    f.close()
    newdata = filedata.replace(line,"")
    f = open('/etc/hosts','w')
    f.write(newdata)
    f.close()

def get_elastic_node_ids():
    elastic_node_ids = [0]
    if os.path.isfile(elastic_ids_file):
        try:
            elastic_node_ids = pickle.load( open( elastic_ids_file, "rb" ))
        except:
            die("The elastic_node_ids file is corrupted.  Please remove file and to restart noded ids.")
    if 0 not in elastic_node_ids:
        die("The elastic_node_ids file is corrupted.  Please remove file and to restart noded ids.")
    return elastic_node_ids

def remove_elastic_node(node_name, elastic_node_ids):
    nodeid = get_node_id_from_name(node_name)
    try:
        elastic_node_ids.remove(int(nodeid))
    except:
        pass
    return nodeid

def get_node_id_from_name(node_name):
    try:
        nodeid = int(re.match('.*?([0-9]+)$', node_name).group(1))
        assert nodeid != 0
    except:
        die("Node name must end with 1 or more digits and cannot be 0.")
    return "%02d" % nodeid

def persist_ids(elastic_node_ids):
    pickle.dump( elastic_node_ids, open( elastic_ids_file, "wb" ) )

def add_elastic_node(elastic_node_ids):
    missing_node_ids = missing_elements(elastic_node_ids)
    if missing_node_ids:
        lowest_missing_id = min(missing_node_ids)
    else:
        lowest_missing_id = max(elastic_node_ids)+1
    elastic_node_ids.append(lowest_missing_id)
    return "%s%02d" % (elastic_node_prefix, lowest_missing_id)

def missing_elements(L):
    L = sorted(L)
    start, end = L[0], L[-1]
    return sorted(set(range(start, end + 1)).difference(L))

def ensure_memory_available(node_count):
    bytes_in_a_mb = 1048576
    free_mb = virtual_memory().free / bytes_in_a_mb
    node_mb = 50
    if (node_count * node_mb) > free_mb:
        die("Insufficient free memory(%dMB) to start %d MOMs. Each MOM takes about %dMB." % (free_mb, node_count, node_mb))

def parse_job_requirements(job):
    parsed_requirements = []
    try:
        for requirements in job['requirements']:
            requirement = {}
            del requirements['resourcesPerTask']['swap']
            del requirements['resourcesPerTask']['disk']
            requirement['dedicated_procs'] = requirements['resourcesPerTask']['processors']['dedicated']
            requirement['task_count'] = requirements['taskCount']
            requirement['tasks_per_node'] = requirements['tasksPerNode']
            if int(requirement['tasks_per_node']) == 0:
                requirement['tasks_per_node'] = 1
            requirement['node_count'] = int(math.ceil(float(requirements['taskCount']) / float(requirement['tasks_per_node'])))
            #Modify resourcesPerTask array to account for node count
            for k,v in requirements['resourcesPerTask'].items():
                total_resources_needed = int(requirements['resourcesPerTask'][k]['dedicated']) * int(requirements['taskCount'])
                requirements['resourcesPerTask'][k] = int(math.ceil(float(total_resources_needed) / float(requirement['node_count'])))
            requirement['proc_count'] = requirements['resourcesPerTask'].pop('processors')
            requirement['memory'] = requirements['resourcesPerTask'].pop('memory')
            requirement['greses'] = requirements['resourcesPerTask']
            requirement['features'] = requirements['features']
            requirement['arch'] = requirements['architecture']
            requirement['os'] = requirements['image']
            parsed_requirements.append(requirement)
    except Exception as e:
        logger.error("Parse error for job: '%s':'%s'" % (job, e.args))
    return parsed_requirements

def parse_request_geometry(request_geometry):
    try:
        d = request_geometry.split('@')
        assert len(d) == 2
        node_count = int(d[0])
        assert node_count > 0
        request_moab_time = d[1]
    except:
        die("Request geometry '%s' is invalid. Please start with a positive number of nodes followed by the @ symbol followed by the time to live." % request_geometry)

    try:
        d = map(int, request_moab_time.split(':'))
        for i in d:
            assert i >= 0
        assert len(d) <= 4
    except:
        die("Request geometry '%s' is invalid. The time to live should be up to 4 non-negative integers separated by colons." % request_geometry)
    while len(d) < 4:
        d.insert(0,0)

    today = datetime.utcnow()
    #Add 1 minute of pading to the TTL
    moab_time = timedelta(days=d[0], hours=d[1], minutes=d[2] + 1, seconds=d[3])
    expected_retire_date = today + moab_time
    return (node_count, expected_retire_date.strftime("%Y-%m-%dT%T+00"))

def die(message):
    logger.error(message)
    exit(message)

def positive_integer(value):
    ivalue = int(value)
    if ivalue <= 0:
         raise argparse.ArgumentTypeError("%s is not a positive integer value" % value)
    return ivalue

description = """
This script is used to experiment with Moab's elastic computing feature without
the need of a real cloud provider.  The script uses torque's multi-mom feature to simulate
the creation and removal of nodes from the cluster.  To use the script with Moab add
the following lines to your moab.cfg and restart Moab.

NODEAVAILABILITYPOLICY	DEDICATED
USERCFG[DEFAULT]        QLIST=elastic
SCHEDCFG[Moab]			FLAGS=enabledynamicnodes
QOSCFG[DEFAULT]			ENABLEPROFILING=TRUE
NODECFG[DEFAULT]		ENABLEPROFILING=TRUE
QOSCFG[elastic]			REQUESTGEOMETRY=PRIORITYJOBSIZE
QOSCFG[elastic]			TRIGGER=EType=threshold,AType=exec,TType=elastic,Action="%s --jobid $JOBID --action add --request-geometry $REQUESTGEOMETRY",Threshold=BACKLOGCOMPLETIONTIME>500,RearmTime=01:00
NODEIDLEPURGETIME		120
NODECFG[DEFAULT]		TRIGGER=EType=end,AType=exec,TType=elastic,Action="%s --action remove --node-name $OID"

Example usage:

./%s  #add 1 multi-mom node with the default request geometry
./%s --action add --request-geometry 16@4:00:00:00 #add 1 multi-mom node with 16 procs for 4 days
./%s --action add --request-geometry 6@10  #add 1 multi-mom node with 6 procs for 10 second.
./%s --action remove --node-name elastic_node01  #remove the node named elastic_node01

Example job submission that uses the qos elastic so Moab will automatically trigger this script:

    echo sleep 600 | msub -l procs=16,qos=elastic,walltime=600

Warning: Make sure you have configured Moab with an rlm server.  If this is a problem you should see the error in the moab logs "node removed because no licenses available".  To check the status of your moab licenses run 'mdiag -l'.
""" % (os.path.abspath(__file__), os.path.abspath(__file__), os.path.basename(__file__), os.path.basename(__file__), os.path.basename(__file__), os.path.basename(__file__))

if __name__ == '__main__':
    config_file_name = '/opt/insight/etc/config.groovy'
    config = ""
    if os.access(config_file_name, os.R_OK):
        config += " "
        with open(config_file_name, 'r') as f:
            config += f.read().replace('\n', ' ')
    else:
        print "Could not open /opt/insight/etc/config.groovy"

    for config_file_name in glob.glob("/opt/insight/etc/config.d/*"):
        if os.access(config_file_name, os.R_OK):
            config += " "
            with open(config_file_name, 'r') as f:
                config += f.read().replace('\n', ' ')
        else:
            print "Could not open %s" % config_file_name

    parser = argparse.ArgumentParser(description=description, formatter_class=RawTextHelpFormatter)
    parser.add_argument('--action',
                        choices=["add", "remove"],
                        default="add",
                        required=True,
                        help="Whether to add or delete a torque elastic multi-mom node. Defaults to add."
                        )
    parser.add_argument('--request-geometry',
                        help="This takes the form <number of cpus>@<time to live for the node>.  Default is 16 cpus for 1 day.",
                        default="16@01:00:00:00"
                        )
    parser.add_argument('--node-name',
                        help="The node to delete.",
                       )
    parser.add_argument('--no-ttl',
                        help="This is to create a node that looks like a permanent node to Moab.",
                        dest="exclude_ttl",
                        action='store_true'
                       )
    parser.add_argument('--jobid',
                        help="The job id of the job that triggered the elastic bursting request.",
                       )
    parser.add_argument('--os',
                        help="Operating System of the node to create.",
                       )
    parser.add_argument('--arch',
                        help="Archecture of the node to create.",
                       )
    parser.add_argument('--greses',
                        help="Comma separated list of generic resources to create on the node. Eg: matlab:1,network:2",
                        type=lambda kv: dict((key, value) for (key, value) in [j.split(":") for j in kv.split(",")])
                       )
    parser.add_argument('--features',
                        help="Comma separated list of features to create on the node.",
                        type=lambda list: list.split(",")
                       )
    parser.add_argument('--torque-sbin',
                        help="The path that contains the torque pbs_mom and momctl binaries.",
                        default="/usr/local/sbin",
                       )
    parser.add_argument('--torque-bin',
                        help="The path that contains the torque qmgr binary file.",
                        default="/usr/local/bin",
                       )
    parser.add_argument('--procs-per-node',
                        help="This is used if you want to create more than 1 node from the request geometry.  If request geometry calls for 16 procs for example and you set procs-per-node to 8 then you should expect this script to create 2 nodes with 8 procs each.",
                        default=10000000000,
                        type=positive_integer
                       )
    add_mongo_parser_arguments(config, parser)
    args = parser.parse_args()
    if args.action == "add":
        logger.debug("Moab has requested a node with a request geometry of '%s'" % args.request_geometry)
        args.proc_count, args.ttl = parse_request_geometry(args.request_geometry)
        args.requirements = []
        if args.jobid:
            logger.debug("The job id of the job that triggered the elastic burst was '%s'" % args.jobid)
            mongo = MongoConnection(args, logger)
            if mongo.connected:
                job = mongo.withConnection(get_job_from_mongo, args.jobid, logger)
                if job:
                    args.requirements = parse_job_requirements(job)
                else:
                    logger.error("There was no job '%s' in the moab mongo database so skipping requirements." % args.jobid)
            else:
                logger.error("Ignoring job id '%s' because not connected to mongo" % args.jobid)
    else:
        if not args.node_name:
            parser.print_help()
            die("\nPlease include the node name that you want to remove.")
        logger.debug("Moab has requested to remove node '%s'" % args.node_name)
    tries = 600
    while(True):
        try:
            with portalocker.Lock('/tmp/.elastic_node_ids.lock') as fh:
                main(**vars(args))
                break
        except Exception as ex:
            if tries == 0:
                if args.action == "add":
                    moab_message = "Moab's requested for a node with a request geometry of '%s' failed." % args.request_geometry
                else:
                    moab_message = "Moab's request to remove node '%s' failed." % args.node_name
                template = "An exception of type {0} occured. Arguments:\n{1!r}"
                message = template.format(type(ex).__name__, ex.args)
                traceback.print_exc()
                die("%s\n%s" % (moab_message, message))
            time.sleep((random.randint(0, 1000) / 1000))
            tries -= 1
