#!/usr/bin/perl

use strict;
use warnings;

package EventParser;

use DBI;
use Pod::Usage;
use Carp qw(carp cluck croak confess);
use Data::Dumper;


#-------------------------------------------------------------------------
# my %event = parseEvent($lineFromEventFile)
#
# Parameters:
#   $lineFromEventFile A line from a Moab event file generated when the
#      moab.cfg WIKIEVENTS parameter is set to true
#
# Parses the event file line and returns a hash of event attribute to value
# pairs.  For example a hash like the following might be returned for a VM
# migrate event associated with vm245
#   {
#     'objectName' => 'vm245',
#     'eventType' => 'VMMIGRATE',
#     'eventID' => '18625413',
#     'params' => {
#       'SOVEREIGN' => 'FALSE',
#       'MSG' => 'migrating VM wcdre37549 to node r06idpx31'
#      },
#      'objectType' => 'vm',
#      'epochTime' => '1304503202'
#    };
#
#-------------------------------------------------------------------------
sub parseEvent($) {
	$_ = shift;
	if (!$_) {
		warn "parseEvent passed undef event.  Returning empty hash";
		return ();
	}
	chomp;
	my @tokens = split( /\s+/, $_, 6 );
	if (scalar @tokens < 6) {
		warn "parseEvent passed event string '$_' which cannot be parsed to a valid event.  Returning empty hash";
		return ();
	}
	return unless scalar @tokens == 6;
	my ( $eventTime, $epochTimeEventID, $objectType, $objectName, $eventType,
		$desc )
	  = @tokens;
	my ( $epochTime, $eventID ) = split( /:/, $epochTimeEventID );
	my @descWords;
	my %params;
	my @descTokens;
	my $quoted;
	my $quoteUsed;
	my $str;

	#force delimeter at the end for cleaner parsing algorithm
	$desc .= ' ';

	#    for my $char ( split( //, $testDesc ) ) {
	for my $char ( split( //, $desc ) ) {

		if ($quoted) {

			#we're in a quote so append verbatim
			$str .= $char;

			#unset $quoted once we find a matching $quoteUsed (could be ' or ")
			$quoted = $quoteUsed ne $char;
		}
		else {
			if ($str) {

				#we're building a string
				if ( $char =~ /\s/ ) {

					#found custom delimiter (whitespace not within quotes)
					#push the new token onto the stack
					push @descTokens, $str;

					#reset the $str
					$str = '';
				}
				else {

					#append $char, check for quotes, then keep going
					$str .= $char;
					$quoted = $quoteUsed = $char if ( $char =~ /['|"]/ );
				}
			}
			elsif ( $char =~ /\S/ ) {

				#we're starting a new string
				if ( $char =~ /['|"]/ ) {

					#we're starting a quoted key
					$quoted = $quoteUsed = $char if ( $char =~ /['|"]/ );
				}
				else {

					#we're starting param or a word
					$str = $char;
				}
			}
		}
	}

	for my $token (@descTokens) {
		if ( $token =~ /=/ ) {

			#parse wiki tokens (v6.0)
			#try quoted regex
			if ( $token =~ /([^\s=]+)=['|"](.*)['|"]/ ) {

				$params{$1} = $2;
			}

			#use regular regex
                                        # /([^\s=]+)=(.*)/
			elsif ( $token =~ /([^\s=]+)=(.*)/ ) {

				$params{$1} = $2;
			}
		}
		else {

      # These are tokens that are not =-separated key/value pairs

			#TODO parse non-wiki formats (v5.0)
			# For now, just lumping them in separate array
			push @descWords, $token;
		}

	}

	my %event = (
		eventTime    => $eventTime,
		epochTime    => $epochTime,
		eventID      => $eventID,
		objectType   => $objectType,
		objectName   => $objectName,
		eventType    => $eventType,
		desc         => $desc,
		descWords    => \@descWords,
		params       => \%params,
	);
	return %event;
}




#---------------------------------------------------------------------
# %hash = getNameToIDMappings($databaseHandle,$tableName);
#
# Parameters
#   databaseHandle : The database handle
#   tableName : The database table name.  This should be a table with
#     an ID and a Name column
#
# Queries a database table with a Name and ID column and returns the ID
# to Name mappings.  For example, if passed 'EventType' then
# %eventTypeHash should contain mappings like( 'GEVENT' => 1,
# 'JOBCANCEL' => 2, 'JOBEND' => 3)
#
#---------------------------------------------------------------------
sub getNameToIDMappings($$) {
	my ($dbh, $tableName) = @_;
	my %retVal;
	my $sth = $dbh->prepare( "select * from " . $tableName);
	$sth->execute();
	while ( my $row = $sth->fetchrow_hashref() ) {
		$retVal{ lc $row->{Name} } = $row->{ID};
	}
	$sth->finish();
	return %retVal;
}


#----------------------------------------------------------------------
# my $numOfLines = countLinesInFile("/tmp/myfile.txt")
#
# Parameters:
#   $filePath the path to a text file
#
# Counts the number of lines in a text file
#----------------------------------------------------------------------
sub countLinesInFile($)
{
  my ($filePath) = @_;
  if (!$filePath)
  {
    cluck '$filePath is undefined, cannot parse an undefined file';
    return 0;
  }
  if (!(-r $filePath))
  {
    cluck "Can't read from event file $filePath";
    return 0;
  }
  if (!(-T $filePath))
  {
    cluck "Event file $filePath is not a text file" ;
    return 0;
  }
  my $fileHandle;
  open($fileHandle,$filePath);
  my @lines=<$fileHandle>;
  return scalar(@lines);
}


1;

__END__

=head1 NAME

EventParser.pm - Contains method used to Moab event files and inserts the
events into a database

=head1 SYNOPSIS

use EventParser;

my %event = EventParser::parseEvent($lineFromEventFile);

my %eventTypes = EventParser::getNameToIDMappings($databaseHandle,"EventType");

my $numOfLines = countLinesInFile("/tmp/myfile.txt")

=head1 DESCRIPTION

This module contains method used to parse Moab event files and insert the
events into a database

=over 4

=item B<my %event = parseEvent($lineFromEventFile);>

Parses the event file line and returns a hash of event attribute to value
pairs.

=item B<my %eventTypes = getNameToIDMappings($databaseHandle,"EventType");>

Queries a database table with a Name and ID column and returns the ID
to Name mappings.  For example, if passed 'EventType' then
%eventTypeHash should contain mappings like( 'GEVENT' => 1,
'JOBCANCEL' => 2, 'JOBEND' => 3)

=item B<my $numOfLines = countLinesInFile("/tmp/myfile.txt");>

Counts the number of lines in a file. If the file is not readable,
0 is returned.

=back

=head1 REQUIRES

DBI
Pod::Usage
Carp
Data::Dumper

=head1 REPORTING BUGS

Report problems to support@adaptivecomputing.com

=head1 AUTHOR

Adaptive Computing Enterprises, Inc.
E<lt>http://www.adaptivecomputing.com/E<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2011-2016 by Adaptive Computing Enterprises, Inc. All Rights Reserved.

=cut

