#!/usr/bin/perl 
########################################################################
#
# Name: events2db.pl
#
# Description: Parses Moab event files and inserts the events into a database.
#
# Copyright (C) 2011-2016 by Adaptive Computing Enterprises, Inc.
# All Rights Reserved.
#
########################################################################

use strict;
use warnings;

use DBI;
use Getopt::Long;
use Pod::Usage;
use Carp qw(carp cluck croak confess);
use Data::Dumper;
use EventParser;

# Set up constants
# 'Events' stores general Event data with foreign keys
# for ObjectType and EventType
use constant EVENTS_TABLE => 'Events';
# 'EventType' maps foreign keys to event type strings
use constant EVENTTYPE_TABLE => 'EventType';
# 'ObjectType' maps foreign keys to object type strings
use constant OBJECTTYPE_TABLE => 'ObjectType';
# 'ParsedEventFiles' stores the files that have already been parsed
use constant PARSEDEVENTFILES_TABLE => 'ParsedEventFiles';
# 'EventsExt' stores VMID, ProxyUser, _USER, and UNAME per event
# This allows for significant improvements to query by user or VMID
use constant EVENTSEXT_TABLE => 'EventsExt';
# 'JobEvents' stores additional wiki information per 'job' event
# This is currently used by Viewpoint's Reporting module
use constant JOBEVENTS_TABLE => 'JobEvents';
# 'VMEvents' stores additional wiki information per 'vm' event
# This is currently used by Viewpoint's Reporting module
use constant VMEVENTS_TABLE => 'VMEvents';

MAIN:
{
  #Parse command line options and print docs if necessary
  my $man = 0;
  my $help = 0;
  my $dbName;
  my $dbHost = "localhost";
  my $dbPort = 3306;
  my $dbUser;
  my $dbPassword;
  my $createSchema = 0;
  my $optParseResult = GetOptions(
    "man|m" => \$man,
    "help|h|?" => \$help,
    "dbname=s" => \$dbName,
    "dbhost=s" => \$dbHost,
    "dbport=i" => \$dbPort,
    "dbuser=s" => \$dbUser,
    "dbpassword=s" => \$dbPassword,
    "create-schema" => \$createSchema
  ) or pod2usage(2);
  pod2usage({-exitval=>0,-verbose => 2}) if $man;
  pod2usage({-exitval=>0,-verbose=>1}) if ($help || scalar(@ARGV!=1));
  if (!$dbName) {
    croak "No database name specified, pass in --dbname parameter or enter ".
          "--help for more details\n";
  }
  if (!$dbUser) {
    croak "No database user specified, pass in --dbuser parameter or enter ".
          "--help for more details\n";
  }
  if (!$dbPassword) {
    print "\n\nPlease enter ${dbUser}'s password for database $dbName\n";
    print "Caution : your password will be visible in the terminal where\n";
    print "   you are typing.\n";
    $dbPassword = <STDIN>;
    chomp $dbPassword;
  }
  my $statsDir = shift(@ARGV);

  #Connect to the database and check the schema
  my $dbh = &createDbConnection($dbName,$dbHost,$dbPort,$dbUser,$dbPassword);
  &verifySchema($dbh,$createSchema);

  #Read the EventTypes and ObjectTypes mappings from the database
  my %eventTypes = &getEventTypes($dbh);
  my %objTypes = &getObjectTypes($dbh);

  #print "Reading new or modified event files from $statsDir.\n";
  croak "$statsDir is not a directory\n" if (! (-d $statsDir));
  croak "$statsDir is not readable\n" if (! (-r $statsDir));

  #Parse the event files that haven't been parsed before or
  #that have been modified since the last time they were parsed
  my $statsdh;
  opendir($statsdh,$statsDir);
  my @evtFileNames = grep(/^events\./,readdir($statsdh));
  closedir($statsdh);

  my @filesToParse = &getFilesToParse($dbh, $statsDir, "^events\.");
  my $filesParsed = 0;
  if (scalar(@filesToParse) > 0 )
  {
    foreach my $fileName (@filesToParse)
    {
      my $filePath = $statsDir;
      $filePath .='/' unless $filePath =~ /\/$/;
      $filePath .=$fileName;
      print "Now parsing file $filePath (".(++$filesParsed)." of ".scalar(@filesToParse).")\n";
      my $success = 0;
      eval { $success =  &parseEventFile($dbh,$filePath,\%eventTypes,\%objTypes)};
      if ($success)
      {
        &recordParsedFile($dbh,PARSEDEVENTFILES_TABLE,$fileName,time());
      }
      else { carp "WARNING: Could not parse $filePath\n$@\n"; }
    }
  } else {
    print "Database is already up to date.  No new or modified event files ".
      " in $statsDir since the last execution of this script\n";
  }

  #Disconnect from the database
  $dbh->disconnect();

  exit 0;
} # END MAIN


#----------------------------------------------------------------------
# my @filesToParse = getFilesToParse($databaseHandle,
#      "/opt/moab/stats","^events\.");
#
# Parameters:
#   $dbh a handle to the database
#   $statsDir The path to the directory where Moab stores event files
#   $eventFileRegEx A regular expression that matches file names of valid
#     event files
#
# Returns an array of all event files that have either never been parsed
# or that have been modified since the last time this script parsed them
#----------------------------------------------------------------------
sub getFilesToParse($$$)
{
  my ($dbh, $statsDir, $eventFileRegEx) = @_;
  confess '$dbh parameter not passed to '.(caller(0))[3] if !$dbh;
  confess '$statsDir parameter not passed to '.(caller(0))[3] if !$statsDir;
  confess '$eventFileRegEx parameter not passed to '.(caller(0))[3] if !$eventFileRegEx;
  my $statsdh;
  opendir($statsdh,$statsDir);
  my @evtFileNames = grep(/$eventFileRegEx/,readdir($statsdh));
  closedir($statsdh);

  my %parseTimes = &getParseTimes($dbh);
  my @filesToParse;
  foreach my $fileName (@evtFileNames)
  {
    my @fileStats = stat($statsDir.'/'.$fileName);
    my $lastModifyTime = $fileStats[9];
    if (!exists $parseTimes{$fileName}  || $parseTimes{$fileName}<$lastModifyTime)
    {
      push @filesToParse, $fileName;
    }
  }
  return @filesToParse;
} # END sub getFilesToParse


#----------------------------------------------------------------------
# my %parseTimes = getParseTimes($databaseHandle)
#
# Returns a hash of event file names mapped to the last epoch time
# (in seconds) since those files were parsed.  For example this
# method might return something like
#       ( 'events.Fri_Jun_03_201' => '1307482860',
#         'events.Sat_Jun_04_2011' => '1307482865' )
#
#----------------------------------------------------------------------
sub getParseTimes($)
{
  my ($dbh) = @_;
  confess '$dbh parameter not passed to '.(caller(0))[3] if !$dbh;
  my $sth = $dbh->prepare( "select FileName,LastParseTime from ".
    PARSEDEVENTFILES_TABLE." order by LastParseTime asc;");

  my %parseTimes;
  $sth->execute();
  while ( my $row = $sth->fetchrow_hashref() )
  {
    my $fileName = $row->{FileName};
    #Since our SQL query orders in ascending order, if we have two
    #rows for the same file we can use the most recent LastParseTime
    my $lastParseTime = $row->{LastParseTime};
    $parseTimes{$fileName} = $lastParseTime;
  }
  $sth->finish();
  return %parseTimes;
} # END sub getParseTimes


#----------------------------------------------------------------------
# verifySchema($databaseHandle,$createSchema)
#
# Verifies that the database schema is correct.
#
# Parameters :
#   $databaseHandle the database handle
#   $createSchema if true (1), the database schema will be created if it does not
#      exist. If false (0) the program will terminate with an error message
#----------------------------------------------------------------------
sub verifySchema($$)
{
  my ($dbh, $createSchema) = @_;
  confess '$dbh parameter not passed to '.(caller(0))[3] if !$dbh;
  if (!&isDbSchemaCorrect($dbh,!$createSchema))
  {
    if ($createSchema)
    {
      &createDbSchema($dbh);
    }
    else
    {
      croak "The database schema is not correct.  To create the schema, pass ".
            "in the --create-schema option\n";
    }
  }
} # END sub verifySchema

#----------------------------------------------------------------------
# createDbSchema($databaseHandle)
#
# Creates tables in the database if they don't already exist
#----------------------------------------------------------------------
sub createDbSchema($)
{
  my ($dbh) = @_;
  confess '$dbh parameter not passed to createDbSchema' if !$dbh;
  if (!&doesTableExist($dbh,EVENTTYPE_TABLE))
  {
    print "Creating database table ".EVENTTYPE_TABLE."\n";
    &createEventTypeTable($dbh,EVENTTYPE_TABLE);
  }

  if (!&doesTableExist($dbh,OBJECTTYPE_TABLE))
  {
    print "Creating database table ".OBJECTTYPE_TABLE."\n";
    &createObjectTypeTable($dbh,OBJECTTYPE_TABLE);
  }

  if (!&doesTableExist($dbh,PARSEDEVENTFILES_TABLE))
  {
    print "Creating database table ".PARSEDEVENTFILES_TABLE."\n";
    &createParsedEventFilesTable($dbh,PARSEDEVENTFILES_TABLE);
  }

  if (!&doesTableExist($dbh,EVENTSEXT_TABLE))
  {
    print "Creating database table ".EVENTSEXT_TABLE."\n";
    &createEventsExtTable($dbh,EVENTSEXT_TABLE);
  }

  if (!&doesTableExist($dbh,EVENTS_TABLE))
  {
    print "Creating database table ".EVENTS_TABLE."\n";
    &createEventsTable($dbh,EVENTS_TABLE);
  }

  if (!&doesTableExist($dbh,VMEVENTS_TABLE))
  {
    print "Creating database table ".VMEVENTS_TABLE."\n";
    &createVMEventsTable($dbh,VMEVENTS_TABLE);
  }

  if (!&doesTableExist($dbh,JOBEVENTS_TABLE))
  {
    print "Creating database table ".JOBEVENTS_TABLE."\n";
    &createJobEventsTable($dbh,JOBEVENTS_TABLE);
  }
} # END sub createDbSchema


#----------------------------------------------------------------------
# my $isSchemaCorrect = isDbSchemaCorrect($databaseHandle,$printWarningsToStderr)
#
# Parameters:
#    $databaseHandle the handle to the database
#    $printWarningsToStderr if true, names of tables that don't exist will
#      be printed to STDERR
#
# Returns true (1) if the database has the correct schema, false (0)
# otherwise
#
#----------------------------------------------------------------------
sub isDbSchemaCorrect($$)
{
  my ($dbh, $printWarnings) = @_;
  confess '$dbh parameter not passed to isDbSchemaCorrect' if !$dbh;
  #TODO Right now we are just checking if the tables exist.  We
  #could also make sure the columns are correct
  my @expectedTables = (EVENTTYPE_TABLE, OBJECTTYPE_TABLE, PARSEDEVENTFILES_TABLE,
    EVENTSEXT_TABLE, EVENTS_TABLE, JOBEVENTS_TABLE, VMEVENTS_TABLE);

  foreach my $table (@expectedTables)
  {
    if (!&doesTableExist($dbh,$table))
    {
      print STDERR "WARNING: The $table table does not exist!\n" if $printWarnings;
      return 0;
    }
  }
  return 1;
} # END sub isDbSchemaCorrect




#---------------------------------------------------------------------
# %eventTypeHash = getEventTypes($databaseHandle);
#
# Queries the database and returns the EventType to EventTypeID mappings.
# After this call %eventTypeHash should contain mappings like('GEVENT' => 1,
# 'JOBCANCEL' => 2, 'JOBEND' => 3)
#
#---------------------------------------------------------------------
sub getEventTypes($)
{
  my ($dbh) = @_;
  confess '$dbh parameter not passed to getEventTypes' if !$dbh;
  if (!&doesTableExist($dbh,EVENTTYPE_TABLE))
  {
    confess EVENTTYPE_TABLE." table does not exist.\n";
  }
  return EventParser::getNameToIDMappings($dbh,EVENTTYPE_TABLE);
} # END sub getEventTypes

#---------------------------------------------------------------------
# %objTypeHash = getObjectTypes($databaseHandle);
#
# Queries the database and returns the ObjectType to ObjectTypeID mappings.
# After this call %objTypeHash should contain mappings like( 'Group' => 5,
# 'Job' => 6, 'Node' => 7)
#
#---------------------------------------------------------------------
sub getObjectTypes($)
{
  my ($dbh) = @_;
  confess '$dbh parameter not passed to getObjectTypes' if !$dbh;
  if (!&doesTableExist($dbh,OBJECTTYPE_TABLE))
  {
    confess OBJECTTYPE_TABLE." table does not exist\n";
  }
  return EventParser::getNameToIDMappings($dbh,OBJECTTYPE_TABLE);
} # END getObjectTypes


#---------------------------------------------------------------------
# createEventTypeTable($databaseHandle, $tableName);
#
# Creates the EventType table in the database
#---------------------------------------------------------------------
sub createEventTypeTable($$)
{
  my ($dbh, $tableName) = @_;
  confess '$dbh parameter not passed to createEventTypeTable' if !$dbh;
  confess '$tableName parameter not passed to createEventTypeTable' if !$dbh;
  $dbh->do("drop table if exists $tableName");
  my $ddl = <<END;
CREATE TABLE $tableName (
  ID INTEGER,
  Name VARCHAR(64),
  PRIMARY KEY (ID)
);
END
  $dbh->do($ddl);
  $dbh->do("CREATE INDEX EVENTTYPE_NAME on $tableName(Name);");
  my %eventTypes = createEventTypes();
  foreach my $eventType (keys %eventTypes)
  {
    my $eventId = $eventTypes{$eventType};
    my $evtSql = "INSERT INTO $tableName (Name,ID) VALUES ('$eventType',$eventId);";
    $dbh->do($evtSql);
  }
} # END sub createEventTypeTable


#---------------------------------------------------------------------
# createObjectTypeTable($databaseHandle, $tableName);
#
# Creates the ObjectType table in the database
#---------------------------------------------------------------------
sub createObjectTypeTable($$)
{
  my ($dbh, $tableName) = @_;
  confess '$dbh parameter not passed to '.(caller(0))[3] if !$dbh;
  confess '$tableName parameter not passed to '.(caller(0))[3] if !$tableName;
  $dbh->do("drop table if exists $tableName");
  my $ddl = <<END;
CREATE TABLE $tableName (
  ID INTEGER,
  Name VARCHAR(64),
  PRIMARY KEY (ID)
);
END
  $dbh->do($ddl);
  $dbh->do("CREATE INDEX OBJECTTYPE_NAME on $tableName(Name);");
  my %objTypes = createObjectTypes();
  foreach my $objType (keys %objTypes)
  {
    my $objId = $objTypes{$objType};
    my $objSql = "INSERT INTO $tableName (Name,ID) VALUES ('$objType',$objId);";
    $dbh->do($objSql);
  }
} # END sub createObjectTypeTable

#---------------------------------------------------------------------
# createParsedEventFilesTable($databaseHandle, $tableName);
#
# Creates the ParsedEventFiles table in the database
#---------------------------------------------------------------------
sub createParsedEventFilesTable($$)
{
  my ($dbh, $tableName) = @_;
  confess '$dbh parameter not passed to '.(caller(0))[3] if !$dbh;
  confess '$tableName parameter not passed to '.(caller(0))[3] if !$tableName;
  my $ddl = <<END;
CREATE TABLE $tableName (
  ID int(11) auto_increment,
  FileName varchar(255),
  LastParseTime int(10),
  PRIMARY KEY (ID),
UNIQUE KEY FileName (FileName)
);
END
  $dbh->do($ddl);
  $dbh->do("CREATE INDEX PARSEDEVENTFILES_NAME on $tableName(FileName);");
} # END sub createParsedEventFilesTable


#---------------------------------------------------------------------
# createEventsExtTable($databaseHandle, $tableName);
#
# Creates the EventsExt table in the database
#---------------------------------------------------------------------
sub createEventsExtTable($$)
{
  my ($dbh, $tableName) = @_;
  confess '$dbh parameter not passed to '.(caller(0))[3] if !$dbh;
  confess '$tableName parameter not passed to '.(caller(0))[3] if !$tableName;
  my $ddl = <<END;
CREATE TABLE `$tableName` (
      `ID` int(11) NOT NULL DEFAULT '0',
      `ProxyUser` varchar(64) DEFAULT NULL,
      `VMID` varchar(64) DEFAULT NULL,
      `_USER` varchar(64) DEFAULT NULL,
      `UNAME` varchar(64) DEFAULT NULL,
       PRIMARY KEY (`ID`),
       KEY `EVENTS_PUSER` (`ProxyUser`),
       KEY `EVENTS_VMID` (`VMID`),
       KEY `EVENTS_USER` (`_USER`),
       KEY `EVENTS_UNAME` (`UNAME`)
)
ENGINE = MyISAM
CHARACTER SET utf8 COLLATE utf8_general_ci;
END
  $dbh->do($ddl);
} # END sub createEventsExtTable


#---------------------------------------------------------------------
# createEventsTable($databaseHandle, $tableName);
#
# Creates the Events table in the database
#---------------------------------------------------------------------
sub createEventsTable($$)
{
  my ($dbh, $tableName) = @_;
  confess '$dbh parameter not passed to '.(caller(0))[3] if !$dbh;
  confess '$tableName parameter not passed to '.(caller(0))[3] if !$tableName;
  my $ddl = <<END;
CREATE TABLE $tableName (
  ID INTEGER,
  ObjectType INTEGER,
  EventType INTEGER,
  EventTime INTEGER UNSIGNED,
  ObjectName VARCHAR(64),
  Name VARCHAR(64),
  Description TEXT,
  PRIMARY KEY (ID)
);
END
  $dbh->do($ddl);
  $dbh->do('CREATE INDEX EVENTS_NAME on Events (Name);');
  $dbh->do('CREATE INDEX EVENTS_TIME on Events (EventTime);');
  $dbh->do('CREATE INDEX EVENTS_ETYPE on Events (EventType);');
  $dbh->do('CREATE INDEX EVENTS_OTYPE on Events (ObjectType);');
  $dbh->do('CREATE INDEX EVENTS_ONAME on Events (ObjectName);');
} # END sub createEventsTable

#---------------------------------------------------------------------
# createVMEventsTable($databaseHandle, $tableName);
#
# Creates the VMEvents table in the database
#---------------------------------------------------------------------
sub createVMEventsTable($$)
{
  my ($dbh, $tableName) = @_;
  confess '$dbh parameter not passed to '.(caller(0))[3] if !$dbh;
  confess '$tableName parameter not passed to '.(caller(0))[3] if !$tableName;
  my $ddl = <<END;
CREATE TABLE `$tableName` (
      `ID` int(11) NOT NULL DEFAULT '0',
      `SOVEREIGN` varchar(64) DEFAULT NULL,
      `POWERSTATE` varchar(64) DEFAULT NULL,
      `CONTAINERNODE` varchar(64) DEFAULT NULL,
      `RCSWAP` varchar(64) DEFAULT NULL,
      `RCMEM` varchar(64) DEFAULT NULL,
      `RCDISK` varchar(64) DEFAULT NULL,
      `RCPROC` varchar(64) DEFAULT NULL,
      `OS` varchar(64) DEFAULT NULL,
      `CLASS` varchar(64) DEFAULT NULL,
      `VARIABLE` varchar(255) DEFAULT NULL,
      `MSG` TEXT DEFAULT NULL,
       PRIMARY KEY (`ID`)
)
ENGINE = MyISAM 
CHARACTER SET utf8 COLLATE utf8_general_ci;
END
  $dbh->do($ddl);
} # END sub createVMEventsTable

#---------------------------------------------------------------------
# createJobEventsTable($databaseHandle, $tableName);
#
# Creates the JobEvents table in the database
#---------------------------------------------------------------------
sub createJobEventsTable($$)
{
  my ($dbh, $tableName) = @_;
  confess '$dbh parameter not passed to '.(caller(0))[3] if !$dbh;
  confess '$tableName parameter not passed to '.(caller(0))[3] if !$tableName;
  my $ddl = <<END;
CREATE TABLE `$tableName` (
      `ID` int(11) NOT NULL DEFAULT '0',
      `JOBID` varchar(64) DEFAULT NULL,
      `JOBTYPE` varchar(64) DEFAULT NULL,
      `ACTION` TEXT DEFAULT NULL,
      `TEMPLATE` varchar(64) DEFAULT NULL,
       PRIMARY KEY (`ID`)
)
ENGINE = MyISAM
CHARACTER SET utf8 COLLATE utf8_general_ci;
END
  $dbh->do($ddl);
} # END sub createJobEventsTable

#---------------------------------------------------------------------
# my %objTypesHash = createObjectTypes();
#
# Creates a hash of ObjectType to ObjectTypeIDs according to the Moab 7.1
# database schema which can be used in creating an ObjectType table in the
# database that is identical to the one created by the Moab 7.1 schema.
#
# Sync with moab-db-mysql-create.sql
#---------------------------------------------------------------------
sub createObjectTypes()
{
  my %objTypes = (
  'Acct' => '1',
  'AM' => '2',
  'Class' => '3',
  'Cluster' => '4',
  'Group' => '5',
  'Job' => '6',
  'Node' => '7',
  'Par' => '8',
  'QOS' => '9',
  'Rsv' => '13',
  'RM' => '14',
  'Sched' => '15',
  'SRsv' => '16',
  'Sys' => '17',
  'TNode' => '18',
  'Trig' => '19',
  'User' => '20',
  'CJob' => '23',
  'GRes' => '30',
  'Gmetric' => '31',
  'Stats' => '39',
  'TJob' => '42',
  'Paction' => '43',
  'VM' => '45',
  'VPC' => '47',
  'JGroup' => '48'
  );
  return %objTypes;
} # END sub createObjectTypes


#---------------------------------------------------------------------
# my %eventTypesHash = createEventTypes();
#
# Creates a hash of EventType to EventTypeIDs according to the Moab 7.1
# database schema which can be used in creating an EventType table in the
# database that is identical to the one created by the Moab 7.1 schema.
#
# Sync with moab-db-mysql-create.sql
#---------------------------------------------------------------------
sub createEventTypes()
{
  my %eventTypes = (
  '[NONE]' => '0',
  'GEVENT' => '1',
  'JOBCANCEL' => '2',
  'JOBEND' => '3',
  'JOBFAILURE' => '4',
  'JOBHOLD' => '5',
  'JOBMIGRATE' => '6',
  'JOBMODIFY' => '7',
  'JOBPREEMPT' => '8',
  'JOBREJECT' => '9',
  'JOBRESUME' => '10',
  'JOBSTART' => '11',
  'JOBSUBMIT' => '12',
  'JOBUPDATE' => '13',
  'JOBVARSET' => '14',
  'JOBVARUNSET' => '15',
  'NODEDOWN' => '16',
  'NODEFAILURE' => '17',
  'NODEUP' => '18',
  'NODEUPDATE' => '19',
  'NOTE' => '20',
  'QOSVIOLATION' => '21',
  'RMDOWN' => '22',
  'RMUP' => '23',
  'RSVCANCEL' => '24',
  'RSVCREATE' => '25',
  'RSVEND' => '26',
  'RSVMODIFY' => '27',
  'RSVSTART' => '28',
  'SCHEDCOMMAND' => '29',
  'SCHEDFAILURE' => '30',
  'SCHEDMODIFY' => '31',
  'SCHEDPAUSE' => '32',
  'SCHEDRECYCLE' => '33',
  'SCHEDRESUME' => '34',
  'SCHEDSTART' => '35',
  'SCHEDSTOP' => '36',
  'TRIGCREATE' => '37',
  'TRIGEND' => '38',
  'TRIGFAILURE' => '39',
  'TRIGSTART' => '40',
  'TRIGTHRESHOLD' => '41',
  'VMCREATE' => '42',
  'VMDESTROY' => '43',
  'VMMIGRATE' => '44',
  'VMPOWERON' => '45',
  'VMPOWEROFF' => '46',
  'NODEMODIFY' => '47',
  'NODEPOWEROFF' => '48',
  'NODEPOWERON' => '49',
  'NODEPROVISION' => '50',
  'ALLSCHEDCOMMAND' => '51',
  'AMDELETE' => '52',
  'AMEND' => '53',
  'AMQUOTE' => '54',
  'AMSTART' => '55',
  'RMPOLLEND' => '56',
  'RMPOLLSTART' => '57',
  'SCHEDCYCLEEND' => '58',
  'SCHEDCYCLESTART' => '59',
  'JOBCHECKPOINT' => '60',
  'AMCREATE' => '61',
  'AMUPDATE' => '62',
  'AMPAUSE' => '63',
  'AMRESUME' => '64',
  'AMMODIFY' => '65',
  'AMCONTINUE' => '66',
  );
  return %eventTypes;
} # END sub createEventTypes


#--------------------------------------------------------------------------
# if (!doesTableExist($databaseHandle,"Events"))
# {
#    # then do something like warn the user or create the table
#
# Returns true (1) if the table exists in the database, false (0) otherwise
#--------------------------------------------------------------------------
sub doesTableExist($$)
{
  my ($dbh, $table) = @_;
  confess '$dbh parameter not passed to doesTableExist' if !$dbh;
  confess '$table parameter not passed to doesTableExist' if !$table;
  my @tables = $dbh->tables('','','','TABLE');
  foreach my $tableEntry (@tables)
  {
    if ( $tableEntry =~ /(`[^`]+`\.)?`([^`]+)`/ )
    {
      my $tableName = $2;
      # Some MySQL engines force table names to lower case
      # Odd, but a necessary workaround
      return 1 if lc $tableName eq lc $table;
    }
  }
  return 0;
} # END sub doesTableExist

#----------------------------------------------------------------------------
# my $databaseHandle = &createDbConnection($dbName,$dbHost,$dbPort,
#   $dbUser,$dbPassword);
#
# Creates a database handle using the passed in parameters
#----------------------------------------------------------------------------
sub createDbConnection($$$$$)
{
  my ($dbName, $dbHost, $dbPort, $dbUser, $dbPassword) = @_;
  confess '$dbName parameter not passed to createDbConnection' if !$dbName;
  confess '$dbHost parameter not passed to createDbConnection' if !$dbHost;
  confess '$dbPort parameter not passed to createDbConnection' if !$dbPort;
  confess '$dbUser parameter not passed to createDbConnection' if !$dbUser;
  confess '$dbPassword parameter not passed to createDbConnection' if !$dbPassword;
  my $db = "DBI:mysql:database=$dbName;host=$dbHost;port=$dbPort";
  my $dbh = DBI->connect($db, $dbUser, $dbPassword,{RaiseError => 1})
	  or confess $!;
  return $dbh;
} # END sub createDbConnection

#----------------------------------------------------------------------------
# my $success = parseEventFile($databaseHandle, $eventFilePath, \%eventTypes, \%objectTypes);
#
# Parses a Moab event file and inserts the events into the database.
# Parameters
#   databaseHandle : the handle to the database
#   eventFilePath : The path to a Moab event file
#   eventTypes : a reference to a hash that maps EventTypes to EventTypeIDs
#   objectTypes : a reference to a hash that maps ObjectTypes to ObjectTypeIDs
#
# Return
#   true (1) if parsing was successful, false (0) otherwise.
#----------------------------------------------------------------------------
sub parseEventFile($$$$)
{
  my ($dbh, $eventFile, $eventTypes, $objectTypes) = @_;
  confess '$dbh parameter not passed to '.(caller(0))[3] if !$dbh;
  confess 'parseFiles not passed $eventTypes reference' if (!$eventTypes);
  confess 'parseFiles not passed $objectTypes reference' if (!$objectTypes);
  if (!$eventFile)
  {
    cluck '$eventFile is undefined, cannot parse an undefined file';
    return 0;
  }
  if (!(-r $eventFile))
  {
    cluck "Can't read from event file $eventFile";
    return 0;
  }
  if (!(-T $eventFile))
  {
    cluck "Event file $eventFile is not a text file" ;
    return 0;
  }
  my $linesInFile = EventParser::countLinesInFile($eventFile);
  my $eventfh;
  open($eventfh,$eventFile) or confess "Can't open $eventFile for reading : $!";
  my $linesParsed = 0;
  while (defined(my $line=<$eventfh>))
  {
    chomp $line;
    eval { &parseAndSaveEvent($dbh,$eventTypes,$objectTypes,$line) if $line };
    carp "$@" if ($@);

    $linesParsed++;
    # Every 1000th line print the percentage done so the user
    # can see progress
    if ($linesParsed % 1000 == 0 && $linesInFile>0) #Avoid divide by 0 later. Make sure this variable is >0
    {
      my $percentageDone = sprintf("%.2f",($linesParsed/$linesInFile)*100);
      local $| = 1;
      print "File progress : $percentageDone% \r" if ($linesInFile>0);
    }
  } # END while
  print "\r";
  return 1;
} # END sub parseEventFile

#----------------------------------------------------------------------------
#  parseAndSaveEvent($dbh,$eventTypes,$objectTypes,$line);
#
#  Parses an event from single line from Moab event file and inserts the
#  event into the database.
#
#  Parameters
#   databaseHandle : the handle to the database
#   eventTypes : a reference to a hash that maps EventTypes to EventTypeIDs
#   objectTypes : a reference to a hash that maps ObjectTypes to ObjectTypeIDs
#   line : A single line from a Moab event file that represents an event
#----------------------------------------------------------------------------
sub parseAndSaveEvent($$$$)
{
  my ($dbh,$eventTypes,$objectTypes,$line) = @_;
  confess '$dbh parameter not passed to '.(caller(0))[3] if !$dbh;
  confess 'parseFiles not passed $eventTypes reference' if (!$eventTypes);
  confess 'parseFiles not passed $objectTypes reference' if (!$objectTypes);
  confess '$line parameter not passed to '.(caller(0))[3] if !$line;
  my %event = EventParser::parseEvent($line);
  #print "Parsed event ".$event{eventID}."\n";

  my $eventTypeId = $$eventTypes{lc $event{eventType}};
  if ($eventTypeId)  { $event{eventTypeID} = $eventTypeId; }
  elsif (!$event{eventID})
  {
    cluck("Unable to determine the event ID from the following line:\n".
          "$line\n");
  }
  elsif (!$event{eventType})
  {
    cluck("Unable to determine the EventType from the following line:\n".
          "$line\n");
  }
  else
  {
    cluck("$event{eventType} is not a recognized Moab EventType.  Cannot".
          "determine EventType ID for event $event{eventID}.\n");
  }
  my $objTypeId = $$objectTypes{lc $event{objectType}};
  if ($objTypeId) { $event{objectTypeID} = $objTypeId; }
  elsif (!$event{eventID})
  {
    cluck("Unable to determine the event ID from the following line:\n".
          "$line\n");
  }
  elsif (!$event{objectType})
  {
    cluck("Unable to determine the ObjectType from the following line:\n".
          "$line\n");
  }
  else
  {
    cluck("$event{objectType} is not a recognized Moab ObjectType. Cannot".
          "determine ObjectType ID for event $event{eventID}\n");
  }
  #print Dumper(\%event);
  eval { &saveEventToEventsTable($dbh,\%event) };
  carp "\n\n$@\n" if $@;
  eval { &saveEventToEventsExtTable($dbh,\%event) };
  carp "\n\n$@\n" if $@;
  for ( lc($event{objectType}) )
  {
    if (/^vm$/)
    {
      eval { &saveEventToVMEventsTable($dbh,\%event) };
      carp "\n\n$@\n" if $@;
    }
    elsif (/^job$/)
    {
      eval { &saveEventToJobEventsTable($dbh,\%event) };
      carp "\n\n$@\n" if $@;
    }
  } # END for
} # END sub parseAndSaveEvent


#-------------------------------------------------------------------------------
# recordParsedFile($databaseHandle,$tableName,$filePath,$parseTime);
#
# Parameters:
#    $databaseHandle a handle to the database
#    $tableName The name of the database table where parsed event information is
#    $fileName the name of the events file
#    $parseTime the epoch time in seconds when the file was parsed
#
# Adds an entry in the database indicating that an events file has been
# parsed
#-------------------------------------------------------------------------------
sub recordParsedFile($$$$)
{
  my ($dbh,$tableName,$fileName,$parseTime) = @_;
  confess '$dbh parameter not passed to '.(caller(0))[3] if !$dbh;
  confess '$tableName parameter not passed to '.(caller(0))[3] if !$tableName;
  confess '$fileName parameter not passed to '.(caller(0))[3] if !$fileName;
  confess '$parseTime parameter not passed to '.(caller(0))[3] if !$parseTime;

  #See if there is already a row in the database for this file
  my $rowCount = 0;
  my $sth = $dbh->prepare( "select count(*) as Count from ".
    PARSEDEVENTFILES_TABLE." where FileName='$fileName'");
  $sth->execute();
  while ( my $row = $sth->fetchrow_hashref() )
  {
    $rowCount = $row->{Count};
  }
  $sth->finish();

  #if there is already a row in the database for this file, update the row
  #otherwise create a new row
  my $evtSql;
  if ($rowCount == 0)
  {
    $evtSql = "INSERT INTO $tableName (FileName,LastParseTime) VALUES ".
      "('$fileName',$parseTime);";
  }
  else
  {
    $evtSql = "UPDATE $tableName SET LastParseTime=$parseTime where FileName='$fileName'";
  }
  $dbh->do($evtSql);
} # END sub recordParsedFile


#-------------------------------------------------------------------------------
# saveEventToEvents($databaseHandle,\%event);
#
# Parameters:
#   $databaseHandle : a handle to the database
#   $event a reference to a hash representing a Moab event
#
# Saves the the portion of the event that goes into the Events table into
# the database
#-------------------------------------------------------------------------------
sub saveEventToEventsTable
{
  my ( $dbh, $event ) = @_;
  return unless $event && $event->{eventID} && $dbh;

  if (!isEventIdInTable($dbh,EVENTS_TABLE,"ID",$event->{eventID}))
  {
    my $sqlInsertEvent = "INSERT INTO ".EVENTS_TABLE." VALUES (?,?,?,?,?,?,?)";
    $dbh->do(
      $sqlInsertEvent,        undef,                 $event->{eventID},
      $event->{objectTypeID}, $event->{eventTypeID}, $event->{epochTime},
      $event->{objectName},   $event->{eventName},   $event->{desc}
    );
  }
} # END sub saveEventToEventsTable

#-------------------------------------------------------------------------------
# saveEventToEventsExtTable($databaseHandle,\%event);
#
# Parameters:
#   $databaseHandle : a handle to the database
#   $event : a reference to a hash representing a Moab event
#
# Saves the the portion of the event that goes into the EventsExt table into
# the database
#-------------------------------------------------------------------------------
sub saveEventToEventsExtTable
{
  my ( $dbh, $event) = @_;
  return unless $event && $event->{eventID} && $dbh;

  if (!isEventIdInTable($dbh,EVENTSEXT_TABLE,"ID",$event->{eventID}))
  {
    my $sql = "INSERT INTO ".EVENTSEXT_TABLE." VALUES (?,?,?,?,?)";
    $dbh->do(
      $sql, undef,
      $event->{eventID},
      $event->{params}->{ProxyUser},
      $event->{params}->{VMID},
      $event->{params}->{USER},
      $event->{params}->{UNAME}
    );
  }
} # END sub saveEventToEventsExtTable

#-------------------------------------------------------------------------------
# saveEventToVMEventsTable($databaseHandle,\%event);
#
# Parameters:
#   $databaseHandle : a handle to the database
#   $event : a reference to a hash representing a Moab event
#
# Saves the the portion of the event that goes into the EventsExt table into
# the database
#-------------------------------------------------------------------------------
sub saveEventToVMEventsTable
{
  my ( $dbh, $event) = @_;
  return unless $event && $event->{eventID} && $dbh;

  if (!isEventIdInTable($dbh,VMEVENTS_TABLE,"ID",$event->{eventID}))
  {
    my $sql = "INSERT INTO ".VMEVENTS_TABLE." VALUES (?,?,?,?,?,?,?,?,?,?,?,?)";
    $dbh->do(
      $sql, undef, $event->{eventID},
      $event->{params}->{SOVEREIGN},     $event->{params}->{POWERSTATE},
      $event->{params}->{CONTAINERNODE}, $event->{params}->{RCSWAP},
      $event->{params}->{RCMEM},         $event->{params}->{RCDISK},
      $event->{params}->{RCPROC},        $event->{params}->{OS},
      $event->{params}->{CLASS},         $event->{params}->{VARIABLE},
      $event->{params}->{MSG}
    );
  }
} # END sub saveEventToVMEventsTable

#-------------------------------------------------------------------------------
# saveEventToJobEventsTable($databaseHandle,\%event);
#
# Parameters:
#   $databaseHandle : a handle to the database
#   $event : a reference to a hash representing a Moab event
#
# Saves the the portion of the event that goes into the EventsExt table into
# the database
#-------------------------------------------------------------------------------
sub saveEventToJobEventsTable
{
  my ( $dbh, $event) = @_;
  return unless $event && $event->{eventID} && $dbh;

  if (!isEventIdInTable($dbh,JOBEVENTS_TABLE,"ID",$event->{eventID}))
  {
    my $sql = "INSERT INTO ".JOBEVENTS_TABLE." VALUES (?,?,?,?,?)";
    $dbh->do(
      $sql, undef,
      $event->{eventID},
      $event->{descWords}->[0], # Assume the first descWord is the JobID
                                # descWords are non-wiki strings in the
                                # description (i.e. no "=" in the string)
      $event->{params}->{JobType},
      $event->{params}->{Action},
      $event->{params}->{TEMPLATE},
    );
  }
} # END sub saveEventToJobEventsTable

#-------------------------------------------------------------------------------
# my $eventExists = isEventIdInTable($databaseHandle,"Events","ID","12385713")
#
# Parameters:
#   $databaseHandle : a handle to the database
#   $tableName : the name of the database table where the Events are stored
#   $idColumnName : the name of the column with the Event ID
#   $eventId : the event ID
#
# Returns true if an event with the eventId passed in already exists in
# the database table, false otherwise
#
#-------------------------------------------------------------------------------
sub isEventIdInTable($$$)
{
  my ($dbh, $tableName, $idColumnName, $eventId) = @_;
  confess '$dbh parameter not passed to '.(caller(0))[3] if !$dbh;
  confess '$tableName parameter not passed to '.(caller(0))[3] if !$tableName;
  confess '$idColumnName parameter not passed to '.(caller(0))[3] if !$idColumnName;
  confess '$eventId parameter not passed to '.(caller(0))[3] if !$eventId;

  my $querySql = "select count(*) as Count from $tableName where $idColumnName='".$eventId."'";
  my $sth = $dbh->prepare($querySql);
  $sth->execute();
  my $eventCount = 0;
  while ( my $row = $sth->fetchrow_hashref() )
  {
    $eventCount = $row->{Count};
  }
  $sth->finish();
  return $eventCount>0;
} # END sub isEventIdInTable

__END__

=head1 NAME

events2db.pl - Parses Moab event files and inserts the events into a database

=head1 SYNOPSIS

B<events2db.pl> [B<-m>|B<--man>] [B<-h>|B<-?>|B<--help>] [B<--create-schema>] B<--dbname=<databasename>>
  B<--dbuser=<databaseuser>> B<--dbpassword=<databasepassword>> B<<moabStatsDir>>

where <moabStatsDir> is the path to the moab stats directory where the event
files are located.  This directory is /opt/moab/stats by default.

=head1 DESCRIPTION

events2db.pl Parses the event files in the Moab stats directory and inserts
these events into a MySql database.  The Moab stats directory is typically
/opt/moab/stats  Event filenames are assumed to begin with the prefix
'events.'.  For example events.Mon_Jun_06_2011, events.Sat_Jun_04_2011, and events.Sun_Jun_05_2011 are
all valid event file names.

The first time you run events2.pl you will probably want to use the --create-schema option so that
the necessary database tables can be created.  For example you might run

events2db.pl --dbname=Moab --dbuser=bob --dbpassword=p@ssw0rd --create-schema /opt/moab/stats

events2db.pl also records which files it has successfully parsed in the database so that it can
be called repeatedly and it will only parse new event files or event files that were modified
since the last time events2db.pl was run.

events2db.pl --dbname=Moab --dbuser=bob --dbpassword=p@ssw0rd /opt/moab/stats


=head1 OPTIONS

=over 4

=item B<dbname> (Required)

The name of the MySql database where the events will be inserted

=item B<dbuser> (Required)

The database user

=item B<dbpassword>

The password of the database user.  If not specified the program will prompt
for the password

=item B<dbhost>

The host where the database is located.  Defaults to localhost if not specified

=item B<dbport>

The database port. Defaults to 3306 if not specified

=item B<create-schema>

If specified, --create-schema will cause the database schema to be created if it
does not already exist before the event files are parsed

=item B<-m>|B<--man>

Displays man page information

=item B<-h>|B<-?>|B<--help>

Displays usage information

=back

=head1 EXAMPLES

events2db.pl --dbname=Moab --dbuser=bob --dbpassword=p@ssw0rd --create-schema /opt/moab/stats

events2db.pl --dbname=Moab --dbuser=bob --dbpassword=p@ssw0rd /opt/moab/stats


=head1 REQUIRES

DBI
Getopt::Long
Pod::Usage
Data::Dumper
EventParser

=head1 REPORTING BUGS

Report problems to support@adaptivecomputing.com

=head1 AUTHOR

Adaptive Computing Enterprises, Inc.
E<lt>http://www.adaptivecomputing.com/E<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2011-2016 by Adaptive Computing Enterprises, Inc. All Rights Reserved.

=cut

