#! /usr/bin/perl -w
################################################################################
#
# Usage: __mon.ipmi.pl [--start] [--stop] [--restart]
#
# This script retrieves IPMI sensor data from cluster nodes.  It is spawned
# from the ipmi.mon.pl script.  It refreshes the .ipmicache.gm file every
# $sleep seconds.
# This script can be invoked in daemon mode which does this in an endless loop.
# If invoked with no arguments, the monitor will run once and exit.
#
# Note: Requires XML::LibXML
#
#   Copyright (C) 2005-2016 by Adaptive Computing Enterprises, Inc. All Rights Reserved.
#
################################################################################
use strict;
use FindBin;
use lib "/make_install/lib/perl5", "/make_install/etc";             # Installed
use lib "${FindBin::Bin}/../../lib/perl5", "${FindBin::Bin}/../../etc";   # Configured
use Getopt::Long;
use XML::LibXML;
#use Data::Dumper;
use Moab::Tools;    # Required before including config to set $homeDir
BEGIN { require "config.ipmi.pl"; }
our ($homeDir, $ipmifile, $daemon, $bmcaddrmap, $pollInterval, $ipmiuser,
    $ipmipass);

# Parse Command Line Arguments
my ($help, $start, $stop, $restart);
my $usage = "Usage: $0 [--start] [--stop] [--restart]\n";
GetOptions(
    'start|s' => \$start,
    'stop|k'  => \$stop,
    'restart' => \$restart,
    'help|?'  => \$help,
) or die $usage;
die $usage if $help;

# If shutdown or restart requested
my $pidFile = "${homeDir}/spool/ipmimon.pid";
if ($stop || ($restart && -e "$pidFile"))
{
    # Get the Pid of the process to kill.
    open(PID, "$pidFile")
      || die "Cannot open $pidFile for reading: $!\nIs the monitor running?\n";
    <PID> =~ /^(\d+)$/;
    my $pid = $1;
    close PID;
    die "Invalid pid in $pidFile" unless $pid;

    # Remove the PID file before the process dies
    unlink $pidFile;

    # Issue the kill to every process of the process group of the stored pid
    print "Killing Pid ($pid)\n";
    kill 15, -$pid or die "Unable to kill monitor with Pid ($pid): $!\n";

    # exit if shutdown option was specified
    exit 0 if $stop;
}

# If called with start or restart option, daemonize and monitor in a loop,
# waiting $pollInterval seconds between invocations
if ($start || $restart)
{
    # Check to see if monitor already running
    if (open(PID, "$pidFile"))
    {
        <PID> =~ /(\d+)/;
        my $pid = $1;
        close PID;
        if ($pid && kill 0, $pid)
        {
            print "Monitor Pid ($pid) is already running\n";
            exit 0;
        }
    }

    # Daemonize
    fork && exit;
    setpgrp();

    # Write pid to pidFile
    open(PID, ">$pidFile") or die "Unable to open $pidFile for writing: $!\n";
    #select(PID); $| = 1; # Turn on autoflushing
    print PID $$;
    close PID;
    print "Starting Pid ($$)\n";

    while (1)
    {
        sleep $pollInterval;
        monitor();
    }
}

# If called without daemonize option, just monitor once and exit
# Print the contents of the node file to STDOUT
else
{
    monitor();
    exit 0;
}

# Routine to grab ipmi data from nodes and dump to file
sub monitor
{
    # Grab node info from Moab
    my $nodesraw = `mdiag -n --format=xml`;

    if ($nodesraw eq "")
    {
        return;
    }

    my $parser = new XML::LibXML();
    my $data   = $parser->parse_string($nodesraw);

    my @nodes;
    my %bmcip;

    my $root        = $data->getDocumentElement();
    my @node_struct = $root->getChildrenByTagName("node");
    foreach my $node (@node_struct)
    {
        my $name = $node->getAttribute("NODEID");
        push(@nodes, $name);
    }

    # Grab each nodes BMC IP address from mapping file
    open(BMCADDR, $bmcaddrmap);
    while (my $line = <BMCADDR>)
    {
        chomp($line);
        if ($line !~ /#/)
        {
            my @bmcinfo = split(/\s+/, $line);
            $bmcip{$bmcinfo[0]} = $bmcinfo[1];
        }
    }
    close(BMCADDR);

    # Grab IPMI info from nodes and write to $ipmifile

    #open tmp output file for writing
    open(TMPOUTFILE, ">${homeDir}/spool/$ipmifile.tmp");

    # Loop through all nodes, grab updated info and write it to file
    foreach my $node (@nodes)
    {
        # Make sure we have the node in our BMC IP map file
        if (exists $bmcip{$node})
        {
            # Refresh temperature data
            my $sensorcmd =
              "ipmitool -I lan -H $bmcip{$node} -U \"$ipmiuser\" -P \"$ipmipass\" sdr type temperature";

            # Parse the temperature data
            #node15 GMETRIC[temp]=113
            foreach my $line (`$sensorcmd`)
            {
                chomp($line);
                my @cutline = split(/\|/, $line);

                # remove trailing whitespace
                $cutline[0] =~ s/\s+$//;

                # replace space with underscore
                $cutline[0] =~ s/\s+/\_/g;

                # grab temperature
                my @temp = split(/\s+/, $cutline[4]);

                print TMPOUTFILE "$node GMETRIC\[$cutline[0]\]=$temp[1]\n";
            }

            # Refresh Power state
            my $powercmd =
              "ipmitool -I lan -H $bmcip{$node} -U \"$ipmiuser\" -P \"$ipmipass\" chassis power status";

            # Parse the chassis status
            foreach my $line (`$powercmd`)
            {
                chomp($line);
                my @cutline = split(/\s+/, $line);

                # remove trailing whitespace
                $cutline[3] =~ s/\s+$//;
                # To uppercase
                $cutline[3] =~ tr/a-z/A-Z/;
                print TMPOUTFILE "$node POWER=$cutline[3]\n";
            }
        }
    }

    close(TMPOUTFILE);
    rename("${homeDir}/spool/$ipmifile.tmp", "${homeDir}/spool/$ipmifile");

}
1;

__END__

=head1 NAME

=head1 SYNOPSIS

=head1 DESCRIPTION

=head2 EXPORT

=head2 EXPORT_OK

=head1 SEE ALSO

Additional information can be found at http://www.adaptivecomputing.com/

=head1 AUTHOR

Adaptive Computing, Inc., E<lt>http://www.adaptivecomputing.com/E<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2005-2016 by Adaptive Computing Enterprises, Inc. All Rights Reserved.

=cut
