#! /usr/bin/perl -w
################################################################################
# Usage: reservationUsage.pl <events file>[ <events file>]...
#
# This script is used to determine procseconds usage due to admin reservations
#
# NOTE: This script is only designed to work with VERSION 400
#
#   Copyright (C) 2010-2016 by Adaptive Computing Enterprises, Inc. All Rights Reserved.
#
###############################################################################
use strict;
use Getopt::Long;

# Populate this array with a list of admins who can set admin reservations
my @admins = qw(dlapine mshow louis enstrom kretzer);

my ($help);
my $usage = "Usage: $0 <events file>[ <events file>]*\n";
GetOptions('help|?' => \$help,)
  or die $usage;
die $usage if $help;

my $totalUsage           = 0;
my $version              = "";
my @rejectedReservations = ();

print "The following reservations used resources on the system:\n";
print
  "================================================================================\n";

# Process each of the files one line at a time
while (my $line = <>)
{
    chomp $line;
    next if $line =~ /^#/;    # Ignore comments
    next unless $line;        # Ignore empty lines

    # Handle version lines
    if ($line =~ /^VERSION/)
    {
        my @words = split /\s+/, $line;
        $version = $words[1];
        next;
    }

    # Break the job line into a attribute hash with named fields
    my %attr = ();

    # Parse Version 500 events
    if ($version eq "500")
    {
        (
            $attr{EventTime},              $attr{Object},
            $attr{ObjectId},               $attr{Event},
            $attr{StartTime},              $attr{EndTime},
            $attr{AllocTaskCount},         $attr{AllocNodeCount},
            $attr{TotalActiveProcSeconds}, $attr{TotalProcSeconds},
            $attr{HostList},               $attr{Owner},
            $attr{ACL}
          )
          = split /\s+/, $line;
    }

    # Parse Version 400 events
    elsif ($version eq "400")
    {
        (
            $attr{Object},                 $attr{Event},
            $attr{ObjectId},               $attr{EventTime},
            $attr{StartTime},              $attr{EndTime},
            $attr{AllocTaskCount},         $attr{AllocNodeCount},
            $attr{TotalActiveProcSeconds}, $attr{TotalProcSeconds},
            $attr{HostList},               $attr{Owner},
            $attr{ACL}
          )
          = split /\s+/, $line;
    }

    # Unknown version
    else
    {
        die "Events file format version could not be determined\n";
    }

    # Extract only rsv end entries
    next
      unless (defined $attr{Object}
        && $attr{Object} eq "rsv"
        && defined $attr{Event}
        && $attr{Event} eq "end");

    # Reject this entry if it has any ACLs outside of the admin list
    my $rejected = 0;
    foreach my $acl (split /;/, $attr{ACL})
    {
        my ($name, $cmp, $value) = split /=/, $acl;
        next unless (defined $name && defined $cmp && defined $value);
        next if $name eq "RSV";    # Ignore RSV acls
        die "Invalid value for compare flag ($cmp)\n" unless $cmp eq "%";
        $rejected = 1 unless $name eq "USER";
        $value =~ s/\+//;          # Remove the plus
        $rejected = 1 unless grep { $_ eq $value } @admins;
    }

    if ($attr{AllocNodeCount} == 0)
    {
        $attr{AllocNodeCount} = @{[split /\|/, $attr{HostList}]};
    }

    unless ($rejected)
    {
        printf("Reservation %s used %.2f proc-seconds on %d nodes\n",
            $attr{ObjectId}, $attr{TotalProcSeconds}, $attr{AllocNodeCount});
        $totalUsage += $attr{TotalProcSeconds};
    }
    else
    {
        push @rejectedReservations, \%attr;
    }
}

print
  "================================================================================\n";
printf("Total Reservation usage: %.2f proc-seconds\n", $totalUsage);
print "\nNote: The following reservations were not included:\n";
print
  "================================================================================\n";
foreach my $res (@rejectedReservations)
{
    printf(
        "Reservation %s used %.2f proc-seconds on %d nodes\n",
        $res->{ObjectId},
        $res->{TotalProcSeconds},
        $res->{AllocNodeCount}
    );
}
print
  "================================================================================\n";
print "\nWarning: This script does not check for overlapping reservations\n";
1;
__END__

=head1 NAME

=head1 SYNOPSIS

=head1 DESCRIPTION

=head2 EXPORT

=head2 EXPORT_OK

=head1 SEE ALSO

Additional information can be found at http://www.adaptivecomputing.com/

=head1 AUTHOR

Adaptive Computing, Inc., E<lt>http://www.adaptivecomputing.com/E<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2010-2016 by Adaptive Computing Enterprises, Inc. All Rights Reserved.

=cut
