#! /usr/bin/perl -w
################################################################################
#
# Usage: scrub.pl [-w <workload trace>] [-r <resource trace>] [-m <map file>]
#
# This script is used to sanitize workload and resource traces, replacing all
# real names with generic names.
#
# If you specify a map file, it will reuse the mapfile appending new values
#
# NOTE: This script is only designed to work with VERSION 5.0
#
#   Copyright (C) 2010-2016 by Adaptive Computing Enterprises, Inc. All Rights Reserved.
#
################################################################################
use strict;
use Data::Dumper;
use Getopt::Long;

# Parse Command Line Arguments
my ($help, $workloadTraceFile, $resourceTraceFile, $mapFile);
my $usage =
  "Usage: $0 [-w <workload trace>] [-r <resource trace>] [-m <map file>]\n";
GetOptions(
    'help|?' => \$help,
    'r=s'    => \$resourceTraceFile,
    'w=s'    => \$workloadTraceFile,
    'm=s'    => \$mapFile,
  )
  or die $usage;
die $usage if $help;

# Initialize map variables
my (
    @jobs,    @nodes,     %job,   %user,        %group, %class,
    %adapter, %arch,      %opsys, %feature,     %qos,   %account,
    %command, %partition, %node,  %reservation, %rm
  )
  = () x 17;

################################################################################
# Parse the map file
################################################################################

if ($mapFile)
{
    if (-r $mapFile)
    {
        open MAPFILE, "<$mapFile"
          or die "Unable to open map file for reading: $!\n";
        select((select(MAPFILE), local $/)[0])
          ;    # Make MAPFILE read all into a scalar
        my $mapData = <MAPFILE>;
        eval($mapData);    # Populate map variables with data from mapfile
        close MAPFILE;
    }
}
else
{
    $mapFile = "scrub.map";
}

################################################################################
# Parse the workload trace -- adding attributes to unique lists
################################################################################

if ($workloadTraceFile)
{
    open WORKLOADTRACE, "<$workloadTraceFile"
      or die "Unable to open workload trace: $!\n";
    foreach my $line (<WORKLOADTRACE>)
    {
        chomp $line;
        next if $line =~ /^VERSION/;    # Ignore version lines
        next if $line =~ /^#/;          # Ignore comments
        next unless $line;                     # Ignore empty lines
        next unless $line =~ /JOBCOMPLETE/;    # Only consider completed jobs

        # Break the job line into a attribute hash with named fields
        my %attr = ();
        (
            $attr{EventTime},         $attr{ObjectType},
            $attr{JobId},             $attr{EventType},
            $attr{NodesRequested},    $attr{TasksRequested},
            $attr{UserName},          $attr{GroupName},
            $attr{WallclockLimit},    $attr{JobState},
            $attr{Class},             $attr{SubmissionTime},
            $attr{DispatchTime},      $attr{StartTime},
            $attr{CompletionTime},    $attr{NetworkAdapter},
            $attr{Arch},              $attr{Opsys},
            $attr{MemComp},           $attr{Memory},
            $attr{DiskComp},          $attr{Disk},
            $attr{Features},          $attr{SystemQueueTime},
            $attr{TaskCount},         $attr{TasksPerNode},
            $attr{QOS},               $attr{Flags},
            $attr{Account},           $attr{Command},
            $attr{RMExt},             $attr{BypassCount},
            $attr{PSUtil},            $attr{Partition},
            $attr{DProcs},            $attr{DMem},
            $attr{DDisk},             $attr{DSwap},
            $attr{StartDate},         $attr{EndDate},
            $attr{AllocatedHostList}, $attr{RMName},
            $attr{RequiredHostList},  $attr{Reservation},
            $attr{SimData},           $attr{Constraints},
            $attr{JobMessage},        $attr{JobCost},
            $attr{History},           $attr{RES1},
            $attr{RES2},              $attr{RES3},
            $attr{RES4}
          )
          = split /\s+/, $line;

        # Add JobId to a unique list
        $job{$attr{JobId}} ||= 'UNDEF';

        # Add UserName to a unique list
        $user{$attr{UserName}} ||= 'UNDEF';

        # Add GroupName to a unique list
        $group{$attr{GroupName}} ||= 'UNDEF';

        # Add Class to a unique list
        $attr{Class} =~
          s/\[(\w+):\d\]/$1/;    # Remove square brackets and colon count
        $class{$attr{Class}} ||= 'UNDEF';

        # Add Network Adapter to a unique list
        $adapter{$attr{NetworkAdapter}} ||= 'UNDEF';

        # Add Arch to a unique list
        $arch{$attr{Arch}} ||= 'UNDEF';

        # Add Opsys to a unique list
        $opsys{$attr{Opsys}} ||= 'UNDEF';

        # Add Features to a unique list
        $attr{Features} =~ s/^\[//;    # Remove leading square bracket
        $attr{Features} =~ s/\]$//;    # Remove trailing square bracket
        my @features = split /\]\[/,
          $attr{Features};             # Split on square brackets (][)
        foreach my $feature (@features)
        {
            $feature{$feature} ||= 'UNDEF';
        }

        # Add QOS to a unique list
        my @QOSes = split /:/, $attr{QOS};  # Handle requested vs. delivered QOS
        foreach my $qos (@QOSes)
        {
            $qos{$qos} ||= 'UNDEF';
        }

        # Add Account to a unique list
        $account{$attr{Account}} ||= 'UNDEF';

        # Add Command to a unique list
        $command{$attr{Command}} ||= 'UNDEF';

        # Add Partition to a unique list
        $partition{$attr{Partition}} ||= 'UNDEF';

        # Add Required nodes to a unique list
        #$attr{RequiredHostList} =~ s/^\[//;    # Remove leading square bracket
        #$attr{RequiredHostList} =~ s/\]$//;    # Remove trailing square bracket
        #my @RequiredNodes = split /\]\[/, $attr{RequiredHostList};
        my @RequiredNodes = split /,/, $attr{RequiredHostList};
        foreach my $node (@RequiredNodes)
        {
            $node{$node} ||= 'UNDEF';
        }

       # Add Allocated nodes to a unique list
       #$attr{AllocatedHostList} =~ s/^\[//;    # Remove leading square bracket
       #$attr{AllocatedHostList} =~ s/\]$//;    # Remove trailing square bracket
       #my @AllocatedNodes = split /\]\[/, $attr{AllocatedHostList};
        my @AllocatedNodes = split /,/, $attr{AllocatedHostList};
        foreach my $node (@AllocatedNodes)
        {
            $node{$node} ||= 'UNDEF';
        }

        # Add Reservation to a unique list
        $reservation{$attr{Reservation}} ||= 'UNDEF';

        # Push job attributes into a job array
        push @jobs, \%attr;
    }
    close WORKLOADTRACE;
}

################################################################################
# Parse the resource trace -- adding attributes to unique lists
################################################################################

if ($resourceTraceFile)
{
    open RESOURCETRACE, "<$resourceTraceFile"
      or die "Unable to open resource trace: $!\n";
    foreach my $line (<RESOURCETRACE>)
    {
        chomp $line;
        next if $line =~ /^VERSION/;    # Ignore version lines
        next if $line =~ /^#/;          # Ignore comments
        next unless $line;              # Ignore empty lines

        # Break the node line into a attribute hash with named fields
        my %attr = ();
        (
            $attr{ResourceType}, $attr{EventType},      $attr{EventTime},
            $attr{NodeId},       $attr{RMName},         $attr{Swap},
            $attr{Memory},       $attr{Disk},           $attr{Procs},
            $attr{Frame},        $attr{Slot},           $attr{Space},
            $attr{Opsys},        $attr{Arch},           $attr{Features},
            $attr{Class},        $attr{NetworkAdapter}, $attr{Speed},
            $attr{RES1},         $attr{RES2},           $attr{RES3}
          )
          = split /\s+/, $line;

        # Add NodeId to a unique list
        $node{$attr{NodeId}} ||= 'UNDEF';

        # Add RMName to a unique list
        $rm{$attr{RMName}} ||= 'UNDEF';

        # Add Opsys to a unique list
        $opsys{$attr{Opsys}} ||= 'UNDEF';

        # Add Arch to a unique list
        $arch{$attr{Arch}} ||= 'UNDEF';

        # Add Features to a unique list
        $attr{Features} =~ s/^\[//;    # Remove leading square bracket
        $attr{Features} =~ s/\]$//;    # Remove trailing square bracket
        my @features = split /\]\[/,
          $attr{Features};             # Split on square brackets (][)
        foreach my $feature (@features)
        {
            $feature{$feature} ||= 'UNDEF';
        }

        # Add Class to a unique list
        $attr{Class} =~
          s/\[(\w+):\d\]/$1/;          # Remove square brackets and colon count
        $class{$attr{Class}} ||= 'UNDEF';

        # Add Network Adapters to a unique list
        $attr{NetworkAdapter} =~ s/^\[//;    # Remove leading square bracket
        $attr{NetworkAdapter} =~ s/\]$//;    # Remove trailing square bracket
        my @adapters = split /\]\[/,
          $attr{NetworkAdapter};             # Split on square brackets (][)
        foreach my $adapter (@adapters)
        {
            $adapter{$adapter} ||= 'UNDEF';
        }

        # Push node attributes into a node array
        push @nodes, \%attr;
    }
    close RESOURCETRACE;
}

################################################################################
# Genericize the unique lists
################################################################################

my $count;

# Genericize job list
$count = grep($_ eq 'UNDEF', keys %job);
foreach my $job (sort keys %job)
{
    if ($job{$job} eq 'UNDEF')
    {
        $job{$job} = 'job' . ++$count;
    }
}

# Genericize user list
$count = grep($_ eq 'UNDEF', keys %user);
foreach my $user (sort keys %user)
{
    if ($user{$user} eq 'UNDEF')
    {
        $user{$user} = 'user' . ++$count;
    }
}

# Genericize group list
$count = grep($_ eq 'UNDEF', keys %group);
foreach my $group (sort keys %group)
{
    if ($group{$group} eq 'UNDEF')
    {
        $group{$group} = 'group' . ++$count;
    }
}

# Genericize class list
$count = grep($_ eq 'UNDEF', keys %class);
foreach my $class (sort keys %class)
{
    if ($class{$class} eq 'UNDEF')
    {
        if ($class eq '[NONE]')
        {
            $class{$class} = $class;
        }
        else
        {
            $class{$class} = 'class' . ++$count;
        }
    }
}

# Genericize adapter list
$count = grep($_ eq 'UNDEF', keys %adapter);
foreach my $adapter (sort keys %adapter)
{
    if ($adapter{$adapter} eq 'UNDEF')
    {
        if ($adapter eq '[NONE]' || $adapter eq '-')
        {
            $adapter{$adapter} = $adapter;
        }
        else
        {
            $adapter{$adapter} = 'adapter' . ++$count;
        }
    }
}

# Genericize arch list
$count = grep($_ eq 'UNDEF', keys %arch);
foreach my $arch (sort keys %arch)
{
    if ($arch{$arch} eq 'UNDEF')
    {
        if ($arch eq '[NONE]' || $arch eq '-')
        {
            $arch{$arch} = $arch;
        }
        else
        {
            $arch{$arch} = 'arch' . ++$count;
        }
    }
}

# Genericize opsys list
$count = grep($_ eq 'UNDEF', keys %opsys);
foreach my $opsys (sort keys %opsys)
{
    if ($opsys{$opsys} eq 'UNDEF')
    {
        if ($opsys eq '[NONE]' || $opsys eq '-')
        {
            $opsys{$opsys} = $opsys;
        }
        else
        {
            $opsys{$opsys} = 'opsys' . ++$count;
        }
    }
}

# Genericize feature list
$count = grep($_ eq 'UNDEF', keys %feature);
foreach my $feature (sort keys %feature)
{
    if ($feature{$feature} eq 'UNDEF')
    {
        if ($feature eq 'NONE')
        {
            $feature{$feature} = $feature;
        }
        else
        {
            $feature{$feature} = 'feature' . ++$count;
        }
    }
}

# Genericize qos list
$count = grep($_ eq 'UNDEF', keys %qos);
foreach my $qos (sort keys %qos)
{
    if ($qos{$qos} eq 'UNDEF')
    {
        if ($qos eq 'NONE' || $qos eq '-')
        {
            $qos{$qos} = $qos;
        }
        else
        {
            $qos{$qos} = 'qos' . ++$count;
        }
    }
}

# Genericize account list
$count = grep($_ eq 'UNDEF', keys %account);
foreach my $account (sort keys %account)
{
    if ($account{$account} eq 'UNDEF')
    {
        if ($account eq '[NONE]' || $account eq '-')
        {
            $account{$account} = $account;
        }
        else
        {
            $account{$account} = 'account' . ++$count;
        }
    }
}

# Genericize command list
$count = grep($_ eq 'UNDEF', keys %command);
foreach my $command (sort keys %command)
{
    if ($command{$command} eq 'UNDEF')
    {
        $command{$command} = 'command' . ++$count;
    }
}

# Genericize partition list
$count = grep($_ eq 'UNDEF', keys %partition);
foreach my $partition (sort keys %partition)
{
    if ($partition{$partition} eq 'UNDEF')
    {
        if ($partition eq '[DEFAULT]' || $partition eq '-')
        {
            $partition{$partition} = $partition;
        }
        else
        {
            $partition{$partition} = 'partition' . ++$count;
        }
    }
}

# Genericize node list
$count = grep($_ eq 'UNDEF', keys %node);
foreach my $node (sort keys %node)
{
    if ($node{$node} eq 'UNDEF')
    {
        if ($node eq 'NONE' || $node eq '-')
        {
            $node{$node} = $node;
        }
        else
        {
            $node{$node} = 'node' . ++$count;
        }
    }
}

# Genericize reservation list
$count = grep($_ eq 'UNDEF', keys %reservation);
foreach my $reservation (sort keys %reservation)
{
    if ($reservation{$reservation} eq 'UNDEF')
    {
        if ($reservation eq '[NONE]' || $reservation eq '-')
        {
            $reservation{$reservation} = $reservation;
        }
        else
        {
            $reservation{$reservation} = 'reservation' . ++$count;
        }
    }
}

# Genericize rm list
$count = grep($_ eq 'UNDEF', keys %rm);
foreach my $rm (sort keys %rm)
{
    if ($rm{$rm} eq 'UNDEF')
    {
        if ($rm eq '[NONE]' || $rm eq '-')
        {
            $rm{$rm} = $rm;
        }
        else
        {
            $rm{$rm} = 'rm' . ++$count;
        }
    }
}

################################################################################
# Update the jobs array with new sanitized values
################################################################################

foreach my $job (@jobs)
{
    # Sanitize JobId
    $job->{JobId} = $job{$job->{JobId}};

    # Sanitize UserName
    $job->{UserName} = $user{$job->{UserName}};

    # Sanitize GroupName
    $job->{GroupName} = $group{$job->{GroupName}};

    # Sanitize Class
    $job->{Class} = "[$class{$job->{Class}}:1]" if $job->{Class} ne "[NONE]";

    # Sanitize Network Adapter
    $job->{NetworkAdapter} = $adapter{$job->{NetworkAdapter}};

    # Sanitize Arch
    $job->{Arch} = $arch{$job->{Arch}};

    # Sanitize Opsys
    $job->{Opsys} = $opsys{$job->{Opsys}};

    # Sanitize Features
    $job->{Features} = '['
      . join('][', map($feature{$_}, split(/\]\[/, $job->{Features}))) . ']';

    # Sanitize QOS
    $job->{QOS} = join(':', map($qos{$_}, split(/:/, $job->{QOS})));

    # Sanitize Account
    $job->{Account} = $account{$job->{Account}};

    # Sanitize Command
    $job->{Command} = $command{$job->{Command}};

    # Sanitize Partition
    $job->{Partition} = $partition{$job->{Partition}};

    # Sanitize RequiredHostList
    $job->{RequiredHostList} =
      join(',', map($node{$_}, split(/,/, $job->{RequiredHostList})));

    # Sanitize AllocatedHostList
    $job->{AllocatedHostList} =
      join(',', map($node{$_}, split(/,/, $job->{AllocatedHostList})));

    # Sanitize Reservation
    $job->{Reservation} = $reservation{$job->{Reservation}};
}

################################################################################
# Update the nodes array with new sanitized values
################################################################################

foreach my $node (@nodes)
{
    # Sanitize NodeId
    $node->{NodeId} = $node{$node->{NodeId}};

    # Sanitize RMName
    $node->{RMName} = $rm{$node->{RMName}};

    # Sanitize Opsys
    $node->{Opsys} = $opsys{$node->{Opsys}};

    # Sanitize Arch
    $node->{Arch} = $arch{$node->{Arch}};

    # Sanitize Features
    $node->{Features} = '['
      . join('][', map($feature{$_}, split(/\]\[/, $node->{Features}))) . ']';

    # Sanitize Class
    $node->{Class} = "[$class{$node->{Class}}:1]";

    # Sanitize Network Adapters
    $node->{NetworkAdapter} = '['
      . join('][', map($adapter{$_}, split(/\]\[/, $node->{NetworkAdapter})))
      . ']';
}

################################################################################
# Write out the scrubbed workload trace
################################################################################

if ($workloadTraceFile)
{
    open SCRUBBEDWORKLOADTRACE, ">${workloadTraceFile}.clean"
      or die "Unable to open scrubbed workload trace file for writing: $!\n";
    foreach my $job (@jobs)
    {
        print SCRUBBEDWORKLOADTRACE
          "$job->{EventTime} $job->{ObjectType} $job->{JobId} $job->{EventType} $job->{NodesRequested} $job->{TasksRequested} $job->{UserName} $job->{GroupName} $job->{WallclockLimit} $job->{JobState} $job->{Class} $job->{SubmissionTime} $job->{DispatchTime} $job->{StartTime} $job->{CompletionTime} $job->{NetworkAdapter} $job->{Arch} $job->{Opsys} $job->{MemComp} $job->{Memory} $job->{DiskComp} $job->{Disk} $job->{Features} $job->{SystemQueueTime} $job->{TaskCount} $job->{TasksPerNode} $job->{QOS} $job->{Flags} $job->{Account} $job->{Command} $job->{RMExt} $job->{BypassCount} $job->{PSUtil} $job->{Partition} $job->{DProcs} $job->{DMem} $job->{DDisk} $job->{DSwap} $job->{StartDate} $job->{EndDate} $job->{AllocatedHostList} $job->{RMName} $job->{RequiredHostList} $job->{Reservation} $job->{SimData} - - 0.0 - - - - -\n";
    }
    close SCRUBBEDWORKLOADTRACE;
}

################################################################################
# Write out the scrubbed resource trace
################################################################################

if ($resourceTraceFile)
{
    open SCRUBBEDRESOURCETRACE, ">${resourceTraceFile}.clean"
      or die "Unable to open scrubbed resource trace file for writing: $!\n";
    foreach my $node (@nodes)
    {
        print SCRUBBEDRESOURCETRACE
          "$node->{ResourceType} $node->{EventType} $node->{EventTime} $node->{NodeId} $node->{RMName} $node->{Swap} $node->{Memory} $node->{Disk} $node->{Procs} $node->{Frame} $node->{Slot} $node->{Space} $node->{Opsys} $node->{Arch} $node->{Features} $node->{Class} $node->{NetworkAdapter} $node->{Speed} $node->{RES1} $node->{RES2} $node->{RES3}\n";
    }
    close SCRUBBEDRESOURCETRACE;
}

################################################################################
# Write out the map file
################################################################################

open MAPFILE, ">$mapFile"
  or die "Unable to open scrub map file for writing: $!\n";
$Data::Dumper::Sortkeys = 1;
print MAPFILE Data::Dumper->Dump(
    [
        \%job,     \%user,      \%group,   \%class,       \%adapter,
        \%arch,    \%opsys,     \%feature, \%qos,         \%account,
        \%command, \%partition, \%node,    \%reservation, \%rm
    ],
    [
        'job',     'user',        'group',   'class',
        'adapter', 'arch',        'opsys',   'feature',
        'qos',     'account',     'command', 'partition',
        'node',    'reservation', 'rm'
    ]
);
close MAPFILE;

1;
__END__

=head1 NAME

=head1 SYNOPSIS

=head1 DESCRIPTION

=head2 EXPORT

=head2 EXPORT_OK

=head1 SEE ALSO

Additional information can be found at http://www.adaptivecomputing.com/

=head1 AUTHOR

Adaptive Computing, Inc., E<lt>http://www.adaptivecomputing.com/E<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2010-2016 by Adaptive Computing Enterprises, Inc. All Rights Reserved.

=cut
