-- This Oracle SQL script creates Moab tables for all Moab suites.
--   Moab HPC Suite - Basic Edition
--   Moab HPC Suite - Enterprise Edition
--   Moab Cloud Suite
--
-- The script is organized into six sections.
--   Moab "pre-defined values" table definitions
--   Moab persistent "historical" table definitions
--   Moab semi-persistent "restart" table definitions
--   Moab non-persistent "current state" table definitions
--   Moab trigger definitions for historical data insertion
--   Moab "pre-defined value" definition insertions
-- Persistence as used above refers to the persistence of table data between Moab restarts.
--====================================================================================================
-- Oracle DBA: Create a "Moab" user name/schema and assign the user name the "CONNECT", "RESOURCE",
-- and "DBA" permissions.  Log in with the Moab user name and execute the following DDL and DML
-- statements to create the Moab database for the current Moab version specified by the first DML
-- statement in this script.  Then remove the "DBA" and "RESOURCE" permissions from the "Moab" user
-- name so the Moab product can only insert, delete, update, and select rows from the Moab tables.
-- 
-- If you choose to use a user other than "Moab", update the schema owner on each of the synonyms
-- in this file.
--
-- This script is for your convenience.  If you have particular business use cases that could benefit
-- from additional indexes for performance purposes, add index creation DDL statements below in the
-- appropriate places.  You should only add indexes to tables that hold PERSISTENT data between Moab
-- restarts; that is, those tables that contain historical information.  Those tables that represent
-- Moab's "current state" should have only a primary key definition for the sake of speed since after
-- each scheduling cycle Moab may delete any row within the "current state" tables and/or insert a new
-- row or version of a row just deleted, and additional indices would slow down this process.



--======================================================-
-- Moab "pre-defined values" tables
--
-- These tables persist between Moab restarts and can be
-- altered by Moab upgrades to reflect the current Moab
-- version's schema.
--=======================================================


CREATE TABLE Moab (
  Version FLOAT(6)
);

CREATE PUBLIC SYNONYM MOAB FOR MOAB.MOAB;


-- Moab pre-defined object type IDs and names
--   Moab definition information
--   PERSISTENT between Moab restarts
--   Possible INSERTs during Moab upgrades
CREATE TABLE ObjectType (
  ID                            NUMBER(10,0),
  Name                          VARCHAR2(64)        NOT NULL,
  CONSTRAINT ObjectType_PK      PRIMARY KEY (ID)
);

CREATE PUBLIC SYNONYM OBJECTTYPE FOR MOAB.OBJECTTYPE;

-- ObjectType indexes
CREATE INDEX ObjectType_Name_IX  on ObjectType(Name);

-- Insert pre-defined object types with their IDs and names
INSERT INTO ObjectType (Name,ID) VALUES ('Acct',         1);
INSERT INTO ObjectType (Name,ID) VALUES ('AM',           2);
INSERT INTO ObjectType (Name,ID) VALUES ('Class',        3);
INSERT INTO ObjectType (Name,ID) VALUES ('Cluster',      4);
INSERT INTO ObjectType (Name,ID) VALUES ('Group',        5);
INSERT INTO ObjectType (Name,ID) VALUES ('Job',          6);
INSERT INTO ObjectType (Name,ID) VALUES ('Node',         7);
INSERT INTO ObjectType (Name,ID) VALUES ('Par',          8);
INSERT INTO ObjectType (Name,ID) VALUES ('QOS',          9);
INSERT INTO ObjectType (Name,ID) VALUES ('Rsv',         13);
INSERT INTO ObjectType (Name,ID) VALUES ('RM',          14);
INSERT INTO ObjectType (Name,ID) VALUES ('Sched',       15);
INSERT INTO ObjectType (Name,ID) VALUES ('SRsv',        16);
INSERT INTO ObjectType (Name,ID) VALUES ('Sys',         17);
INSERT INTO ObjectType (Name,ID) VALUES ('TNode',       18);
INSERT INTO ObjectType (Name,ID) VALUES ('Trig',        19);
INSERT INTO ObjectType (Name,ID) VALUES ('User',        20);
INSERT INTO ObjectType (Name,ID) VALUES ('ID',          21);
INSERT INTO ObjectType (Name,ID) VALUES ('CJob',        23);
INSERT INTO ObjectType (Name,ID) VALUES ('GRes',        30);
INSERT INTO ObjectType (Name,ID) VALUES ('Gmetric',     31);
INSERT INTO ObjectType (Name,ID) VALUES ('Stats',       39);
INSERT INTO ObjectType (Name,ID) VALUES ('TJob',        42);
INSERT INTO ObjectType (Name,ID) VALUES ('Paction',     43);
INSERT INTO ObjectType (Name,ID) VALUES ('VM',          45);
INSERT INTO ObjectType (Name,ID) VALUES ('VPC',         47);
INSERT INTO ObjectType (Name,ID) VALUES ('JGroup',      48);
COMMIT;


-- Moab pre-defined event type IDs and names
--   Moab definition information
--   PERSISTENT between Moab restarts
--   Possible INSERTs during Moab upgrades
CREATE TABLE EventType (
  ID                            NUMBER(10,0),
  Name                          VARCHAR2(64)        NOT NULL,
  CONSTRAINT EventType_PK       PRIMARY KEY (ID)
);

CREATE PUBLIC SYNONYM EVENTTYPE FOR MOAB.EVENTTYPE;

-- EventType indexes
CREATE INDEX EventType_Name_IX  on EventType(Name);

-- Insert pre-defined event types with their IDs and names --
INSERT INTO EventType (Name,ID) VALUES ('[NONE]',                 0);
INSERT INTO EventType (Name,ID) VALUES ('GEVENT',                 1);
INSERT INTO EventType (Name,ID) VALUES ('JOBCANCEL',              2);
INSERT INTO EventType (Name,ID) VALUES ('JOBEND',                 3);
INSERT INTO EventType (Name,ID) VALUES ('JOBFAILURE',             4);
INSERT INTO EventType (Name,ID) VALUES ('JOBHOLD',                5);
INSERT INTO EventType (Name,ID) VALUES ('JOBMIGRATE',             6);
INSERT INTO EventType (Name,ID) VALUES ('JOBMODIFY',              7);
INSERT INTO EventType (Name,ID) VALUES ('JOBPREEMPT',             8);
INSERT INTO EventType (Name,ID) VALUES ('JOBREJECT',              9);
INSERT INTO EventType (Name,ID) VALUES ('JOBRESUME',             10);
INSERT INTO EventType (Name,ID) VALUES ('JOBSTART',              11);
INSERT INTO EventType (Name,ID) VALUES ('JOBSUBMIT',             12);
INSERT INTO EventType (Name,ID) VALUES ('JOBUPDATE',             13);
INSERT INTO EventType (Name,ID) VALUES ('JOBVARSET',             14);
INSERT INTO EventType (Name,ID) VALUES ('JOBVARUNSET',           15);
INSERT INTO EventType (Name,ID) VALUES ('NODEDOWN',              16);
INSERT INTO EventType (Name,ID) VALUES ('NODEFAILURE',           17);
INSERT INTO EventType (Name,ID) VALUES ('NODEUP',                18);
INSERT INTO EventType (Name,ID) VALUES ('NODEUPDATE',            19);
INSERT INTO EventType (Name,ID) VALUES ('NOTE',                  20);
INSERT INTO EventType (Name,ID) VALUES ('QOSVIOLATION',          21);
INSERT INTO EventType (Name,ID) VALUES ('RMDOWN',                22);
INSERT INTO EventType (Name,ID) VALUES ('RMUP',                  23);
INSERT INTO EventType (Name,ID) VALUES ('RSVCANCEL',             24);
INSERT INTO EventType (Name,ID) VALUES ('RSVCREATE',             25);
INSERT INTO EventType (Name,ID) VALUES ('RSVEND',                26);
INSERT INTO EventType (Name,ID) VALUES ('RSVMODIFY',             27);
INSERT INTO EventType (Name,ID) VALUES ('RSVSTART',              28);
INSERT INTO EventType (Name,ID) VALUES ('SCHEDCOMMAND',          29);
INSERT INTO EventType (Name,ID) VALUES ('SCHEDFAILURE',          30);
INSERT INTO EventType (Name,ID) VALUES ('SCHEDMODIFY',           31);
INSERT INTO EventType (Name,ID) VALUES ('SCHEDPAUSE',            32);
INSERT INTO EventType (Name,ID) VALUES ('SCHEDRECYCLE',          33);
INSERT INTO EventType (Name,ID) VALUES ('SCHEDRESUME',           34);
INSERT INTO EventType (Name,ID) VALUES ('SCHEDSTART',            35);
INSERT INTO EventType (Name,ID) VALUES ('SCHEDSTOP',             36);
INSERT INTO EventType (Name,ID) VALUES ('TRIGCREATE',            37);
INSERT INTO EventType (Name,ID) VALUES ('TRIGEND',               38);
INSERT INTO EventType (Name,ID) VALUES ('TRIGFAILURE',           39);
INSERT INTO EventType (Name,ID) VALUES ('TRIGSTART',             40);
INSERT INTO EventType (Name,ID) VALUES ('TRIGTHRESHOLD',         41);
INSERT INTO EventType (Name,ID) VALUES ('VMCREATE',              42);
INSERT INTO EventType (Name,ID) VALUES ('VMDESTROY',             43);
INSERT INTO EventType (Name,ID) VALUES ('VMMIGRATE',             44);
INSERT INTO EventType (Name,ID) VALUES ('VMPOWERON',             45);
INSERT INTO EventType (Name,ID) VALUES ('VMPOWEROFF',            46);
INSERT INTO EventType (Name,ID) VALUES ('NODEMODIFY',            47);
INSERT INTO EventType (Name,ID) VALUES ('NODEPOWEROFF',          48);
INSERT INTO EventType (Name,ID) VALUES ('NODEPOWERON',           49);
INSERT INTO EventType (Name,ID) VALUES ('NODEPROVISION',         50);
INSERT INTO EventType (Name,ID) VALUES ('ALLSCHEDCOMMAND',       51);
INSERT INTO EventType (Name,ID) VALUES ('AMDELETE',              52);
INSERT INTO EventType (Name,ID) VALUES ('AMEND',                 53);
INSERT INTO EventType (Name,ID) VALUES ('AMQUOTE',               54);
INSERT INTO EventType (Name,ID) VALUES ('AMSTART',               55);
INSERT INTO EventType (Name,ID) VALUES ('RMPOLLEND',             56);
INSERT INTO EventType (Name,ID) VALUES ('RMPOLLSTART',           57);
INSERT INTO EventType (Name,ID) VALUES ('SCHEDCYCLEEND',         58);
INSERT INTO EventType (Name,ID) VALUES ('SCHEDCYCLESTART',       59);
INSERT INTO EventType (Name,ID) VALUES ('JOBCHECKPOINT',         60);
INSERT INTO EventType (Name,ID) VALUES ('AMCREATE',              61);
INSERT INTO EventType (Name,ID) VALUES ('AMUPDATE',              62);
INSERT INTO EventType (Name,ID) VALUES ('AMPAUSE',               63);
INSERT INTO EventType (Name,ID) VALUES ('AMRESUME',              64);
INSERT INTO EventType (Name,ID) VALUES ('AMMODIFY',              65);
INSERT INTO EventType (Name,ID) VALUES ('AMCONTINUE',            66);
INSERT INTO EventType (Name,ID) VALUES ('NODEREMOVE',            67);
INSERT INTO EventType (Name,ID) VALUES ('NODEADD',               68);
COMMIT;



--======================================================-
-- Moab "historical" tables
--
-- These tables persist between Moab restarts and can be
-- altered by Moab upgrades to reflect the current Moab
-- version's schema.
--
-- The DBA may want to add indexes and even foreign keys
-- to these tables to address query speed for enterprise-
-- specific use cases.  Be aware additional indexes slow
-- insert operations.
--=======================================================


-- Moab general statistics by object type, object name, and epoch time
--   Moab historical information
--   PERSISTENT between Moab restarts
--   Possible ALTERs during Moab upgrades
CREATE TABLE GeneralStats (
  ObjectType                    NUMBER(10,0),
  ObjectName                    VARCHAR2(64),
  Time                          NUMBER(19,0),
  StatDuration                  NUMBER(10,0),
  IterationCount                NUMBER(10,0),
  IntervalStatCount             NUMBER(10,0),
  IntervalStatDuration          NUMBER(10,0),
  TotalJobCount                 NUMBER(10,0),
  JobsSubmitted                 NUMBER(10,0),
  JobsActive                    NUMBER(10,0),
  JobsEligible                  NUMBER(10,0),
  JobsIdle                      NUMBER(10,0),
  JobsCompletedByQOS            NUMBER(10,0),
  JobsQueued                    NUMBER(10,0),
  JobsPreempted                 NUMBER(10,0),
  TotalJobAccuracy              BINARY_DOUBLE,
  TotalNodeLoad                 BINARY_DOUBLE,
  TotalNodeMemory               BINARY_DOUBLE,
  TotalNodeCount                NUMBER(10,0),
  TotalAllocatedNodeCount       NUMBER(10,0),
  TotalActiveNodeCount          NUMBER(10,0),
  TotalActiveProcCount          NUMBER(10,0),
  TotalConfiguredProcCount      NUMBER(10,0),
  TotalAvailableCapacity        NUMBER(10,0),
  ProcSecondsDedicated          BINARY_DOUBLE,
  ProcSecondsExecuted           BINARY_DOUBLE,
  ProcSecondsRequested          BINARY_DOUBLE,
  ProcSecondsUtilized           BINARY_DOUBLE,
  ProcHoursPreempted            BINARY_DOUBLE,
  ProcHoursSubmitted            BINARY_DOUBLE,
  ProcHoursQueued               BINARY_DOUBLE,
  ProcHoursAvailable            BINARY_DOUBLE,
  ProcHoursDedicated            BINARY_DOUBLE,
  ProcHoursSuccessful           BINARY_DOUBLE,
  ProcHoursUtilized             BINARY_DOUBLE,
  DiskSecondsAvailable          BINARY_DOUBLE,
  DiskSecondsDedicated          BINARY_DOUBLE,
  DiskSecondsUtilized           BINARY_DOUBLE,
  DiskSecondsUtilizedInstant    BINARY_DOUBLE,
  MemSecondsAvailable           BINARY_DOUBLE,
  MemSecondsDedicated           BINARY_DOUBLE,
  MemSecondsUtilized            BINARY_DOUBLE,
  AvgBypass                     NUMBER(10,0),
  MaxBypass                     NUMBER(10,0),
  TotalBypass                   NUMBER(10,0),
  AvgXFactor                    BINARY_DOUBLE,
  MaxXFactor                    BINARY_DOUBLE,
  AvgQueueTime                  BINARY_DOUBLE,
  MaxQueueTime                  NUMBER(10,0),
  TotalQueueTime                NUMBER(10,0),
  TotalRequestedWalltime        BINARY_DOUBLE,
  TotalExecutionWalltime        BINARY_DOUBLE,
  MinEfficiency                 BINARY_DOUBLE,
  MinEfficiencyDuration         BINARY_DOUBLE,
  StartedJobCount               NUMBER(10,0),
  StartedJobsQueueTime          NUMBER(10,0),
  StartedJobsProcCount          NUMBER(10,0),
  StartedJobsXFactor            BINARY_DOUBLE,
  TotalNodeWeightedXFactor      BINARY_DOUBLE,
  TotalNodeWeightedJobAccuracy  BINARY_DOUBLE,
  QOSCredits                    BINARY_DOUBLE,
  AppBacklog                    BINARY_DOUBLE,
  AppLoad                       BINARY_DOUBLE,
  AppResponseTime               BINARY_DOUBLE,
  AppThroughput                 BINARY_DOUBLE,
  TotalXFactor                  BINARY_DOUBLE,
  Other                         VARCHAR2(256),
  CONSTRAINT GeneralStats_PK    PRIMARY KEY (ObjectType, ObjectName, Time),
  CONSTRAINT GeneralStats_OT_FK FOREIGN KEY (ObjectType)
                                  REFERENCES ObjectType (ID)
);

CREATE PUBLIC SYNONYM GENERALSTATS FOR MOAB.GENERALSTATS;


-- Moab generic metrics by object type, object name, generic metric name, and epoch time
--   Moab historical information
--   PERSISTENT between Moab restarts
--   Possible ALTERs during Moab upgrades
CREATE TABLE GenericMetrics (
  ObjectType                    NUMBER(10,0),
  ObjectName                    VARCHAR2(64),
  GMetricName                   VARCHAR2(64),
  Time                          NUMBER(19,0),
  GMetricValue                  BINARY_DOUBLE,
  CONSTRAINT GenericMetrics_PK          PRIMARY KEY (ObjectType, ObjectName, GMetricName, Time),
  CONSTRAINT GenericMetrics_OTONTM_FK   FOREIGN KEY (ObjectType, ObjectName, Time) 
                                          REFERENCES GeneralStats(ObjectType, ObjectName, Time)
);

CREATE PUBLIC SYNONYM GENERICMETRICS FOR MOAB.GENERICMETRICS;


-- Moab node statistics by node ID and epoch time
--   Moab historical information
--   PERSISTENT between Moab restarts
--   Possible ALTERs during Moab upgrades
CREATE TABLE NodeStats (
  NodeID                        VARCHAR2(64),
  Time                          NUMBER(19,0),
  NodeState                     NUMBER(10,0),
  IterationCount                NUMBER(10,0),
  IntervalStatCount             NUMBER(10,0),
  IntervalStatDuration          NUMBER(10,0),
  InternalNodeCategory          VARCHAR2(64),
  ExternalNodeCategory          VARCHAR2(64),
  TotalQueueTime                NUMBER(19,0),
  TotalXFactor                  BINARY_DOUBLE,
  ProcsAvailable                NUMBER(10,0),
  ProcsConfigured               NUMBER(10,0),
  ProcsDedicated                NUMBER(10,0),
  MaxMemoryUsed                 NUMBER(10,0),
  MinMemoryUsed                 NUMBER(10,0),
  TotalMemoryUsed               NUMBER(10,0),
  StartJC                       NUMBER(10,0),
  FailureJC                     NUMBER(10,0),
  AvgCPULoad                    BINARY_DOUBLE,
  MaxCPULoad                    BINARY_DOUBLE,
  MinCPULoad                    BINARY_DOUBLE,
  Category                      NUMBER(10,0),
  RCategory                     NUMBER(10,0),
  Other                         VARCHAR2(256),
  CPUClock                      NUMBER(10,0),
  CPUMaxClock                   NUMBER(10,0),
  CPUMinClock                   NUMBER(10,0),
  CONSTRAINT NodeStats_PK       PRIMARY KEY (NodeID, Time)
);

CREATE PUBLIC SYNONYM NODESTATS FOR MOAB.NODESTATS;


-- Moab node generic resource statistics by node ID, generic resource name, and epoch time
--   Moab historical information
--   PERSISTENT between Moab restarts
--   Possible ALTERs during Moab upgrades
CREATE TABLE NodeStatsGenericResources (
  NodeID                        VARCHAR2(64),
  GResName                      VARCHAR2(64),
  Time                          NUMBER(19,0),
  GenericResourcesAvailable     NUMBER(10,0),
  GenericResourcesConfigured    NUMBER(10,0),
  GenericResourcesDedicated     NUMBER(10,0),
  CONSTRAINT NodeStatsGRES_PK           PRIMARY KEY (NodeID, GResName, Time),
  CONSTRAINT NodeStatsGRES_IDTM_FK      FOREIGN KEY (NodeID, Time)
                                          REFERENCES NodeStats(NodeID, Time)
);

CREATE PUBLIC SYNONYM NODESTATSGENERICRESOURCES FOR MOAB.NODESTATSGENERICRESOURCES;


-- Moab events by event ID
--   Moab historical information
--   PERSISTENT between Moab restarts
--   Possible ALTERs during Moab upgrades
CREATE TABLE Events (
  ID                            NUMBER(10,0),
  ObjectType                    NUMBER(10,0),
  EventType                     NUMBER(10,0),
  EventTime                     NUMBER(19,0),
  ObjectName                    VARCHAR2(64),
  Name                          VARCHAR2(64),
  Description                   VARCHAR2(4000),
  CONSTRAINT Events_PK          PRIMARY KEY (ID)
);

CREATE PUBLIC SYNONYM EVENTS FOR MOAB.EVENTS;

-- Events indexes (used to speed up queries)
CREATE INDEX Events_Name_IX     on Events (Name);
CREATE INDEX Events_Time_IX     on Events (EventTime);
CREATE INDEX Events_EType_IX    on Events (EventType);
CREATE INDEX Events_OType_IX    on Events (ObjectType);
CREATE INDEX Events_OName_IX    on Events (ObjectName);


-- Moab job history by job ID and submission time, which allows the job ID to
-- wrap around and keep multiple historical jobs with the same job ID separate
--   Moab historical information
--   PERSISTENT between Moab restarts
--   Possible ALTERs during Moab upgrades
CREATE TABLE JobHistory (
  ID                            VARCHAR2(64),
  SourceRMJobID                 VARCHAR2(64),
  DestinationRMJobID            VARCHAR2(64),
  GridJobID                     VARCHAR2(64),
  AName                         VARCHAR2(64),
  UserName                      VARCHAR2(64),
  Account                       VARCHAR2(64),
  Class                         VARCHAR2(64),
  QOS                           VARCHAR2(64),
  OwnerGroup                    VARCHAR2(64),
  JobGroup                      VARCHAR2(64),
  State                         VARCHAR2(64),
  EState                        VARCHAR2(64),
  SubState                      VARCHAR2(64),
  UserPriority                  NUMBER(19,0),
  SystemPriority                NUMBER(19,0),
  CurrentStartPriority          NUMBER(19,0),
  RunPriority                   NUMBER(19,0),
  PerPartitionPriority          VARCHAR2(4000),
  SubmitTime                    NUMBER(19,0), 
  QueueTime                     NUMBER(19,0),
  StartTime                     NUMBER(19,0),
  CompletionTime                NUMBER(19,0),
  CompletionCode                NUMBER(10,0),
  UsedWalltime                  NUMBER(19,0),
  RequestedMinWalltime          NUMBER(19,0),
  RequestedMaxWalltime          NUMBER(19,0),
  CPULimit                      NUMBER(19,0),
  SuspendTime                   NUMBER(19,0),
  HoldTime                      NUMBER(19,0),
  ProcessorCount                NUMBER(10,0),
  RequestedNodes                NUMBER(10,0),
  ActivePartition               VARCHAR2(64),
  SpecPAL                       VARCHAR2(64),
  DestinationRM                 VARCHAR2(64),
  SourceRM                      VARCHAR2(64),
  Flags                         VARCHAR2(4000),
  MinPreemptTime                NUMBER(19,0),
  Dependencies                  VARCHAR2(4000),
  RequestedHostList             CLOB,
  ExcludedHostList              CLOB,
  MasterHostName                VARCHAR2(64),
  GenericAttributes             VARCHAR2(4000),
  Holds                         VARCHAR2(4000),
  Cost                          BINARY_DOUBLE,
  Description                   VARCHAR2(4000),
  Messages                      VARCHAR2(4000),
  NotificationAddress           VARCHAR2(4000),
  StartCount                    NUMBER(19,0),
  BypassCount                   NUMBER(19,0),
  CommandFile                   VARCHAR2(4000),
  Arguments                     VARCHAR2(4000),
  RMSubmitLanguage              VARCHAR2(4000),
  StdIn                         VARCHAR2(4000),
  StdOut                        VARCHAR2(4000),
  StdErr                        VARCHAR2(4000),
  RMOutput                      VARCHAR2(4000),
  RMError                       VARCHAR2(4000),
  InitialWorkingDirectory       VARCHAR2(4000),
  UMask                         NUMBER(19,0),
  RsvStartTime                  NUMBER(19,0),
  BlockReason                   VARCHAR2(4000),
  BlockMsg                      VARCHAR2(4000),
  PSDedicated                   BINARY_DOUBLE,
  PSUtilized                    BINARY_DOUBLE,
  CONSTRAINT JobHistory_PK      PRIMARY KEY (ID, SubmitTime)
);

CREATE PUBLIC SYNONYM JOBHISTORY FOR MOAB.JOBHISTORY;


--=======================================================
-- Moab "restart" tables
--
-- Should only have primary key index for speed!
--=======================================================


-- Moab object checkpoint by object type and object name
--   Moab "current state/restart" information
--   All rows PERSISTENT between Moab restarts
--   Individual rows NOT PERSISTENT during Moab operation
--   Possible ALTERs during Moab upgrades
CREATE TABLE  mcheckpoint (
  object_type                   VARCHAR2(64)        NOT NULL,
  object_name                   VARCHAR2(64)        NOT NULL,
  timestamp                     NUMBER(19,0)        DEFAULT '0',
  data                          CLOB                NOT NULL,
  CONSTRAINT MCheckPoint_PK     PRIMARY KEY (object_type, object_name)
);

CREATE PUBLIC SYNONYM MCHECKPOINT FOR MOAB.MCHECKPOINT;


--======================================================
-- Moab "current state" tables
--
-- Should only have primary key index for speed!
--======================================================

-- Moab nodes by node ID
--   Moab "current state" information
--   NOT PERSISTENT between Moab restarts
--   Possible ALTERs during Moab upgrades
CREATE TABLE Nodes (
  ID                            VARCHAR2(64),
  State                         VARCHAR2(64),
  OperatingSystem               VARCHAR2(64),
  ConfiguredProcessors          NUMBER(19,0),
  AvailableProcessors           NUMBER(19,0),
  ConfiguredMemory              NUMBER(19,0),
  AvailableMemory               NUMBER(19,0),
  Architecture                  VARCHAR2(64),
  AvailGres                     VARCHAR2(1024),
  ConfigGres                    VARCHAR2(1024),
  AvailClasses                  VARCHAR2(1024),
  ConfigClasses                 VARCHAR2(1024),
  ChargeRate                    BINARY_DOUBLE,
  DynamicPriority               BINARY_DOUBLE,
  EnableProfiling               NUMBER(19,0),
  Features                      VARCHAR2(1024),
  GMetric                       VARCHAR2(1024),
  HopCount                      NUMBER(19,0),
  HypervisorType                VARCHAR2(64),
  IsDeleted                     NUMBER(19,0),
  IsDynamic                     NUMBER(19,0),
  JobList                       VARCHAR2(1024),
  LastUpdateTime                NUMBER(19,0),
  LoadAvg                       BINARY_DOUBLE,
  MaxLoad                       BINARY_DOUBLE,
  MaxJob                        NUMBER(19,0),
  MaxJobPerUser                 NUMBER(19,0),
  MaxProc                       NUMBER(19,0),
  MaxProcPerUser                NUMBER(19,0),
  OldMessages                   VARCHAR2(1024),
  NetworkAddress                VARCHAR2(64),
  NodeSubstate                  VARCHAR2(64),
  Operations                    VARCHAR2(1024),
  OSList                        VARCHAR2(1024),
  Owner                         VARCHAR2(64),
  ResOvercommitFactor           VARCHAR2(64),
  PartitionName                 VARCHAR2(64),
  PowerIsEnabled                NUMBER(19,0),
  PowerPolicy                   VARCHAR2(64),
  PowerSelectState              VARCHAR2(64),
  PowerState                    VARCHAR2(64),
  Priority                      NUMBER(19,0),
  PriorityFunction              VARCHAR2(64),
  ProcessorSpeed                NUMBER(19,0),
  ProvisioningData              VARCHAR2(1024),
  AvailableDisk                 NUMBER(19,0),  
  AvailableSwap                 NUMBER(19,0),  
  ConfiguredDisk                NUMBER(19,0),
  ConfiguredSwap                NUMBER(19,0),
  ReservationCount              NUMBER(19,0),
  ReservationList               VARCHAR2(1024),
  ResourceManagerList           VARCHAR2(1024),
  SlotsUsed                     NUMBER(19,0),
  Speed                         BINARY_DOUBLE,
  SpeedWeight                   BINARY_DOUBLE,
  TotalNodeActiveTime           NUMBER(19,0),
  LastModifyTime                NUMBER(19,0),
  TotalTimeTracked              NUMBER(19,0),
  TotalNodeUpTime               NUMBER(19,0),
  TaskCount                     NUMBER(19,0),
  VMOSList                      VARCHAR2(1024),          
  CPUClock                      VARCHAR2(64),          
  CONSTRAINT Nodes_PK           PRIMARY KEY (ID)
);

CREATE PUBLIC SYNONYM NODES FOR MOAB.NODES;


-- Moab jobs by job ID
--   Moab "current state" information
--   NOT PERSISTENT between Moab restarts
--   Possible ALTERs during Moab upgrades
CREATE TABLE Jobs (
  ID                            VARCHAR2(64),
  SourceRMJobID                 VARCHAR2(64),
  DestinationRMJobID            VARCHAR2(64),
  GridJobID                     VARCHAR2(64),
  AName                         VARCHAR2(64),
  UserName                      VARCHAR2(64),
  Account                       VARCHAR2(64),
  Class                         VARCHAR2(64),
  QOS                           VARCHAR2(64),
  OwnerGroup                    VARCHAR2(64),
  JobGroup                      VARCHAR2(64),
  State                         VARCHAR2(64),
  EState                        VARCHAR2(64),
  SubState                      VARCHAR2(64),
  UserPriority                  NUMBER(19,0),
  SystemPriority                NUMBER(19,0),
  CurrentStartPriority          NUMBER(19,0),
  RunPriority                   NUMBER(19,0),
  PerPartitionPriority          VARCHAR2(4000),
  SubmitTime                    NUMBER(19,0), 
  QueueTime                     NUMBER(19,0),
  StartTime                     NUMBER(19,0),
  CompletionTime                NUMBER(19,0),
  CompletionCode                NUMBER(10,0),
  UsedWalltime                  NUMBER(19,0),
  RequestedMinWalltime          NUMBER(19,0),
  RequestedMaxWalltime          NUMBER(19,0),
  CPULimit                      NUMBER(19,0),
  SuspendTime                   NUMBER(19,0),
  HoldTime                      NUMBER(19,0),
  ProcessorCount                NUMBER(10,0),
  RequestedNodes                NUMBER(10,0),
  ActivePartition               VARCHAR2(64),
  SpecPAL                       VARCHAR2(64),
  DestinationRM                 VARCHAR2(64),
  SourceRM                      VARCHAR2(64),
  Flags                         VARCHAR2(4000),
  MinPreemptTime                NUMBER(19,0),
  Dependencies                  VARCHAR2(4000),
  RequestedHostList             CLOB,
  ExcludedHostList              CLOB,
  MasterHostName                VARCHAR2(64),
  GenericAttributes             VARCHAR2(4000),
  Holds                         VARCHAR2(4000),
  Cost                          BINARY_DOUBLE,
  Description                   VARCHAR2(4000),
  Messages                      VARCHAR2(4000),
  NotificationAddress           VARCHAR2(4000),
  StartCount                    NUMBER(19,0),
  BypassCount                   NUMBER(19,0),
  CommandFile                   VARCHAR2(4000),
  Arguments                     VARCHAR2(4000),
  RMSubmitLanguage              VARCHAR2(4000),
  StdIn                         VARCHAR2(4000),
  StdOut                        VARCHAR2(4000),
  StdErr                        VARCHAR2(4000),
  RMOutput                      VARCHAR2(4000),
  RMError                       VARCHAR2(4000),
  InitialWorkingDirectory       VARCHAR2(4000),
  UMask                         NUMBER(19,0),
  RsvStartTime                  NUMBER(19,0),
  BlockReason                   VARCHAR2(4000),
  BlockMsg                      VARCHAR2(4000),
  PSDedicated                   BINARY_DOUBLE,
  PSUtilized                    BINARY_DOUBLE,
  Internal_ID                   NUMBER(10) NOT NULL,
  CONSTRAINT Jobs_PK            PRIMARY KEY (Internal_ID)
);

CREATE PUBLIC SYNONYM JOBS FOR MOAB.JOBS;

CREATE SEQUENCE jobs_pk_seq
  MINVALUE 1
  START WITH 1
  INCREMENT BY 1
  CACHE 100;

CREATE OR REPLACE TRIGGER jobs_pk_incr
BEFORE INSERT ON Jobs
FOR EACH ROW
BEGIN
  SELECT jobs_pk_seq.NEXTVAL
  INTO :new.Internal_ID
  FROM dual;
END;
/


-- Moab job resource requests by job ID and relative index
--   Moab "current state" information
--   NOT PERSISTENT between Moab restarts
--   Possible ALTERs during Moab upgrades
CREATE TABLE Requests (
  JobID                         VARCHAR2(64),
  RIndex                        NUMBER(19,0),
  AllocNodeList                 CLOB,
  AllocPartition                VARCHAR2(64),
  PartitionIndex                NUMBER(19,0),
  NodeAccessPolicy              VARCHAR2(64),
  PreferredFeatures             VARCHAR2(4000),
  RequestedApp                  VARCHAR2(64),
  RequestedArch                 VARCHAR2(64),
  ReqOS                         VARCHAR2(64),
  ReqNodeSet                    VARCHAR2(64),
  ReqPartition                  VARCHAR2(64),
  MinNodeCount                  NUMBER(19,0),
  MinTaskCount                  NUMBER(19,0),
  TaskCount                     NUMBER(19,0),
  TasksPerNode                  NUMBER(19,0),
  DiskPerTask                   NUMBER(19,0),
  MemPerTask                    NUMBER(19,0),
  ProcsPerTask                  NUMBER(19,0),
  SwapPerTask                   NUMBER(19,0),
  NodeDisk                      VARCHAR2(64),
  NodeFeatures                  VARCHAR2(4000),
  NodeMemory                    VARCHAR2(64),
  NodeSwap                      VARCHAR2(64),
  NodeProcs                     VARCHAR2(64),
  GenericResources              VARCHAR2(4000),
  ConfiguredGenericResources    VARCHAR2(4000),
  CPUClock                      VARCHAR2(64),
  CONSTRAINT Requests_PK        PRIMARY KEY (JobID, RIndex)
);

CREATE PUBLIC SYNONYM REQUESTS FOR MOAB.REQUESTS;


-- Moab reservations by name
--   Moab "current state" information
--   NOT PERSISTENT between Moab restarts
--   Possible ALTERs during Moab upgrades
CREATE TABLE Reservations (
  Name                          VARCHAR2(64),
  ACL                           VARCHAR2(4000),
  AAccount                      VARCHAR2(64),
  AGroup                        VARCHAR2(64),
  AUser                         VARCHAR2(64),
  AQOS                          VARCHAR2(64),
  AllocNodeCount                NUMBER(19,0),
  AllocNodeList                 VARCHAR2(1024),
  AllocProcCount                NUMBER(19,0),
  AllocTaskCount                NUMBER(19,0),
  CL                            VARCHAR2(4000),
  ReservationComment            VARCHAR2(4000),
  Cost                          BINARY_DOUBLE,
  CTime                         NUMBER(19,0),
  Duration                      NUMBER(19,0),
  EndTime                       NUMBER(19,0),
  ExcludeRsv                    VARCHAR2(4000),
  ExpireTime                    NUMBER(19,0),
  Flags                         VARCHAR2(4000),
  GlobalID                      VARCHAR2(4000),
  HostExp                       VARCHAR2(4000),
  History                       VARCHAR2(4000),
  Label                         VARCHAR2(4000),
  LastChargeTime                NUMBER(19,0),
  LogLevel                      VARCHAR2(64),
  Messages                      VARCHAR2(4000),
  Owner                         VARCHAR2(64),
  PartitionName                 VARCHAR2(64),
  Profile                       VARCHAR2(64),
  ReqArch                       VARCHAR2(64),
  ReqFeatureList                VARCHAR2(4000),
  ReqMemory                     VARCHAR2(64),
  ReqNodeCount                  VARCHAR2(64),
  ReqNodeList                   CLOB,
  ReqOS                         VARCHAR2(64),
  ReqTaskCount                  VARCHAR2(64),
  ReqTPN                        VARCHAR2(64),
  Resources                     VARCHAR2(64),
  RsvAccessList                 VARCHAR2(4000),
  RsvGroup                      VARCHAR2(64),
  RsvParent                     VARCHAR2(64),
  SID                           VARCHAR2(64),
  StartTime                     NUMBER(19,0),
  StatCBPS                      NUMBER(19,0),
  StatCRPS                      NUMBER(19,0),
  StatTBPS                      NUMBER(19,0),
  StatTRPS                      NUMBER(19,0),
  SubType                       VARCHAR2(64),
  TTrigger                      VARCHAR2(64),
  TType                         VARCHAR2(64),
  Variables                     VARCHAR2(1024),
  VMList                        VARCHAR2(4000),
  CONSTRAINT Reservations_PK    PRIMARY KEY (NAME)
);

CREATE PUBLIC SYNONYM RESERVATIONS FOR MOAB.RESERVATIONS;


-- Moab triggers by trigger ID and name
--   Moab "current state" information
--   NOT PERSISTENT between Moab restarts
--   Possible ALTERs during Moab upgrades
CREATE TABLE Triggers (
  TrigID                        VARCHAR2(64),
  Name                          VARCHAR2(64),
  ActionData                    VARCHAR2(1024),
  ActionType                    VARCHAR2(64),
  BlockTime                     NUMBER(19,0),
  Description                   VARCHAR2(1024),
  Disabled                      NUMBER(19,0),
  EBuf                          VARCHAR2(1024),
  EventTime                     NUMBER(19,0),
  EventType                     VARCHAR2(64),
  FailOffset                    VARCHAR2(64),
  FailureDetected               NUMBER(19,0),
  Flags                         VARCHAR2(1024),
  IsComplete                    NUMBER(19,0),
  IsInterval                    NUMBER(19,0),
  LaunchTime                    NUMBER(19,0),
  Message                       VARCHAR2(1024),
  MultiFire                     VARCHAR2(64),
  ObjectID                      VARCHAR2(64),
  ObjectType                    VARCHAR2(64),
  OBuf                          VARCHAR2(1024),
  TriggerOffset                 VARCHAR2(64),
  Period                        VARCHAR2(64),
  PID                           NUMBER(19,0),
  RearmTime                     NUMBER(19,0),
  Requires                      VARCHAR2(1024),
  Sets                          VARCHAR2(1024),
  State                         VARCHAR2(64),
  Threshold                     VARCHAR2(64),
  Timeout                       NUMBER(19,0),
  Unsets                        VARCHAR2(1024),
  CONSTRAINT Triggers_PK        PRIMARY KEY (TrigID)
);

CREATE PUBLIC SYNONYM TRIGGERS FOR MOAB.TRIGGERS;


-- Moab virtual containers by name
--   Moab "current state" information
--   NOT PERSISTENT between Moab restarts
--   Possible ALTERs during Moab upgrades
CREATE TABLE VCs (
  Name                          VARCHAR2(64),
  Description                   VARCHAR2(1024),
  Jobs                          VARCHAR2(4000),
  Nodes                         VARCHAR2(4000),
  VMs                           VARCHAR2(4000),
  Rsvs                          VARCHAR2(4000),
  VCs                           VARCHAR2(4000),
  Variables                     VARCHAR2(4000),
  Flags                         VARCHAR2(4000),
  CONSTRAINT VCs_PK             PRIMARY KEY (Name)
);

CREATE PUBLIC SYNONYM VCS FOR MOAB.VCS;


--=======================================================
-- Moab stored procedure definitions
--
-- These procedures insert or update certain "state"
-- data in the historical tables to persist historical
-- information.  These procedures can be replaced by
-- Moab upgrades to reflect the current Moab version's
-- schema.
--=======================================================

-- Updates the JobHistory record
CREATE OR REPLACE PROCEDURE update_job_history
(
  NewID                         IN VARCHAR2,
  NewSourceRMJobID              IN VARCHAR2,
  NewDestinationRMJobID         IN VARCHAR2,
  NewGridJobID                  IN VARCHAR2,
  NewAName                      IN VARCHAR2,
  NewUserName                   IN VARCHAR2,
  NewAccount                    IN VARCHAR2,
  NewClass                      IN VARCHAR2,
  NewQOS                        IN VARCHAR2,
  NewOwnerGroup                 IN VARCHAR2,
  NewJobGroup                   IN VARCHAR2,
  NewState                      IN VARCHAR2,
  NewEState                     IN VARCHAR2,
  NewSubState                   IN VARCHAR2,
  NewUserPriority               IN NUMBER,
  NewSystemPriority             IN NUMBER,
  NewCurrentStartPriority       IN NUMBER,
  NewRunPriority                IN NUMBER,
  NewPerPartitionPriority       IN VARCHAR2,
  NewSubmitTime                 IN NUMBER, 
  NewQueueTime                  IN NUMBER,
  NewStartTime                  IN NUMBER,
  NewCompletionTime             IN NUMBER,
  NewCompletionCode             IN NUMBER,
  NewUsedWalltime               IN NUMBER,
  NewRequestedMinWalltime       IN NUMBER,
  NewRequestedMaxWalltime       IN NUMBER,
  NewCPULimit                   IN NUMBER,
  NewSuspendTime                IN NUMBER,
  NewHoldTime                   IN NUMBER,
  NewProcessorCount             IN NUMBER,
  NewRequestedNodes             IN NUMBER,
  NewActivePartition            IN VARCHAR2,
  NewSpecPAL                    IN VARCHAR2,
  NewDestinationRM              IN VARCHAR2,
  NewSourceRM                   IN VARCHAR2,
  NewFlags                      IN VARCHAR2,
  NewMinPreemptTime             IN NUMBER,
  NewDependencies               IN VARCHAR2,
  NewRequestedHostList          IN CLOB,
  NewExcludedHostList           IN CLOB,
  NewMasterHostName             IN VARCHAR2,
  NewGenericAttributes          IN VARCHAR2,
  NewHolds                      IN VARCHAR2,
  NewCost                       IN BINARY_DOUBLE,
  NewDescription                IN VARCHAR2,
  NewMessages                   IN VARCHAR2,
  NewNotificationAddress        IN VARCHAR2,
  NewStartCount                 IN NUMBER,
  NewBypassCount                IN NUMBER,
  NewCommandFile                IN VARCHAR2,
  NewArguments                  IN VARCHAR2,
  NewRMSubmitLanguage           IN VARCHAR2,
  NewStdIn                      IN VARCHAR2,
  NewStdOut                     IN VARCHAR2,
  NewStdErr                     IN VARCHAR2,
  NewRMOutput                   IN VARCHAR2,
  NewRMError                    IN VARCHAR2,
  NewInitialWorkingDirectory    IN VARCHAR2,
  NewUMask                      IN NUMBER,
  NewRsvStartTime               IN NUMBER,
  NewBlockReason                IN VARCHAR2,
  NewBlockMsg                   IN VARCHAR2,
  NewPSDedicated                IN BINARY_DOUBLE,
  NewPSUtilized                 IN BINARY_DOUBLE
)
IS
--  ErrMsg             VARCHAR2(512);
  
BEGIN
  UPDATE JobHistory 
    SET 
      SourceRMJobID             = NewSourceRMJobID,
      DestinationRMJobID        = NewDestinationRMJobID,
      GridJobID                 = NewGridJobID,
      AName                     = NewAName,
      UserName                  = NewUserName,
      Account                   = NewAccount,
      Class                     = NewClass,
      QOS                       = NewQOS,
      OwnerGroup                = NewOwnerGroup,
      JobGroup                  = NewJobGroup,
      State                     = NewState,
      EState                    = NewEState,
      SubState                  = NewSubState,
      UserPriority              = NewUserPriority,
      SystemPriority            = NewSystemPriority,
      CurrentStartPriority      = NewCurrentStartPriority,
      RunPriority               = NewRunPriority,
      PerPartitionPriority      = NewPerPartitionPriority,
      QueueTime                 = NewQueueTime,
      StartTime                 = NewStartTime,
      CompletionTime            = NewCompletionTime,
      CompletionCode            = NewCompletionCode,
      UsedWalltime              = NewUsedWalltime,
      RequestedMinWalltime      = NewRequestedMinWalltime,
      RequestedMaxWalltime      = NewRequestedMaxWalltime,
      CPULimit                  = NewCPULimit,
      SuspendTime               = NewSuspendTime,
      HoldTime                  = NewHoldTime,
      ProcessorCount            = NewProcessorCount,
      RequestedNodes            = NewRequestedNodes,
      ActivePartition           = NewActivePartition,
      SpecPAL                   = NewSpecPAL,
      DestinationRM             = NewDestinationRM,
      SourceRM                  = NewSourceRM,
      Flags                     = NewFlags,
      MinPreemptTime            = NewMinPreemptTime,
      Dependencies              = NewDependencies,
      RequestedHostList         = NewRequestedHostList,
      ExcludedHostList          = NewExcludedHostList,
      MasterHostName            = NewMasterHostName,
      GenericAttributes         = NewGenericAttributes,
      Holds                     = NewHolds,
      Cost                      = NewCost,
      Description               = NewDescription,
      Messages                  = NewMessages,
      NotificationAddress       = NewNotificationAddress,
      StartCount                = NewStartCount,
      BypassCount               = NewBypassCount,
      CommandFile               = NewCommandFile,
      Arguments                 = NewArguments,
      RMSubmitLanguage          = NewRMSubmitLanguage,
      StdIn                     = NewStdIn,
      StdOut                    = NewStdOut,
      StdErr                    = NewStdErr,
      RMOutput                  = NewRMOutput,
      RMError                   = NewRMError,
      InitialWorkingDirectory   = NewInitialWorkingDirectory,
      UMask                     = NewUMask,
      RsvStartTime              = NewRsvStartTime,
      BlockReason               = NewBlockReason,
      BlockMsg                  = NewBlockMsg,
      PSDedicated               = NewPSDedicated,
      PSUtilized                = NewPSUtilized
    WHERE ID = NewID AND SubmitTime = NewSubmitTime;
 
--EXCEPTION
--  WHEN OTHERS THEN
--    ErrMsg := SQLERRM;
--    DBMS_OUTPUT.PUT_LINE('Update JobHistory Error Message: ' || ErrMsg || ' - record not updated');
END;
/


-- Inserts a JobHistory table record
-- If a duplicate key error occurs, calls the
-- update_job_history procedure.
CREATE OR REPLACE PROCEDURE insert_job_history
(
  NewID                         IN VARCHAR2,
  NewSourceRMJobID              IN VARCHAR2,
  NewDestinationRMJobID         IN VARCHAR2,
  NewGridJobID                  IN VARCHAR2,
  NewAName                      IN VARCHAR2,
  NewUserName                   IN VARCHAR2,
  NewAccount                    IN VARCHAR2,
  NewClass                      IN VARCHAR2,
  NewQOS                        IN VARCHAR2,
  NewOwnerGroup                 IN VARCHAR2,
  NewJobGroup                   IN VARCHAR2,
  NewState                      IN VARCHAR2,
  NewEState                     IN VARCHAR2,
  NewSubState                   IN VARCHAR2,
  NewUserPriority               IN NUMBER,
  NewSystemPriority             IN NUMBER,
  NewCurrentStartPriority       IN NUMBER,
  NewRunPriority                IN NUMBER,
  NewPerPartitionPriority       IN VARCHAR2,
  NewSubmitTime                 IN NUMBER, 
  NewQueueTime                  IN NUMBER,
  NewStartTime                  IN NUMBER,
  NewCompletionTime             IN NUMBER,
  NewCompletionCode             IN NUMBER,
  NewUsedWalltime               IN NUMBER,
  NewRequestedMinWalltime       IN NUMBER,
  NewRequestedMaxWalltime       IN NUMBER,
  NewCPULimit                   IN NUMBER,
  NewSuspendTime                IN NUMBER,
  NewHoldTime                   IN NUMBER,
  NewProcessorCount             IN NUMBER,
  NewRequestedNodes             IN NUMBER,
  NewActivePartition            IN VARCHAR2,
  NewSpecPAL                    IN VARCHAR2,
  NewDestinationRM              IN VARCHAR2,
  NewSourceRM                   IN VARCHAR2,
  NewFlags                      IN VARCHAR2,
  NewMinPreemptTime             IN NUMBER,
  NewDependencies               IN VARCHAR2,
  NewRequestedHostList          IN CLOB,
  NewExcludedHostList           IN CLOB,
  NewMasterHostName             IN VARCHAR2,
  NewGenericAttributes          IN VARCHAR2,
  NewHolds                      IN VARCHAR2,
  NewCost                       IN BINARY_DOUBLE,
  NewDescription                IN VARCHAR2,
  NewMessages                   IN VARCHAR2,
  NewNotificationAddress        IN VARCHAR2,
  NewStartCount                 IN NUMBER,
  NewBypassCount                IN NUMBER,
  NewCommandFile                IN VARCHAR2,
  NewArguments                  IN VARCHAR2,
  NewRMSubmitLanguage           IN VARCHAR2,
  NewStdIn                      IN VARCHAR2,
  NewStdOut                     IN VARCHAR2,
  NewStdErr                     IN VARCHAR2,
  NewRMOutput                   IN VARCHAR2,
  NewRMError                    IN VARCHAR2,
  NewInitialWorkingDirectory    IN VARCHAR2,
  NewUMask                      IN NUMBER,
  NewRsvStartTime               IN NUMBER,
  NewBlockReason                IN VARCHAR2,
  NewBlockMsg                   IN VARCHAR2,
  NewPSDedicated                IN BINARY_DOUBLE,
  NewPSUtilized                 IN BINARY_DOUBLE
)
IS
--  ErrMsg             VARCHAR2(512);
  
  e_non_unique_error EXCEPTION;
  PRAGMA             EXCEPTION_INIT (e_non_unique_error, -1);
  
BEGIN
  INSERT INTO JobHistory (
    ID,
    SourceRMJobID,
    DestinationRMJobID,
    GridJobID,
    AName,
    UserName,
    Account,
    Class,
    QOS,
    OwnerGroup,
    JobGroup,
    State,
    EState,
    SubState,
    UserPriority,
    SystemPriority,
    CurrentStartPriority,
    RunPriority,
    PerPartitionPriority,
    SubmitTime,
    QueueTime,
    StartTime,
    CompletionTime,
    CompletionCode,
    UsedWalltime,
    RequestedMinWalltime,
    RequestedMaxWalltime,
    CPULimit,
    SuspendTime,
    HoldTime,
    ProcessorCount,
    RequestedNodes,
    ActivePartition,
    SpecPAL,
    DestinationRM,
    SourceRM,
    Flags,
    MinPreemptTime,
    Dependencies,
    RequestedHostList,
    ExcludedHostList,
    MasterHostName,
    GenericAttributes,
    Holds,
    Cost,
    Description,
    Messages,
    NotificationAddress,
    StartCount,
    BypassCount,
    CommandFile,
    Arguments,
    RMSubmitLanguage,
    StdIn,
    StdOut,
    StdErr,
    RMOutput,
    RMError,
    InitialWorkingDirectory,
    UMask,
    RsvStartTime,
    BlockReason,
    BlockMsg,
    PSDedicated,
    PSUtilized)
  VALUES (
    NewID,
    NewSourceRMJobID,
    NewDestinationRMJobID,
    NewGridJobID,
    NewAName,
    NewUserName,
    NewAccount,
    NewClass,
    NewQOS,
    NewOwnerGroup,
    NewJobGroup,
    NewState,
    NewEState,
    NewSubState,
    NewUserPriority,
    NewSystemPriority,
    NewCurrentStartPriority,
    NewRunPriority,
    NewPerPartitionPriority,
    NewSubmitTime,
    NewQueueTime,
    NewStartTime,
    NewCompletionTime,
    NewCompletionCode,
    NewUsedWalltime,
    NewRequestedMinWalltime,
    NewRequestedMaxWalltime,
    NewCPULimit,
    NewSuspendTime,
    NewHoldTime,
    NewProcessorCount,
    NewRequestedNodes,
    NewActivePartition,
    NewSpecPAL,
    NewDestinationRM,
    NewSourceRM,
    NewFlags,
    NewMinPreemptTime,
    NewDependencies,
    NewRequestedHostList,
    NewExcludedHostList,
    NewMasterHostName,
    NewGenericAttributes,
    NewHolds,
    NewCost,
    NewDescription,
    NewMessages,
    NewNotificationAddress,
    NewStartCount,
    NewBypassCount,
    NewCommandFile,
    NewArguments,
    NewRMSubmitLanguage,
    NewStdIn,
    NewStdOut,
    NewStdErr,
    NewRMOutput,
    NewRMError,
    NewInitialWorkingDirectory,
    NewUMask,
    NewRsvStartTime,
    NewBlockReason,
    NewBlockMsg,
    NewPSDedicated,
    NewPSUtilized);
--  DBMS_OUTPUT.PUT_LINE('Insert JobHistory successful');
  
EXCEPTION
  WHEN e_non_unique_error THEN      
--    ErrMsg := SQLERRM;
--    DBMS_OUTPUT.PUT_LINE('Insert JobHistory failed - Error Message: ' || ErrMsg);
--    DBMS_OUTPUT.PUT_LINE('Attempting to update JobHistory');
    update_job_history
      (
        NewID,
        NewSourceRMJobID,
        NewDestinationRMJobID,
        NewGridJobID,
        NewAName,
        NewUserName,
        NewAccount,
        NewClass,
        NewQOS,
        NewOwnerGroup,
        NewJobGroup,
        NewState,
        NewEState,
        NewSubState,
        NewUserPriority,
        NewSystemPriority,
        NewCurrentStartPriority,
        NewRunPriority,
        NewPerPartitionPriority,
        NewSubmitTime,
        NewQueueTime,
        NewStartTime,
        NewCompletionTime,
        NewCompletionCode,
        NewUsedWalltime,
        NewRequestedMinWalltime,
        NewRequestedMaxWalltime,
        NewCPULimit,
        NewSuspendTime,
        NewHoldTime,
        NewProcessorCount,
        NewRequestedNodes,
        NewActivePartition,
        NewSpecPAL,
        NewDestinationRM,
        NewSourceRM,
        NewFlags,
        NewMinPreemptTime,
        NewDependencies,
        NewRequestedHostList,
        NewExcludedHostList,
        NewMasterHostName,
        NewGenericAttributes,
        NewHolds,
        NewCost,
        NewDescription,
        NewMessages,
        NewNotificationAddress,
        NewStartCount,
        NewBypassCount,
        NewCommandFile,
        NewArguments,
        NewRMSubmitLanguage,
        NewStdIn,
        NewStdOut,
        NewStdErr,
        NewRMOutput,
        NewRMError,
        NewInitialWorkingDirectory,
        NewUMask,
        NewRsvStartTime,
        NewBlockReason,
        NewBlockMsg,
        NewPSDedicated,
        NewPSUtilized
      );
      
--  WHEN OTHERS THEN
--    ErrMsg := SQLERRM;
--    DBMS_OUTPUT.PUT_LINE('Insert JobHistory failed - Error Message: ' || ErrMsg);
END;
/



--=======================================================
-- Moab table trigger definitions
--
-- These triggers copy certain "state" data into the
-- historical tables to persist historical information.
-- These triggers can be replaced by Moab upgrades to
-- reflect the current Moab version's schema.
--=======================================================

-- Create a JobHistory table record if a completed job.
CREATE OR REPLACE TRIGGER job_history_trig
  BEFORE INSERT ON Jobs
    FOR EACH ROW WHEN (NEW.CompletionTime > 0)
    BEGIN
      insert_job_history
        (
          :NEW.ID,
          :NEW.SourceRMJobID,
          :NEW.DestinationRMJobID,
          :NEW.GridJobID,
          :NEW.AName,
          :NEW.UserName,
          :NEW.Account,
          :NEW.Class,
          :NEW.QOS,
          :NEW.OwnerGroup,
          :NEW.JobGroup,
          :NEW.State,
          :NEW.EState,
          :NEW.SubState,
          :NEW.UserPriority,
          :NEW.SystemPriority,
          :NEW.CurrentStartPriority,
          :NEW.RunPriority,
          :NEW.PerPartitionPriority,
          :NEW.SubmitTime,
          :NEW.QueueTime,
          :NEW.StartTime,
          :NEW.CompletionTime,
          :NEW.CompletionCode,
          :NEW.UsedWalltime,
          :NEW.RequestedMinWalltime,
          :NEW.RequestedMaxWalltime,
          :NEW.CPULimit,
          :NEW.SuspendTime,
          :NEW.HoldTime,
          :NEW.ProcessorCount,
          :NEW.RequestedNodes,
          :NEW.ActivePartition,
          :NEW.SpecPAL,
          :NEW.DestinationRM,
          :NEW.SourceRM,
          :NEW.Flags,
          :NEW.MinPreemptTime,
          :NEW.Dependencies,
          :NEW.RequestedHostList,
          :NEW.ExcludedHostList,
          :NEW.MasterHostName,
          :NEW.GenericAttributes,
          :NEW.Holds,
          :NEW.Cost,
          :NEW.Description,
          :NEW.Messages,
          :NEW.NotificationAddress,
          :NEW.StartCount,
          :NEW.BypassCount,
          :NEW.CommandFile,
          :NEW.Arguments,
          :NEW.RMSubmitLanguage,
          :NEW.StdIn,
          :NEW.StdOut,
          :NEW.StdErr,
          :NEW.RMOutput,
          :NEW.RMError,
          :NEW.InitialWorkingDirectory,
          :NEW.UMask,
          :NEW.RsvStartTime,
          :NEW.BlockReason,
          :NEW.BlockMsg,
          :NEW.PSDedicated,
          :NEW.PSUtilized
        );
--      EXCEPTION
--      WHEN OTHERS THEN
--        NULL;
    END;
/
