import os, sys, re

testable_pat = re.compile('moab-db-([a-z]+)-create.sql$', re.IGNORECASE)
table_pat = re.compile(r'\Wcreate\s+table\s+([a-z_]+)[^(]*\((.*?)\)\s*;', re.IGNORECASE | re.DOTALL)
end_of_field_pat = re.compile(r',\s*\n')
fieldname_pat = re.compile(r'^(?:[a-z_]+|"[a-z_]+"|`[a-z_]+`|' + "'" + '[a-z_]+' + "'" + ')\s+.*')
ws_pat = re.compile(r'\s+')

def load_schema(f):
    schema = {}
    f = open(f, 'r')
    try:
        txt = f.read()
    finally:
        f.close()
    for match in table_pat.finditer(txt):
        table_name = match.group(1).lower()
        body = match.group(2)
        fields = end_of_field_pat.split(body)
        fields = [x.strip().lower() for x in fields]
        fields = [x for x in fields if fieldname_pat.match(x)]
        fields = [x[0:ws_pat.search(x).start()] for x in fields]
        fields = [x.replace('[','').replace(']','').replace('"','').replace("'",'').replace('`','') for x in fields]
        fields = [x for x in fields if x not in ['constraint', 'primary', 'foreign']]
        schema[table_name] = fields
    return schema

def show_diff_fields(tbl, master, other, a_tbl, b_tbl):
    sys.stderr.write('\nFields in table %s differ:\n\n%s  %s\n------------------------------  ------------------------------\n' % 
                     (tbl.upper(), master.ljust(30), other.ljust(30)))
    for i in xrange(500):
        if i >= len(a_tbl):
            a = ''
        else:
            a = a_tbl[i]
        if i >= len(b_tbl):
            b = ''
        else:
            b = b_tbl[i]
        if not a and not b:
            break
        mark = ''
        if a != b:
            a = a.upper()
            b = b.upper()
            mark = '***'
        sys.stderr.write('%s  %s %s\n' % (a.ljust(30), b.ljust(30), mark))    

def find_diff(schemas, master, other):
    result = 0
    schema_a = schemas[master]
    schema_b = schemas[other]
    for tbl in schema_a.keys():
        if tbl in schema_b:
            a = str(schema_a[tbl])
            b = str(schema_b[tbl])
            if (a != b):
                show_diff_fields(tbl, master, other, schema_a[tbl], schema_b[tbl])
                result = 1
        else:
            sys.stderr.write('\nTable %s is present in master schema (%s), but missing from %s.\n' % (tbl, master, other))
            result = 1
    for tbl in schema_b.keys():
        if tbl not in schema_a.keys():
            sys.stderr.write('\nTable %s is missing from master schema (%s), but present in %s.\n' % (tbl, master, other))
            result = 1
    return result

def test_schema_sync():
    problem = 0
    if len(sys.argv) > 1:
        folder = sys.argv[1]
    else:
        folder = '.'
    items = [x for x in os.listdir(folder) if testable_pat.match(x)]
    schemas = {}
    loaded = []
    for item in items:
        dbms = testable_pat.match(item).group(1)
        fpath = os.path.abspath(os.path.join(folder, item))
        schema = load_schema(fpath)
        if not schema:
            sys.stderr.write('No schema loaded from %s.\n' % fpath)
            problem = 1
        else:
            schemas[dbms] = schema
            loaded.append(fpath)
    print(loaded)
    if len(loaded) < 2 or ('mysql' not in schemas):
        loaded_txt = str(loaded).replace('[','').replace(']','').replace(',', '\n  ').replace("'",'')
        sys.stderr.write('Expected to load at least 2 schemas, including mysql. Instead, loaded:\n   %s\n' % loaded_txt)
        problem = 1
    else:
        master = schemas['mysql']
        for dbms in schemas.keys():
            if dbms != master:
                if find_diff(schemas, 'mysql', dbms):
                    problem = 1
    if problem:
        sys.stderr.write('\nFAIL. Cannot prove schemas are in sync.\n')
    else:
        print('\nSUCCESS. %d schemas are in sync.' % len(schemas))
    return problem

if __name__ == '__main__':
    sys.exit(test_schema_sync())