#!/usr/bin/python

#THIS SOFTWARE SCRIPT IS PROVIDED BY ADAPTIVE COMPUTING ENTERPRISES, INC. "AS IS"
#AND IS FOR REFERENCE USE ONLY.  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
#NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
#PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ADAPTIVE COMPUTING ENTERPRISES, INC.
#BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#DAMAGES ARISING FROM THE USE THEREOF.

"""
This module contains a library of routines that can be used by various
data staging utilities.  They include parsing routines as well as remote
shell copying routines.
"""

import re
import string
import sys
import xml.etree.ElementTree as ET
import paramiko
import subprocess
from optparse import OptionParser
import ds_config


def ParseDS(string):
  """
  Parse a data staging strings into source and destination parts.
  The two separate parts are returned.
  Raises an exception if there are problems.
  """
  parts = string.split('%')
  if len(parts) != 2:
    raise Exception("Expecting two parts separated by '%%'.  Instead, found %d" % (parts.len))
  try:
    src = ParseLocation(parts[0])
  except Exception as e:
    raise Exception("Error parsing substring '%s': (%s)" % (parts[0], str(e)))
  try:
    dst = ParseLocation(parts[1])
  except Exception as e:
    raise Exception("Error parsing substring '%s': (%s)" % (parts[1], str(e)))

  # check to see that we are not trying to go from a directory to a file.  (other combos are OK)
  if not src.filename and dst.filename:
    raise Exception("Cannot stage from a directory to a file. '%s'" % (string))

  return (src,dst)

def ParseLocation(string):
  """
  Parse a single location string and create a Location object.
  The string is of the form user@host:/path/filename
  user is optional
  host is not optional
  path is not optional (and must end with a trailing '/')
  filename is optional
  """
  user = None
  host = None
  path = None
  filename = None
  m = re.search("^(.*@)*(.*:)*(.*/)(.*)$",string)
  if m:
    if m.group(1):
      user     = m.group(1).rstrip('@')
    if m.group(2):
      host     = m.group(2).rstrip(':')
    else:
       raise Exception("Missing host component.")
    if m.group(3):
      path     = m.group(3)
    else:
       raise Exception("Missing path component.")
    if m.group(4):
      filename = m.group(4)
    return Location(user, host, path, filename)

  #if we reach this there was a parsing error:
  raise Exception("Unable to parse '%s'." % (string))

class Location:
  """
  A simple container for information describing a data source/destination.
  """
  def __init__(self, _user=None, _host=None, _path=None, _filename=None):
    self.user = _user
    self.host = _host
    self.path = _path
    self.filename = _filename


def ValidateSSHConnectivity(srcdst):
  """
  Validate that it is possible to connect via SSH from this server
  to the source server and then to the destination server.   This is the
  pathway used for both rsync and scp move functions.
  """
  (src, dst) = ParseDS(srcdst)

  # Build the command string that will connect from src to dst
  command = "ssh"

  # From the src server make sure that you have a host key and can connect
  # to the dst server.
  # (this has the side effect of storing the host key if it isn't present.)
  command += " -o StrictHostKeyChecking=no"

  command += ' '
  if dst.user:
    command += dst.user
    command += '@'
  if dst.host:    # should always have a host???
    command += dst.host

  # add a dummy/do-nothing command to the ssh command..
  command += ' /bin/true'


  # Open a ssh connection to the src server
  ssh = paramiko.SSHClient()
  # Auto-add a host key if the one presented is not recognized.  (needed for non-interactive)
  ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
  try:
    ssh.connect(src.host, username=src.user)
  except Exception as e:
    raise Exception("Error while creating ssh connection to: '%s' as user: '%s': (%s)" % (src.host, src.user, str(e)))

  # From the src server, run the command that connects to the dst server.
  try:
    stdin, stdout, stderr = ssh.exec_command(command)
    rc = stdout.channel.recv_exit_status()
    if rc != 0:
      # TODO: all kinds of error checking need to be added here.
      raise Exception("ssh exit code: %s, stderr: '%s'" % (rc, stderr.read()))
  except Exception as e:
    raise Exception("Error running: '%s' on host: '%s' as user: '%s': (%s)" % (command, src.host, src.user, str(e)))

  result = stdout.read()
  return result


def MoveDataRsync(srcdst, dryrun = True):
  """
  Copy the data using rsync from the source to the destination
  srcdst is the data staging string in the form "<SRC>%<DST>".
  dryrun indicates whether to do the transfer for real.
  Returns a tuple composed of:
    1) The stdout from the rsync command.
    2) The actual number of bytes transferred.
    3) The total size (uncompressed & un-replicated) of files to transfer.
  """
  (src, dst) = ParseDS(srcdst)

  # Build the command string that will push files from src to dst
  command = "rsync -avz"
  if dryrun:
    command += 'n'

  # From the src server make sure that you have a host key and can connect
  # to the dst server.
  # (Do this by modifying the ssh call used for rsync transport.
  #  it has the side effect of storing the host key if it isn't present.)
  command += " --rsh='ssh -o StrictHostKeyChecking=no'"

  command += ' '
  # NOTE: don't use user@host: here because we are already on that server as that user.
  command += src.path  # should always have a path
  if src.filename:
    command += src.filename

  command += ' '

  if dst.user:
    command += dst.user
    command += '@'
  if dst.host:    # should always have a host???
    command += dst.host
    command += ':'
  command += dst.path  # should always have a path
  if dst.filename:
    command += dst.filename

  # Open a ssh connection to the src server
  ssh = paramiko.SSHClient()
  # Auto-add a host key if the one presented is not recognized.  (needed for non-interactive)
  ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
  try:
    ssh.connect(src.host, username=src.user)
  except Exception as e:
    raise Exception("Error while creating ssh connection to: '%s' as user: '%s': (%s)" % (src.host, src.user, str(e)))

  # From the src server, run the command that pushes data to the dst server.
  try:
    stdin, stdout, stderr = ssh.exec_command(command)
    rc = stdout.channel.recv_exit_status()
    if rc != 0:
      # TODO: all kinds of error checking may be added here.
      raise Exception("Rsync exit code: %s, stderr: '%s'" % (rc, stderr.read()))
  except Exception as e:
    raise Exception("Error running: '%s' on host: '%s' as user: '%s': (%s)" % (command, src.host, src.user, str(e)))


  # TODO: (maybe) stdout.read() blocks until all the data is present.
  #       we may need to use readlines() to clear the buffer and get updated results along the way....
  #for line in stderr.readlines():
  #  print line
  #for line in stdout.readlines():
  #  result += line

  result = stdout.read()

  transfer_size = 0
  m = re.search(".*sent ([0-9]+) bytes.*",result)
  if m:
    if m.group(1):
      transfer_size = int(m.group(1))

  total_size = 0
  m = re.search(".*total size is ([0-9]+) .*",result)
  if m:
    if m.group(1):
      total_size = int(m.group(1))

  return (result, transfer_size, total_size)



def MoveDataSCP(srcdst):
  """
  Copy the data using scp from the source to the destination.
  srcdst is the data staging string in the form "<SRC>%<DST>"
  the stdout from the scp command will be returned.
  """
  (src, dst) = ParseDS(srcdst)

  # Build the scp command string that will push files from src to dst
  #  -p      Preserves modification times, access times, and modes from the original file.
  #  -r      Recursively copy entire directories.  Note that scp follows symbolic links encountered in the
  #          tree traversal.
  #  -o ssh_option
  #          Can be used to pass options to ssh in the format used in ssh_config(5).  This is useful for
  #          specifying options for which there is no separate scp command-line flag.  For full details of
  #          the options listed below, and their possible values, see ssh_config(5).

  command = "scp -pr"
  command += " -o StrictHostKeyChecking=no"

  # create the source and destination arguments
  command += ' '
  if src.user:
    command += src.user
    command += '@'
  if src.host:    # should always have a host???
    command += src.host
    command += ':'
  command += src.path  # should always have a path
  if src.filename:
    command += src.filename

  command += ' '

  if dst.user:
    command += dst.user
    command += '@'
  if dst.host:    # should always have a host???
    command += dst.host
    command += ':'
  command += dst.path  # should always have a path
  if dst.filename:
    command += dst.filename

  p = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
  (stdoutdata,stderrdata) = p.communicate()
  if p.returncode != 0:
    raise Exception("Error copying data using command: '%s' returncode: '%s' stderr:(%s)" %
                    (command, str(p.returncode), stderrdata))
  return stdoutdata



def FilterXML(xmlstring):
  """
  xmlstring should be the job submission string as constructed by msub

  parse the XML looking for:
  1) A <stageinsize></stageinsize> entity.
       If this entity is found, no alteration is needed.
       If multiple entites are found, this is an error condition.

  2) A <stageoutsize></stageoutsize> entity.
       If multiple entites are found, this is an error condition.

  3) Zero or more <stageout></stageout> entities.
     Attempt to connect to assure that hosts are reachable.

  4) Zero or more <stagein></stagein> entities.
     If these are found, the size of the source file needs to be
     determined and then that size should be inserted into a <stageinsize>
     entity in the returned XML.
     Also attempt to connect to assure that hosts are reachable.

  5) If a <stageoutsize> is present then there must also be a <stageout>.

  6) If a <stageout> is present then there must also be a <stageoutsize>.

  7) If a <stageinsize> is present then there must also be a <stagein>.

  8) If a <stageinsize> and <stageoutsize> are present, they must be
     properly formatted. i.e. "^[0-9]+[kKmMgGtT]?[bB]"

  9) If <stagein> and/or <stageout> were detected, insert (or append) a
     default data staging job template if needed.
  """

  root = ET.fromstring(xmlstring)

  # ensure only one stageinsize parameter exists
  stageinsize_count = 0
  for sizein in root.getiterator('stageinsize'):
    stageinsize_count += 1
  if stageinsize_count > 1:
    raise Exception("Multiple 'stageinsize' parameters specified.  Only one is allowed.")

  # ensure only one stageoutsize parameter exists
  stageoutsize_count = 0
  for sizeout in root.getiterator('stageoutsize'):
    stageoutsize_count += 1
  if stageoutsize_count > 1:
    raise Exception("Multiple 'stageoutsize' parameters specified.  Only one is allowed.")

  # loop through the stageout entries and test that they are accessible.
  stageout_count=0
  for stageout in root.getiterator('stageout'):
    stageout_count += 1
    if re.search(r"^.*\$JOBHOST.*", stageout.text) or re.search(r"^.*\$CLUSTERHOST.*", stageout.text):
      # We cannot reliably test accessibility until the hosts are known. Skip for now.
      # TODO: log that we are skipping?
      continue
    try:
      # This validates connectivity
      result = ValidateSSHConnectivity(stageout.text)
    except Exception as e:
      raise Exception("Error while checking accessibility of stageout '%s': (%s)" % (stageout.text, str(e)))


  # loop through the stagein entries and test that they are accessible.
  computed_stageinsize = 0
  stagein_count = 0
  for stagein in root.getiterator('stagein'):
    size = 0
    stagein_count += 1
    if re.search(r"^.*\$JOBHOST.*", stagein.text) or re.search(r"^.*\$CLUSTERHOST.*", stagein.text):
      # We cannot reliably test accessibility until the hosts are known. Skip for now.
      # TODO: log that we are skipping?
      continue
    try:
      (result, bytes_transferred, size) = MoveDataRsync(stagein.text, dryrun=True)
    except Exception as e:
      raise Exception("Error while checking accessibility of stagein '%s': (%s)" % (stagein.text, str(e)))

    # Examine the output and keep track of how much data would have been transferred on a real run.
    computed_stageinsize += size

  # If stageoutsize specified make sure there is a stageout.
  if stageoutsize_count > 0 and stageout_count < 1:
    raise Exception("No <stageout> provided with specified <stageoutsize>.")

  # If stageout is specified make sure that stageoutsize is also specified.
  if stageout_count > 0 and stageoutsize_count < 1:
    raise Exception("No <stageoutsize> provided with specified <stageout>.")

  # If stageinsize is specified make sure there is a stagein.
  if stageinsize_count > 0 and stagein_count < 1:
    raise Exception("No <stagein> provided with specified <stageinsize>.")

  if stageinsize_count < 1 and stagein_count > 0:
    # We need to add a <stageinsize> entity to the XML.
    if computed_stageinsize == 0:
      # This could be because failures above (specifically skipping because of $JOBHOST | $CLUSTERHOST).
      raise Exception("No <stageinsize> was input and the value could not be determined automatically--please specify manually.")

    # Add a stageinsize subelement to root.
    sub = ET.SubElement(root, 'stageinsize')
    sub.text = repr(computed_stageinsize) + 'b'
    stagein_count += 1

  # Ensure stageinsize is correctly formatted.
  if stagein_count > 0:
    size_element = root.find('stageinsize')
    if not (re.search(r"^[0-9]+[kKmMgGtT]?[bB]", size_element.text) or re.search(r"^[0-9]+$", size_element.text)):
      raise Exception("Illegal stagein format (%s)." % (size_element.text))

  # ensure stageoutsize is correctly formatted
  if stageout_count > 0:
    size_element = root.find('stageoutsize')
    if not (re.search(r"^[0-9]+[kKmMgGtT]?[bB]", size_element.text) or re.search(r"^[0-9]+$", size_element.text)):
      raise Exception("Illegal stageout format (%s)." % (size_element.text))

  # add in a template if needed
  #  if there is not a default defined, just skip the potential addition and let the server catch it
  if (stagein_count + stageout_count > 0) and hasattr(ds_config, 'DEFAULT_TEMPLATE'):
    # we have a data staging job so check for a template

    # first, find Extension element
    extension_element = root.find('Extension')

    # see if it contains a value
    if extension_element is None:
      # no template was specified so insert one
      sub = ET.SubElement(root, 'Extension')
      sub.text = "x=template:" + ds_config.DEFAULT_TEMPLATE
    elif not re.search(r"template:", extension_element.text):
      # extension element exists but contains no template so append one
      extension_element.text += ";template:" + ds_config.DEFAULT_TEMPLATE

  output = ET.tostring(root)
  return output


def GetMoveDirectives(argv, xmlstream):
  """
  Uses the commandline arguments to determine whether this is
  being invoked as a --stagein or --stageout.  It then parses
  the appropriate stage* move directives from the XML and
  returns them as a list of strings.
  """

  usage = "usage: %prog [-i|--stagein] | [-o|--stageout] | [[-t|--test] TEST_SRCDST]"
  parser = OptionParser()
  parser.add_option("-i", "--stagein",
                    action="store_true",
                    dest="dostagein",
                    help="Select files for 'StageIn' from the XML.",
                    default=False)
  parser.add_option("-o", "--stageout",
                    action="store_true",
                    dest="dostageout",
                    help="Select files for 'StageOut' from the XML.",
                    default=False)
  parser.add_option("-t", "--test",
                    action="store",
                    type="string",
                    dest="test_srcdst",
                    help="Test run with a single argument from the commandline.")
  (options, args) = parser.parse_args(argv[1:])

  returnlist = []
  if options.dostagein and options.dostageout:
    sys.stderr.write("Cannot specify both --stagein and --stageout.\n")
    parser.print_help()
    sys.exit(1)
  if options.test_srcdst:
    returnlist.append(options.test_srcdst)
    # at this point we are done--don't try to parse any XML.
    return returnlist
  search_element_name=""
  if options.dostagein:
    search_element_name="StageIn"
  if options.dostageout:
    search_element_name="StageOut"

  # now search the XML for all the elements that match.
  xmlstring = xmlstream.read(-1)
  root = ET.fromstring(xmlstring)
  for stage in root.getiterator(search_element_name):
    returnlist.append(stage.text)
  ExpandJobHost(root,returnlist)
  ExpandClusterHost(root,returnlist)
  return returnlist



def ExpandJobHost(xmltree,srcdstlist):
  """
  Replace any $JOBHOST strings in the source/destination list with the first node in the AllocNodeList.
  """
  try:
    found = False
    for i in xrange (0,len(srcdstlist)):
      if re.search(r"^.*\$JOBHOST.*", srcdstlist[i]):
        found = True
        break

    if not found:
      return

    # AllocNodeList XML format is 'AllocNodeList="node01,node02,node03"'
    # Sometimes a task count is added: 'AllocNodeList="node01:2,node02:2,node03:2"'
    # in this case we are only expecting a single node, but code defensively.
    nodelist = xmltree.get('AllocNodeList')
    node = nodelist.split(',')
    node = node[0].split(':')
    for i in xrange (0,len(srcdstlist)):
      srcdstlist[i] = srcdstlist[i].replace('$JOBHOST',node[0])
  except Exception as e:
    raise Exception("Could not expand $JOBHOST (%s)." % (str(e)))


def GetPartition(xmltree):
  """
  Find the partition associated with this job.  This funciton utilizes the XML
  associated with a job and examines that to determine the partition.
  """
  partition = None

  # First, attempt to determine the partition from the DATA_STAGING_PARTITION variable if present...
  variables = xmltree.find('Variables')
  if variables is not None:
    for variable in variables.getiterator('Variable'):
      if variable.get('name') == 'DATA_STAGING_PARTITION':
        partition = variable.text

  # ...next fallback to the partition list.
  if partition is None:
    # PAL XML format is 'PAL="foo,bar,baz"' although in this case we really only
    # expect one partition to be present, but code defensively.
    pal = xmltree.get('PAL')
    partition = pal.split(',')[0]

  return partition



def ExpandClusterHost(xmltree,srcdstlist):
  """
  Replace any $CLUSTERHOST strings in the source/destination list with the fileserver for the
  first partition in the PAL.  The mapping between partition name and fileserver name is
  contained in PARTITION_TO_HOST.
  """
  try:
    found = False
    for i in xrange (0,len(srcdstlist)):
      if re.search(r"^.*\$CLUSTERHOST.*", srcdstlist[i]):
        found = True
        break

    if not found:
      return

    partition = GetPartition(xmltree)

    if partition not in ds_config.PARTITION_TO_HOST:
      raise Exception("Partition '%s' not found in PARTITION_TO_HOST mapping--Check datastaging script configuration." % (partition))
    for i in xrange (0,len(srcdstlist)):
      srcdstlist[i] = srcdstlist[i].replace('$CLUSTERHOST',ds_config.PARTITION_TO_HOST[partition])
  except Exception as e:
    raise Exception("Could not expand $CLUSTERHOST (%s)." % (str(e)))



def DoMoveRsyncMain():
  """
  Contains the body of the move rsync script.
  """

  movelist = GetMoveDirectives(sys.argv, sys.stdin)

  if len(movelist) == 0:
    # Nothing to do--return without error.
    return

  total_bytes = 0
  total_size = 0
  for srcdst in movelist:
    output = ""
    bytes_transferred = 0
    size = 0
    try:
      (output, bytes_transferred, size) = MoveDataRsync(srcdst, dryrun = False)
    except Exception as e:
      raise Exception("Could not move data: '%s'  (%s)" % (srcdst, str(e)))
    sys.stdout.write(output)
    total_bytes += bytes_transferred
    total_size += size

  sys.stdout.write("""######End of Rsync output.
# The following four variables may be interpreted by the trigger 'Sets'
# functionality to import the variable onto the Moab job.  Select the
# appropriate variable (*_IN or *_OUT) depending on whether the trigger
# is for a stage-in or a stage-out job.
# (Because the same script is used for both in/out jobs both variables are always available)
#######\n""")
  sys.stdout.write("BYTES_IN=%d\n" % total_bytes)
  sys.stdout.write("BYTES_OUT=%d\n" % total_bytes)
  sys.stdout.write("DATA_SIZE_IN=%d\n" % total_size)
  sys.stdout.write("DATA_SIZE_OUT=%d\n" % total_size)


def DoMoveSCPMain():
  """
  Contains the body of the move scp script.
  """

  movelist = GetMoveDirectives(sys.argv, sys.stdin)

  if len(movelist) == 0:
    # nothing to do--return without error.
    return

  for srcdst in movelist:
    output = ""
    try:
      output = MoveDataSCP(srcdst)
    except Exception as e:
      raise Exception("Could not move data: '%s'  (%s)." % (srcdst, str(e)))
    # scp does not write output to stdout with no tty so the following is
    # usually a no-op except when run in test mode.
    sys.stdout.write(output)



def DoMoveMultiplexMain():
  """
  Contains the body of the move multiplex script.
  """

  xmlstring = sys.stdin.read(-1)
  xmltree = ET.fromstring(xmlstring)
  partition = GetPartition(xmltree)

  # Give a fallback as "DEFAULT" if it is not in the dictionary...
  if partition not in ds_config.PARTITION_TO_SCRIPT:
    partition = "DEFAULT"
  # Allow for the case where the admin has removed "DEFAULT" from the dictionary...
  if partition not in ds_config.PARTITION_TO_SCRIPT:
    raise Exception("Partition '%s' not found in ds_config.PARTITION_TO_SCRIPT mapping--Check datastaging script configuration." % (partition))

  script = ds_config.PARTITION_TO_SCRIPT[partition]



  # Replace the new script name in place of this name.  All other arguments remain the same.
  sys.argv[0] = script
  p = subprocess.Popen(sys.argv, shell=False, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

  # Re-inject the stdin back into the child process.
  (stdoutdata,stderrdata) = p.communicate(input=xmlstring)
  rc = p.returncode
  sys.stdout.write(stdoutdata)
  sys.stderr.write(stderrdata)
  sys.exit(rc)


